/*
  fan.cpp

  Here we implement a function that outputs, in sparse SDPA format, an
  SDP whose optimal value is the sum of the k largest eigenvalues of a
  symmetric matrix.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <sdpsl.h>

using namespace std;
using namespace sdp;

//
// Prototypes
//
static void output_fan_sdp(const Matrix<double>& A, int k);

/*
  main
  
  We apply function output_fan_sdp to a sample matrix.
*/
int main()
{
  Matrix<double> A({ { -2,  4,  3, -5, -1, -6 }, 
      	             {  4,  8,  0,  0,  6,  3 }, 
		     {  3,  0,  0,  0, -5,  3 }, 
		     { -5,  0,  0, -2, -2, -4 }, 
		     { -1,  6, -5, -2, -2,  1 }, 
		     { -6,  3,  3, -4,  1, -2 } });
  
  output_fan_sdp(A, 3);  // sum of 3 largest eigenvalues

  return 0;
}

/*
  output_fan_sdp

  Outputs an SDP whose optimal value is the sum of the k largest
  eigenvalues of symmetric matrix A.
*/
void output_fan_sdp(const Matrix<double>& A, int k)
{
  int n = A.num_rows();
  SDPVariable X(n), Y(n);
  SDPConstraintContainer cc;

  /* First the trace constraint for X */
  cc.add_constraint(Matrix<double>::identity(n) * X == k);

  /* Then the correspondence between X and Y */
  Matrix<double> E(n, n, 1);
  MatrixMask E_mask(n);

  for (int i = 0; i < n; i++)
    for (int j = i; j < n; j++) {
      double rhs = i == j ? 1 : 0;

      E_mask.turn_on(i, j);
      cc.add_constraint(SDPMaskedMatrix(E, E_mask) * X
			+ SDPMaskedMatrix(E, E_mask) * Y == rhs);
      E_mask.turn_off(i, j);      
    }

  /* Then we create the problem and output it */
  SDPProblem problem;

  problem.set_objective(A * X);
  problem << cc;

  cout << SDPAWriter(problem);
}
