/*
  spack.cpp

  A semidefinite programming approach to compute the Cohn-Elkies
  sphere-packing bound.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <fstream>
#include <mpfsdpsl.h>

using namespace std;
using namespace mpf_sdp;

//
// This class does all the work of problem generation, etc.
//
class Kernel {
  /* Problem data */
  int n;         // Dimension of the space
  int deg;       // Degree of polynomials to be used
  mpf_class pi;  // An approximation of pi

  /* Bases of polynomials, etc. */
  PolynomialVariable x;

  SparseMonomialBasis mon_basis;     // Basis of monomials we work with
  vector<Polynomial> lag_basis;      // Basis of Laguerre polynomials
  vector<Polynomial> fourier_polys;  // Polynomials for computing the Fourier inverse

  /* Moment matrices & co. */
  Matrix<Polynomial> V;     // Moment matrix
  Matrix<Polynomial> F;     // Inverses of V
  Matrix<Polynomial> x2_F;  // Inverses of x^2 F

  /* SDP stuff */
  SDPVariable S0, S1, Q;  // Variables of the problem

  /* Utility functions */
  Polynomial fourier_inverse(const Polynomial& p) const;
  void make_constraints(SDPConstraintContainer& cc) const;
  SDPLinearCombination make_objective(int fac) const;
  void show_info(const SDPSolution<mpf_class>& sol, const SDPLinearCombination& obj,
		 const SDPConstraintContainer& cc, ostream& out) const;
    
public:
  Kernel(int _n, int d);

  void output_opt(ostream& out) const;
  void output_feas(istream& in, double eps, ostream& out) const;
  void project(istream& in) const;
};

//
// Prototypes
//
static void get_and_apply_opts(int argc, char **argv);
static void show_usage(const char *prog);

//
// Functions
//

/*
  This function parses options from the command line and applies them.
*/
void get_and_apply_opts(int argc, char **argv)
{
  int prec = 256;
  double output_eps = 1e-15;
  double tolerance_eps = 1e-15;

  if (argc > 0) {
    for (int i = 0; i < argc; i += 2) {
      if (strcmp(argv[i], "-p") == 0) {
	if (i + 1 == argc) show_usage(argv[0]);

	prec = atoi(argv[i + 1]);

	if (prec <= 53) show_usage(argv[0]);
      }
      else if (strcmp(argv[i], "-oeps") == 0) {
	if (i + 1 == argc) show_usage(argv[0]);

	output_eps = atof(argv[i + 1]);

	if (output_eps < 0) show_usage(argv[0]);
      }
      else if (strcmp(argv[i], "-teps") == 0) {
	if (i + 1 == argc) show_usage(argv[0]);

	tolerance_eps = atof(argv[i + 1]);

	if(tolerance_eps < 0) show_usage(argv[0]);
      }
      else show_usage(argv[0]);
    }
  }

  cout << "Precision is set to " << prec << " bits." << endl
       << "Output eps    = " << output_eps << endl
       << "Tolerance eps = " << tolerance_eps << endl;

  mpf_set_default_prec(prec);
  SDPEnvironment::set_output_eps(output_eps);
  SDPEnvironment::set_tolerance_eps(tolerance_eps);
}

/*
  Shows usage information.
*/
void show_usage(const char *prog) 
{ 
  cerr << "Usage: " << prog << " <command> <args> [-p <prec>] [-oeps <eps>] [-teps <eps]" << endl;
  cerr << endl;
  cerr << "Possible commands: n is always >= 2 and d >= 1 is always odd." << endl << endl;
  cerr << "   opt <n> <d> <outfile>" << endl;
  cerr << "     Writes to outfile, in sparse SDPA format, the SDP approximating the" << endl
       << "     Cohn-Elkies bound for dimension n using polynomials of degree <= 2d." << endl
       << endl;

  cerr << "   feas <n> <d> <eps> <infile> <outfile>" << endl;
  cerr << "     Writes to outfile the feasibility version of the Cohn-Elkies bound" << endl
       << "     for dimension n with polynomials of degree up to 2d, requiring that" << endl
       << "     the objective value should be at most the objective value of the   " << endl
       << "     solution contained in infile (in CSDP format) plus the positive eps."
       << endl << endl;

  cerr << "   project <n> <d> <infile>" << endl;
  cerr << "     Reads a solution from infile (in CSDP format) and projects it onto" << endl
       << "     the subspace generated by the constraints of the Cohn-Elkies SDP."  << endl
       << "     Shows feasibility information of the solution before and after the" << endl
       << "     projection." << endl;

  cerr << endl;

  cerr << "Optional paramters:" << endl << endl;
  cerr << "   -p <prec>   : sets the GMP precision to prec; default is 256." << endl;
  cerr << "   -oeps <eps> : sets output_eps to eps > 0; default is 1e-15." << endl;
  cerr << "   -teps <eps> : sets tolerance_eps to eps > 0; default is 1e-15." << endl
       << "                 When d grows, tolerance_eps should be made smaller." << endl;

  exit(-1); 
}

/*
  main
*/
int main(int argc, char **argv)
{
  if (argc <= 1) show_usage(argv[0]);

  if (strcmp(argv[1], "opt") == 0) {
    if (argc < 5) show_usage(argv[0]);

    int n = atoi(argv[2]);
    int d = atoi(argv[3]);
    const char *outfilename = argv[4];

    if (n < 2) show_usage(argv[0]);
    if (d <= 0 || d % 2 == 0) show_usage(argv[0]);

    ofstream outfile(outfilename);

    if (!outfile) {
      cerr << "Error opening file \"" << outfilename << "\"" << endl;
      exit(-1);
    }
    
    outfile.precision(20);

    get_and_apply_opts(argc - 5, argv + 5);

    /* Executes command */
    Kernel K(n, d);
    
    K.output_opt(outfile);
  }
  else if (strcmp(argv[1], "project") == 0) {
    if (argc < 5) show_usage(argv[0]);

    int n = atoi(argv[2]);
    int d = atoi(argv[3]);
    const char *infilename = argv[4];

    if (n < 2) show_usage(argv[0]);
    if (d <= 0 || d % 2 == 0) show_usage(argv[0]);

    ifstream infile(infilename);

    if (!infile) {
      cerr << "Error opening file \"" << infilename << "\"" << endl;
      exit(-1);
    }

    get_and_apply_opts(argc - 5, argv + 5);

    /* Executes command */
    Kernel K(n, d);

    cout.precision(20);
    
    K.project(infile);
  }
  else if (strcmp(argv[1], "feas") == 0) {
    if (argc < 7) show_usage(argv[0]);

    int n = atoi(argv[2]);
    int d = atoi(argv[3]);
    double eps = atof(argv[4]);
    const char *infilename = argv[5];
    const char *outfilename = argv[6];

    if (n < 2) show_usage(argv[0]);
    if (d <= 0 || d % 2 == 0) show_usage(argv[0]);
    if (eps < 0) show_usage(argv[0]);

    ifstream infile(infilename);

    if (!infile) {
      cerr << "Error opening file \"" << infilename << "\"" << endl;
      exit(-1);
    }

    ofstream outfile(outfilename);

    if (!outfile) {
      cerr << "Error opening file \"" << outfilename << "\"" << endl;
      exit(-1);
    }

    outfile.precision(20);

    get_and_apply_opts(argc - 7, argv + 7);

    Kernel K(n, d);
    
    K.output_feas(infile, eps, outfile);
  }
  else show_usage(argv[0]);

  return 0;
}

/* 
   Returns the polynomial part of the Fourier inverse of a given
   polynomial.
*/
Polynomial Kernel::fourier_inverse(const Polynomial& p) const
{
  Polynomial ret;

  for (auto& c : p) {
    int k = c.first[x] / 2;

    ret += c.second * fourier_polys[k];
  }

  return ret;
}

/*
  Initializes the kernel to deal with a sphere packing problem in
  dimension n using polynomials of degree up to 2d.
*/
Kernel::Kernel(int _n, int d)
  : n(_n), deg(d), x("x"), V(1 + d / 2), F(1 + d / 2),
    x2_F(1 + d / 2), S0(1 + d / 2), S1(1 + d / 2), Q(1 + d / 2)
{
  /* Computes an approximation for pi */
  pi = acos(-1);

  /* Fills in mon_basis */
  for (int k = 0; k <= deg; k++)
    mon_basis.add_monomial(Monomial(x(2 * k)));

  /* Fills in lag_basis; notice the normalization */
  compute_laguerre_polys(deg, x, n / 2.0 - 1, lag_basis, mpf_class(2 * pi), 2);

  for (auto& p : lag_basis) {
    /* Finds the maximum coefficient in absolute value */
    mpf_class max_coeff = 0;
    for (auto& c : p)
      if (NumberOperations::abs(c.second) > max_coeff)
	max_coeff = NumberOperations::abs(c.second);

    p /= max_coeff;
  }

  /* Fills in fourier_polys */
  compute_laguerre_polys(deg, x, n / 2.0 - 1, fourier_polys, pi, 2);

  for (int k = 0; k <= deg; k++)
    for (int i = 1; i <= k; i++)
      fourier_polys[k] *= mpf_class(i / pi);

  /* Fills in the moment matrix V */
  for (int i = 0; i <= deg / 2; i++)
    for (int j = 0; j <= deg / 2; j++)
      V[i][j] = lag_basis[i] * lag_basis[j];

  /* Computes the Fourier inverse matrices */
  for (int i = 0; i <= deg / 2; i++)
    for (int j = 0; j <= deg / 2; j++) {
      F[i][j] = fourier_inverse(V[i][j]);
      x2_F[i][j] = fourier_inverse(x(2) * V[i][j]);
    }
}

/*
  Adds to a constraint container the constraints of the Cohn-Elkies
  bound SDP using a basis of normalized Laguerre polynomials for the
  polynomial constraint.  
*/
void Kernel::make_constraints(SDPConstraintContainer& cc) const
{
  /* Adds the constraint on f^(0) */
  Matrix<mpf_class> A(1 + deg / 2);

  for (int i = 0; i <= deg / 2; i++)
    for (int j = 0; j <= deg / 2; j++)
      A[i][j] = V[i][j].coefficient(Monomial());

  cc.add_constraint(A * S0 == 1);

  /* Adds the polynomial constraint */
  try {
    SparseBasisChanger sc(mon_basis, lag_basis);

    cc.add_constraint(F * S0 + x2_F * S1 + (x(2) - 1) * V * Q == 0, sc);
  }
  catch(InvalidBasis) {
    cerr << endl << "Error: The basis change matrix is singular!" << endl;
    cerr << "Most likely, d is too big; try making tolerance_eps smaller." << endl;

    exit(-1);
  }
}

/*
  Returns a linear combination corresponding to the objective function
  of the problem multiplied by an integer fac.
*/
SDPLinearCombination Kernel::make_objective(int fac) const
{
  Matrix<mpf_class> A(1 + deg / 2), x2_A(1 + deg / 2);

  for (int i = 0; i <= deg / 2; i++)
    for (int j = 0; j <= deg / 2; j++) {
      A[i][j] = mpf_class(fac * F[i][j].coefficient(Monomial()));
      x2_A[i][j] = mpf_class(fac * x2_F[i][j].coefficient(Monomial()));
    }

  return A * S0 + x2_A * S1;
}

/*
  Outputs the basic Cohn-Elkies SDP to a stream.
*/
void Kernel::output_opt(ostream& out) const
{
  SDPProblem problem;
  SDPConstraintContainer cc;

  make_constraints(cc);
  
  problem << cc;
  problem.set_objective(make_objective(-1));

  out << SDPAWriter(problem);
}

/*
  Given an input stream from where to read a solution of the
  Cohn-Elkies SDP and a positive number eps, writes to an output
  stream a feasibility version of the problem setting the objective
  function as a constraint so as to ensure that the objective value is
  epsilon-close to the value of the solution read.
*/
void Kernel::output_feas(istream& in, double eps, ostream& out) const
{
  /* Reads the solution to find out the objective value */
  SDPSolution<mpf_class> sol;

  sol.create({ S0, S1, Q });

  in >> CSDPReader<mpf_class>({ S0, S1, Q }, sol);

  SDPLinearCombination objective = make_objective(1);
  mpf_class obj_value = sol.evaluate_at(objective);

  /* Outputs the feasibiliy problem */
  SDPProblem problem;
  SDPConstraintContainer cc;

  make_constraints(cc);
  cc.add_constraint(objective <= obj_value + eps);
  problem << cc;

  out << SDPAWriter(problem);
}

/*
  Writes to the output stream information about a solution: its
  objective value, the minimum eigenvalue of each matrix, and the
  maximum violation of the constraints in the container.
*/
void Kernel::show_info(const SDPSolution<mpf_class>& sol, const SDPLinearCombination& obj,
		       const SDPConstraintContainer& cc, ostream& out) const
{
  SDPSolutionAnalyzer analyzer;

  /* The minimum eigenvalues */
  vector<double> eigs;

  out << "Minimum eigenvalues..." << endl;
  out << "  of S0 = " << analyzer.double_eigenvalues(sol[S0]->matrix(), eigs) << endl;
  out << "  of S1 = " << analyzer.double_eigenvalues(sol[S1]->matrix(), eigs) << endl;
  out << "  of Q  = " << analyzer.double_eigenvalues(sol[Q]->matrix(), eigs) << endl;
  out << endl;

  /* The feasibility measures and objective value */
  vector<mpf_class> measures;

  out << "Maximum constraint violation = "
      << analyzer.feasibility_measures({ &cc }, sol, measures) << endl;
  out << "Objective value              = "
      << sol.evaluate_at(obj) << endl;
}

/*
  Reads a solution from a stream, computes its feasibility measures
  and shows them, then computes a projection and the feasibility
  measures of the projection.
*/
void Kernel::project(istream& in) const
{
  /* Reads a solution */
  SDPSolution<mpf_class> sol;

  sol.create({ S0, S1, Q });

  in >> CSDPReader<mpf_class>({ S0, S1, Q }, sol);

  /* The objective function, used to find the objective value */
  SDPLinearCombination objective = make_objective(1);

  /* The constraints used for the measures and projection */
  SDPConstraintContainer cc;

  make_constraints(cc);

  /* Computes measures for the solution */
  cout << endl;
  cout << "Info on the original solution:" << endl;
  cout << "------------------------------" << endl << endl;

  show_info(sol, objective, cc, cout);

  /* Projects and computes measures for projection */
  SDPSolution<mpf_class> proj_sol;
  SDPSolutionAnalyzer analyzer;

  analyzer.project_solution({ &cc }, sol, proj_sol);

  cout << endl;
  cout << "Info on the projected solution:" << endl;
  cout << "-------------------------------" << endl << endl;

  show_info(proj_sol, objective, cc, cout);

  cout << endl;
}
