/*!
  \file analyzer.h

  \brief Tools to analyze solutions of SDPs.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __ANALYZER_H__
#define __ANALYZER_H__

#include <list>
#include <vector>
#include <utility>
#include <initializer_list>
#include <gmpxx.h>
#include <sdpsl/sdpenv.h>
#include <sdpsl/sdperror.h>
#include <sdpsl/number.h>
#include <sdpsl/matrix.h>
#include <sdpsl/sdpcon.h>
#include <sdpsl/sdpsolution.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//! Error in SDPSolutionAnalyzer.
class AnalyzerError : public SDPError {
  
public:
  AnalyzerError(const std::string& msg = std::string()) : SDPError(msg) { }
};

/*!
  \brief Tools to analyze solutions of SDPs.

  This class provides tools that can be used to analyze solutions of
  semidefinite programming problems. These tools can be used to:

  - Compute the eigenvalues of symmetric matrices, what can be used to
    check that the matrices are positive semidefinite. Here we employ an
    LAPACK routine to compute the eigenvalues, so computations are
    done using double-precision arithmetic.

  - Compute <em>feasibility measures</em> of a solution, meaning that
    given some constraints, we compute how violated they are. This can
    be done using double-precision or arbitrary-precision arithmetic.

  - Project a given solution on a subspace generated by some
    constraints. This can be used to improve the feasibility measures
    of a solution. Computations can be carried out in double-precision
    arithmetic (using LAPACK), or in arbitrary precision arithmetic,
    using a custom routine.

  \section eigs Computing eigenvalues

  Eigenvalues are computed using double-precision arithmetic, even if
  the matrix uses \c mpf_class. We use an LAPACK routine to compute
  eigenvalues, so numerical issues may arise.

  Two different kinds of functions to compute eigenvalues are
  provided. Their prototypes are:

  - <tt>double double_eigenvalues(const Matrix<T>& A,
    std::vector<double>& eigenvalues) const</tt>

  - <tt>double double_eigenvalues(const Matrix<T>& A, const
    MatrixMask& mask, std::vector<double>& eigenvalues) const</tt>

  \param A is the input matrix, with elements of type \c T (either \c
  double or \c mpf_class).

  \param mask is an instance of \c MatrixMask which tells the
  positions where the matrix is to be considered nonzero.

  \param eigenvalues is the vector that is filled with the eigenvalues
  computed; it is resized if necessary.

  \returns The smallest eigenvalue of the matrix.

  \throws AnalyzerError if the computation fails.

  \throws ValueError if the matrix supplied is not square, of if the
  matrix and the mask have different sizes.

  \note %Matrix \a A is assumed to be symmetric; in any case only the
  lower-diagonal part is used.

  \section feas_measures Feasibility measures

  The measure of violation of a constraint depends on its type:

  - For \c == constraints, it equals the absolute value of the 
    difference of the right-hand side and the left-hand side.

  - For \c <= constraints, it equals the maximum of zero and the
    left-hand side minus the right-hand side.

  - For \c >= constraints, it equals the maximum of zero and the
    right-hand side minus the left-hand side.

  This means that the measure of violation is always a nonnegative
  number, and the more positive it is the more violated the
  constraint is.

  Notice that we have described the measure of violation of a linear
  constraint. %Polynomial constraints are transformed into linear
  equality constraints (see \ref polycon and \ref polybasics), so
  measures of violation can also be applied to them.

  There are two functions that can be used to compute feasibility
  measures, depending on the type of the solution provided. Their
  prototypes are:

  - <tt>double feasibility_measures(const std::initializer_list<const
    SDPConstraintContainer *>& containers, const SDPSolution<T>&
    solution, std::vector<T>& measures) const</tt>

  \param containers is a list of pointers to constraint containers,
  containing the constraints that are going to be tested.

  \param solution is the solution to be tested.

  \param measures is a vector that will contain the measure of
  violation of each constraint (in the order the constraints are
  provided); the vector will be resized if necessary.

  \returns The largest violation.

  When \c T is \c double, computations are done with double-precision
  arithmetic. When \c T is \c mpf_class, computations use
  arbitrary-precision arithmetic via GMP.

  \section solution_proj Projections

  Solvers use numerical methods and return solutions that are not
  truly feasible. One way to improve the feasibility measures of a
  solution for some given set of constraints is to <em>project</em>
  the solution onto the affine subspace generated by the
  constraints. This means that we find the closest point to the
  solution lying on the affine subspace.

  \c SDPSolutionAnalyzer allows for the projection of solutions onto
  the affine subspace generated by linear equality constraints. These
  are specified by a constraint container, and might come from a
  polynomial constraint for instance.

  Suppose we have a semidefinite programming problem with variables
  \f$X_1, \ldots, X_r\f$ and also some linear constraints

  \f[
  \sum_{j=1}^r \langle A_{ij}, X_j\rangle = b_i\quad\text{for $i = 1,
  \ldots, m$}.
  \f]

  Recall that the variable matrices are symmetric. Say \f$X_j\f$ is an
  \f$n_j \times n_j\f$ matrix, and let

  \f[
  N = \sum_{j=1}^r \frac{n_j (n_j + 1)}{2}.
  \f]

  Notice \f$N\f$ is the sum of the number of diagonal and
  upper-diagonal entries in the matrices \f$X_j\f$. So \f$(X_1,
  \ldots, X_r)\f$ can be naturally identified with a vector \f$x \in
  \mathbb{R}^N\f$. This means that the system above can be represented
  as \f$F x = b\f$, where \f$F \in \mathbb{R}^{m \times N}\f$ and \f$b
  \in \mathbb{R}^m\f$.

  We assume that the system \f$F x = b\f$ is under-determined, that
  is, that \f$m \leq N\f$. We use two different approaches to compute
  the projection, depending on whether double-precision or
  arbitrary-precision computations are desired; we will go into
  details below.

  There are four different functions provided to compute projections;
  their prototypes are of the form:

  - <tt>void project_solution(const std::initializer_list<const
    SDPConstraintContainer *>& containers, const SDPSolution<T>&
    solution, SDPSolution<U>& result) const</tt>

  Here, both \c T and \c U can be either \c double or \c
  mpf_class. 

  \param containers is a list of pointers to constraint containers
  holding the constraints that will define the system giving the
  affine subspace onto which the projection will be made. Only the
  equality constraints contained in the container will be considered
  for the projection.

  \param solution contains the solution to be projected.

  \param result will contain the result of the projection. This instance
  of \c SDPSolution will be adapted to contain the projected
  solution. Namely, variables in the projection that do not occur in
  \a result will be created (without masks), and variables in \a
  result will be reset. If a mask is specified for some matrix in \a
  result, then it will be set by \c project_solution. For this,
  environment parameter \c output_eps is used: only numbers whose
  absolute value is at least \c output_eps are considered
  nonzero. Notice that it is usually not a good idea to specify masks
  for the projection, because it is not <em>a priori</em> clear how
  sparse it is.

  \throws ValueError if the system provided is over-determined (see
  above) or empty.

  \throws AnalyzerError if a call do LAPACK fails, or if the system
  given by the constraints is infeasible.

  Finally, notice also that the projection satisfies all constraints
  with equality by definition. Since however we use floating-point
  computations, even after the projection the constraints are not
  exactly satisfied, but their violations will be very small --- one
  would hope orders of magnitude smaller than the violations of the
  original solution.

  \note
  The parameter that determines which method is used for the
  projection is \c U. When \c U is \c double, then LAPACK is used for
  the projection. When \c U is \c mpf_class, a custom routine is used,
  as we explain now.

  \subsection lapack_proj Double-precision projections using LAPACK

  To project a given point \f$z\f$ onto the affine subspace given by
  \f$F x = b\f$, we first find a minimum-norm \f$x^*\f$ such that \f$F
  x^* = b - Fz\f$, and then the projection is \f$\overline{x} = z + x^*\f$.

  LAPACK is used to find this minimum-norm solution. If the solution
  to be projected is given using \c mpf_class, then the system is
  computed using arbitrary-precision arithmetic, but at the end it is
  converted to a double-precision system for LAPACK to be called. It
  is possible that there is a gain in precision during the computation
  of the system, however.

  \subsection gmp_proj Arbitrary-precision projections with GMP

  If \a result is of type <tt>SDPSolution<mpf_class></tt> then the
  projection is made using arbitrary-precision arithmetic. To project
  a point \f$z\f$ onto the afine subspace given by \f$Fx = b\f$ we:

  <ol>
  <li>Find a point \f$x_0\f$ such that \f$Fx_0 = b\f$;</li>

  <li>Project \f$z - x_0\f$ onto the linear subspace given by \f$Fx =
  0\f$. This we do by finding the optimal solution \f$v^*\f$ to the
  least-squares problem \f$\min_v \| F^{\sf T} v - (z - x_0) \|^2\f$;
  then the projection is \f$u = (z - x_0) - F^{\sf T} v^*\f$;</li>

  <li>The projection of \f$z\f$ is then \f$u + x_0 = z - F^{\sf T}
  v^*\f$.</li>
  </ol>

  To execute the above steps, we use the QR decomposition of \f$F\f$.
*/
class SDPSolutionAnalyzer {
  SDPEnvironment env;
  
public:
  //! Creates an analyzer given an environment.
  explicit SDPSolutionAnalyzer(const SDPEnvironment& e = SDPEnvironment())
    : env(e) { }

  //! Computes eigenvalues of a symmetric matrix of doubles; see \ref eigs.
  double double_eigenvalues(const Matrix<double>& A, std::vector<double>& eigenvalues) const;

  //! Computes eigenvalues of a symmetric matrix of doubles with a mask; see \ref eigs.
  double double_eigenvalues(const Matrix<double>& A, const MatrixMask& mask, 
                            std::vector<double>& eigenvalues) const;

  //! Computes eigenvalues of a symmetric matrix of \c mpf_class; see \ref eigs.
  double double_eigenvalues(const Matrix<mpf_class>& A, std::vector<double>& eigenvalues) const;
  
  //! Computes eigenvalues of a symmetric matrix of \c mpf_class with a mask; see \ref eigs.
  double double_eigenvalues(const Matrix<mpf_class>& A, const MatrixMask& mask, 
                            std::vector<double>& eigenvalues) const;

  //! Computes the feasibility measures of a \c double solution; see \ref feas_measures.
  double feasibility_measures(const std::initializer_list<const SDPConstraintContainer *>& containers,
			      const SDPSolution<double>& solution,
			      std::vector<double>& measures) const;

  //! Computes the feasibility measures of a \c mpf_class solution; see \ref feas_measures.
  mpf_class feasibility_measures(const std::list<const SDPConstraintContainer *>& containers,
				 const SDPSolution<mpf_class>& solution,
				 std::vector<mpf_class>& measures) const;

  //! Computes the feasibility measures of a \c mpf_class solution; see \ref feas_measures.
  mpf_class feasibility_measures(const std::initializer_list<const SDPConstraintContainer *>& containers,
				 const SDPSolution<mpf_class>& solution,
				 std::vector<mpf_class>& measures) const;

  /*!
    \brief Projects a double solution onto an affine subspace generated
    by equality constraints using double-precision arithmetic; see
    \ref solution_proj.
  */
  void project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
			const SDPSolution<double>& solution, SDPSolution<double>& result) const;

  /*!
    \brief Projects a GMP solution onto an affine subspace generated by equality
    constraints using double-precision arithmetic; see \ref solution_proj.
  */
  void project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
			const SDPSolution<mpf_class>& solution, SDPSolution<double>& result) const;

  /*!
    \brief Projects a double solution onto an affine subspace generated
    by equality constraints using arbitrary-precision arithmetic; see
    \ref solution_proj.
  */
  void project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
			const SDPSolution<double>& solution, SDPSolution<mpf_class>& result) const;

  /*!
    \brief Projects a GMP solution onto an affine subspace generated by equality
    constraints using arbitrary-precision arithmetic; see \ref solution_proj.
  */
  void project_solution(const std::list<const SDPConstraintContainer *>& containers,
			const SDPSolution<mpf_class>& solution, SDPSolution<mpf_class>& result) const;

  /*!
    \brief Projects a GMP solution onto an affine subspace generated by equality
    constraints using arbitrary-precision arithmetic; see \ref solution_proj.
  */
  void project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
			const SDPSolution<mpf_class>& solution, SDPSolution<mpf_class>& result) const;
};

} // namespace

#endif

// Local variables:
// mode: c++
// End:
