/*
  conv.h

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __CONV_H__
#define __CONV_H__

#include <gmpxx.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*
  !!! CAVEAT LECTOR !!!

  Some functions should work on double or GMP solutions and produce
  either double or GMP results. To avoid duplicating code, we use
  templates. This requires some trickery because of how conversions
  from GMP types to double work.

  Suppose we have an instruction:

  lhs = rhs

  where lhs is a variable of type double or mpf_class, and rhs is an
  expression involving doubles and instances of mpf_class. The point
  here is that rhs has to be converted to the type of lhs for the
  assignment. There are some cases to consider:

  1. lhs = *, rhs = double: Then the right-hand side has to be
  converted to a double (i.e., no conversion needed), or to mpf_class,
  and also no special conversion is needed. So the right-hand side can
  remain unaltered;

  2. lhs = mpf_class, rhs = *: No matter what the right-hand side is
  (a double, mpf_class, or some GMP operation class), then it can be
  assigned to an mpf_class. So no conversion is needed;

  3. lhs = double, rhs = GMP class: This happens when the right-hand
  side is mpf_class, or some GMP operation class. Then rhs must be
  converted to a double. 

  The following templates implement the 3 cases above at compile time,
  removing the need of duplicating code.
*/
template <class T>
inline double convert_to_double(const T& num) { return mpf_class(num).get_d(); }

template <>
inline double convert_to_double(const double& num) { return num; }

template <class T>
class gmp_conv { };   // A tricky class

template <>
class gmp_conv<mpf_class> {
public:
  template <class U>
  static const U& convert(const U& val) { return val; }
};

template <>
class gmp_conv<double> {
public:
  template <class U>
  static double convert(const U& val) { return convert_to_double(val); }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
