/*!
  \file linsys.h

  \brief Solver for linear systems using LU decomposition.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __LINSYS_H__
#define __LINSYS_H__

#include <vector>
#include <sdpsl/matrix.h>
#include <sdpsl/sdperror.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Solver for linear systems using LU decomposition.

  This class provides a solver for linear systems. It is a template
  where type \c T gives the type of the numbers in the system (this
  can be, for instance, \c float, \c double, or \c mpf_class), and
  type \c Op provides a class that implements a static method with
  prototype

  <tt>T Op::abs(const T& val)</tt>

  that returns the absolute value of a number of type \c T.

  To solve the system, we use LU decomposition. When the solver is
  created, the decomposition is computed, and used afterwards to solve
  the system for any given right-hand side.
*/
template <class T, class Op>
class LinearSystemSolver {
  int n;
  T sfactor;  // Scaling factor
  Matrix<T> LU;
  std::vector<int> perm;

public:
  //! Exception thrown when matrix is singular.
  class SingularMatrix { };

  /*!
    \brief Creates a solver given a matrix.

    Initializes the solver by computing the LU decomposition of matrix
    \a A, which is the system matrix. Parameter \a eps is used to
    decide whether the matrix is singular: if during a pivot of the LU
    decomposition algorithm all entries on the column have absolute
    value <= \a eps, then the matrix is said to be singular.

    \note Yes, we are using a row-oriented algorithm!

    \throws SingularMatrix if the matrix is singular.
    \throws ValueError if the matrix given is not square.
  */
  explicit LinearSystemSolver(const Matrix<T>& A, double eps = 1e-10)
    : n(A.num_rows()), LU(A), perm(A.num_rows())
  {
    if (!A.is_square())
      throw ValueError(ERROR_MSG("Matrix is not square"));

    /* Scales the matrix */
    sfactor = 0;
    T *vec = LU.vector_data();

    for (int i = 0; i < LU.vector_size(); i++)
      if (NumberOperations::abs(vec[i]) > sfactor)
	sfactor = NumberOperations::abs(vec[i]);

    if (sfactor <= eps) throw SingularMatrix();

    for (int i = 0; i < LU.vector_size(); i++) vec[i] /= sfactor;

    /* Initilizes the permutation vector */
    for (int i = 0; i < n; i++)
      perm[i] = i;

    /* Does the LU decomposition in place */
    for (int k = 0; k < n; k++) {
      /* Finds pivot element and performs the pivot */
      int pivot = k;
      for (int i = k + 1; i < n; i++)
	if (Op::abs(LU[perm[i]][k]) > Op::abs(LU[perm[pivot]][k]))
	  pivot = i;

      if (Op::abs(LU[perm[pivot]][k]) <= eps) 
	throw SingularMatrix();
    
      { int tmp = perm[k];
	perm[k] = perm[pivot];
	perm[pivot] = tmp;
      }
    
      /* Makes zeros below position (k, k) */
      for (int i = k + 1; i < n; i++) {
	LU[perm[i]][k] /= LU[perm[k]][k];

	const T l = LU[perm[i]][k];
	for (int j = k + 1; j < n; j++)
	  LU[perm[i]][j] -= l * LU[perm[k]][j];
      }
    }
  }

  /*!
    \brief Solves a system for a given right-hand side.

    Given a right-hand side vector \a b, solves the system Ax = b, and
    puts the solution in the vector \a x. Vector \a x must be of the
    appropriate size.

    \throws ValueError if the sizes of b or x are wrong.
  */
  void solve_for(const std::vector<T>& b, std::vector<T>& x) const
  {
    if (b.size() != LU.num_rows() 
	|| x.size() != LU.num_rows())
      throw ValueError(ERROR_MSG("Wrong dimension for b or x"));
  
    /* Solves L y = P b */
    for (int i = 0; i < n; i++) {
      x[i] = b[perm[i]] / sfactor;

      for (int j = 0; j < i; j++)
	x[i] -= LU[perm[i]][j] * x[j];
    }

    /* Solves U x = y */
    for (int i = n - 1; i >= 0; i--) {
      for (int j = n - 1; j > i; j--)
	x[i] -= LU[perm[i]][j] * x[j];

      x[i] /= LU[perm[i]][i];
    }
  }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
