/*!
  \file monomial.h

  \brief Class for representing a monomial and functions dealing with
  monomials.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __MONOMIAL_H__
#define __MONOMIAL_H__

#include <vector>
#include <iostream>
#include <sdpsl/polyvar.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//
// Class declaration
//

//! Represents a monomial
/*!
  A monomial is a product of powers of the variables declared. A
  monomial behaves like a vector of integers associating with each
  variable its exponent.

  \remark In a monomial variables can have negative exponents, since
  we deal with Laurent polynomials.
*/
class Monomial {
  std::vector<int> exps;  // Exponents

public:
  //! Creates the constant monomial, where all variables have exponento zero.
  Monomial() { }

  //! Creates a monomial where variable x has exponent 1.
  explicit Monomial(const PolynomialVariable& x) 
    : exps(x.idx() + 1, 0) { exps[x.idx()] = 1; } 

  //! Creates a monomial given a variable power.
  explicit Monomial(const PolynomialVariable::Power& pw)
    : exps(pw.idx() + 1, 0) { exps[pw.idx()] = pw.power(); }

  /* Get and set */

  //! Returns true if the monomial is the constant monomial (i.e., all variables have exponent zero).
  bool is_constant() const;

  /* Operators */

  //! In-place multiplication by another monomial.
  /*!
    Multiplies the monomial by another monomial, adding the exponents of
    the variables. If the vector containing the exponents of the other
    monomial is bigger, then the size of the vector of this monomial is
    increased.

    \param other is the monomial by which this monomial is multiplied.

    \returns A reference to this monomial.
  */
  Monomial& operator*=(const Monomial& other);

  //! Raises a monomial to the power k.
  /*!
    Raises the monomial to power \a k, multiplying the exponent of every
    variable by \a k, and returns the resulting monomial.

    \param k is the power to which the monomial is raised.

    \returns The resulting monomial.
  */
  Monomial operator()(int k);

  friend Monomial operator*(const Monomial& a, const Monomial& b);

  //! Returns true if the monomial is lexicographically less than the other monomial.
  /*!
    The variables in a monomial are ordered according to their
    indices. This makes it possible to order monomials lexicographically
    according to the exponents of the corresponding variables.
  
    \param other is the other monomial with which this is compared.
  
    \returns True if this < other, false otherwise.
  */
  bool operator<(const Monomial& other) const;

  //! Returns the exponent of a variable given its index.
  int operator[](int i) const { return i >= exps.size() ? 0 : exps[i]; }

  //! Returns the exponent of a given variable.
  int operator[](const PolynomialVariable& x) const { return x.idx() >= exps.size() ? 0 : exps[x.idx()]; }

  //! Outputs monomial to the given stream.
  /*!
    Outputs the monomial to the stream and returns the stream. The
    monomial is output as a string like x(k) * y(l), where x and y are
    variables and k and l are their exponents. If k is 1, then "x" is
    output instead of "x(1)", for instance. Only variables with nonzero
    exponent our output. Variables are ordered according to their
    indices.
  */
  friend std::ostream& operator<<(std::ostream& out, const Monomial& mon);
};

//
// Prototypes
//
extern std::ostream& operator<<(std::ostream& out, const Monomial& mon);

//! Multiplies two monomials and returns the resulting monomial.
extern Monomial operator*(const Monomial& a, const Monomial& b);

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
