/*!
  \file number.h

  \brief A generic number type, and operations performed on it.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __NUMBER_H__
#define __NUMBER_H__

#include <string>
#include <iostream>
#include <gmpxx.h>
#include <math.h>

#ifdef __MPF_BUILD__

namespace mpf_sdp {

//! A number in \sdpsl.
typedef mpf_class Number;

#else

namespace sdp {

//! A number in \sdpsl.
typedef double Number;

#endif

//
// Functions
//

/*!
  \brief Converts a number of type \c U to a number of type \c T.
*/
template <class T, class U>
inline T convert_number(const U& num) { return (T) num; }

/*!
  \brief Specialization of \c convert to <tt>U = mpf_class</tt> and
  <tt>T = double</tt>.
*/
template <>
inline double convert_number(const mpf_class& num) { return num.get_d(); }

/*!
  \brief A converter class for \c double to \c mpf_class.

  This can be used, for instance, with \c Matrix and \c SDPSolution.
*/
class double_to_mpf_class {
public:
  mpf_class operator()(double x) const { return mpf_class(x); }
};

/*!  
  \brief Used to output the coefficient of a polynomial, when it is
  not the first being output.
*/
class InsideCoefficient {
  const Number& num;

public:
  //! Initializes the inside coefficient to a number \a n.
  explicit InsideCoefficient(const Number& n) : num(n) { }

  //! Outputs the coefficient to the stream given.
  friend std::ostream& operator<<(std::ostream& out, const InsideCoefficient& ic);
};

/*!
  \brief Implements common operations on type \c Number.
*/
class NumberOperations {
public:
  /*! 
    \brief Returns the absolute value of a \c double.
  */
  static double abs(double num) { return num >= 0 ? num : -num; }

  /*!
    \brief Returns the absolute value of an \c mpf_class.
  */
  static mpf_class abs(const mpf_class& num)
  {
    mpf_class ret;

    mpf_abs(ret.get_mpf_t(), num.get_mpf_t());

    return ret;
  }

  /*!
    \brief Returns the square root of a \c double.
  */
  static double sqrt(double num) { return ::sqrt(num); }

  /*!
    \brief Returns the square root of an \c mpf_class.
  */
  static mpf_class sqrt(const mpf_class& num)
  {
    mpf_class ret;
    
    mpf_sqrt(ret.get_mpf_t(), num.get_mpf_t());

    return ret;
  }

  /*!
    \brief Converts string to \c double and puts it in \a out.

    \throws ValueError if the conversion fails.
  */
  static void from_string(const std::string& str, double& out);

  /*!
    \brief Converts string to \c mpf_class and puts it in \a out.

    \throws ValueError if the conversion fails.
  */  
  static void from_string(const std::string& str, mpf_class& out);
};

//
// Prototypes
//
extern std::ostream& operator<<(std::ostream& out, const InsideCoefficient& ic);

}  // namespace

#endif

// Local variables:
// mode: c++
// End:

