/*!
  \file outmatrix.h

  \brief A matrix and mask pair, used for output.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __OUTMATRIX_H__
#define __OUTMATRIX_H__

#include <memory>
#include <sdpsl/matrix.h>
#include <sdpsl/matrixmask.h>
#include <sdpsl/sdpvar.h>


#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A matrix and mask pair, used for output.

  Some functions of \sdpsl require a list of matrices with optional
  associated masks, and it must be possible to access these
  matrices. This is the case, for instance, in functions that load
  solutions or deal with them.

  \c SDPOutputMatrix is a pair containing a matrix and an optional
  associated mask of the same size. For both the matrix and the mask,
  a <em>shared reference</em> is stored (see e.g. tbe documentation of
  class \c Matrix or \c MatrixMask; see also \c shared_ref and \c
  const_shared_ref).
*/
template <class T>
class SDPOutputMatrix {
  std::unique_ptr<shared_ref<Matrix<T> > > _matrix_ptr;
  std::unique_ptr<shared_ref<MatrixMask> > _mask_ptr;

public:
  /*!
    \brief Copy constructor.

    We must be careful to make new shared references!
  */
  SDPOutputMatrix(const SDPOutputMatrix& other)
  {
    _matrix_ptr.reset(new shared_ref<Matrix<T> >(*(other._matrix_ptr->ptr())));
    
    if (other._mask_ptr)
      _mask_ptr.reset(new shared_ref<MatrixMask>(*(other._mask_ptr->ptr())));
  }

  /*!
    \brief Creates a pair given a matrix and a mask.

    \throws ValueError if the dimensions of the matrix and mask
    differ, or if the matrix is not square.
  */
  SDPOutputMatrix(Matrix<T>& A, MatrixMask& mask)
  {
    if (!A.is_square())
      throw ValueError(ERROR_MSG("Matrix is not square"));

    if (A.num_rows() != mask.num_rows()
	|| A.num_cols() != mask.num_cols())
      throw ValueError(ERROR_MSG("Matrix and mask have different dimensions"));

    _matrix_ptr.reset(new shared_ref<Matrix<T> >(A));
    _mask_ptr.reset(new shared_ref<MatrixMask>(mask));
  }
  
  /*!
    \brief Creates a pair given a matrix, possibly creating a mask.
    
    If \a create_mask is \c true, then a mask of the same size of the
    matrix is created.
    
    \throws ValueError if the matrix is not square.
  */
  explicit SDPOutputMatrix(Matrix<T>& A, bool create_mask = false)
  {
    if (!A.is_square())
      throw ValueError(ERROR_MSG("Matrix is not square"));

    _matrix_ptr.reset(new shared_ref<Matrix<T> >(A));
    
    if (create_mask) {
      MatrixMask mask(A.num_rows());

      _mask_ptr.reset(new shared_ref<MatrixMask>(mask));
    }
  }
  
  /*!
    \brief Given an instance of \c SDPVariable, creates a pair
    containing a matrix of the corresponding size, together with an
    optional mask.

    If \a create_mask is \c true, then a mask of the same size as the
    matrix will be created.
  */
  explicit SDPOutputMatrix(const SDPVariable& X, bool create_mask = false)
  {
    Matrix<T> A(X.size());
    _matrix_ptr.reset(new shared_ref<Matrix<T> >(A));
    
    if (create_mask) {
      MatrixMask mask(A.num_rows());

      _mask_ptr.reset(new shared_ref<MatrixMask>(mask));
    }
  }
  
  //! Returns a reference to the matrix contained.
  Matrix<T>& matrix() { return *(_matrix_ptr->ptr()); }

  //! Returns a constant reference to the matrix contained.
  const Matrix<T>& matrix() const { return *(_matrix_ptr->ptr()); }

  //! Returns a pointer to the mask, which might be \c nullptr.
  MatrixMask *mask() 
  { 
    if (_mask_ptr) return _mask_ptr->ptr(); 
    else return nullptr;
  }

  //! Returns a constant pointer to the mask, which might be \c nullptr.
  const MatrixMask *mask() const
  {
    if (_mask_ptr) return _mask_ptr->ptr();
    else return nullptr;
  }

  //! Assignment operator.
  SDPOutputMatrix& operator=(const SDPOutputMatrix& other)
  {
    _matrix_ptr.reset(new shared_ref<Matrix<T> >(*(other._matrix_ptr->ptr())));

    if (other._mask_ptr)
      _mask_ptr.reset(new shared_ref<MatrixMask>(*(other._mask_ptr->ptr())));
    else
      _mask_ptr.reset();

    return *this;
  }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
