/*!
  \file polyvar.h

  \brief Classes implementing variables of polynomials.  

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __POLYVAR_H__
#define __POLYVAR_H__

#include <vector>
#include <string>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//! A variable in a polynomial.
/*!
  Variables have to be declared so that polynomials can be
  created. Each variable has a name, which is a string that is used
  to represent the variable when polynomials are output, for
  instance. Variables are uniquely identified by their indices, which
  start at 0, and not by their names.

  Once a variable is created, it can be used in any polynomial,
  whether the polynomial was created before the variable or
  not. Variables cannot be erased once created.
*/
class PolynomialVariable {
  //! Holds the name of each variable declared.
  static std::vector<std::string> variable_names;

  //! The index of this variable.
  int _idx;

public:
  //! Creates a new variable given its name.
  explicit PolynomialVariable(const std::string& name)
    : _idx(new_variable(name).idx()) { }
  
  //! Creates a variable given its index.
  explicit PolynomialVariable(int i);

  /* Get and set */
  int idx() const { return _idx; }  //!< Returns the index of the variable. 
  const std::string& name() const { return variable_names[_idx]; } //!< Returns the name of the variable.

  //! Represents a power of a variable.
  class Power {
    friend class PolynomialVariable;

    int _idx;
    int k;

    Power(int i, int kk) 
      : _idx(i), k(kk) { }

  public:
    int idx() const { return _idx; } //!< Returns the index of the variable.
    int power() const { return k; }  //!< Returns the exponent of the power.
  };

  //! Raises a variable to an integer power.
  /*! Returns an instance of class Power, representing an integer
      power (whether negative or positive) of the variable. Instances
      of class Power are then used to construct instances of classes
      Monomial and Polynomial.

      \param k the power to which the variable is raised.
      \return Instance of class Power (see above).
  */
  Power operator()(int k) const { return Power(_idx, k); }

  /* Static functions  */
  static int num_variables() { return variable_names.size(); }  //!< Returns number of variables declared.

  //! Declares a new variable, given its name.
  /*!
    Creates a new variable, given a name to it, assigning to it an
    unused index. This effectively increases the number of variables
    that can be used in declaring polynomials by 1.

    \param name the name of the variable. No check is made to see
    whether this name is unique. The name however is only used in
    representations of polynomials, it does not identify the variable
    itself.

    \return An instance of PolynomialVariable corresponding to the newly
    created variable.
  */
  static PolynomialVariable new_variable(const std::string& name);

  //! Returns name of i-th variable
  /*! \param i index of the variable whose name should be returned. It
      is not checked whether this variable really exists.
      \return Name of the variable.
  */
  static const std::string& var_name(int i) { return variable_names[i]; }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
