/*!
  \file rawconstraint.h

  \brief Classes for internal storage of SDP constraints and utility
  functions for the creation of such data.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __RAWCONSTRAINT_H__
#define __RAWCONSTRAINT_H__

#include <list>
#include <cassert>
#include <sdpsl/number.h>
#include <sdpsl/sdplin.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//
// Types
//

/*!
  \brief The relation of a constraint: ==, >=, or <=.
*/
enum SDPRelation {
  sdp_leq, 
  sdp_eq, 
  sdp_geq
};

/*!
  \brief An entry in a constraint matrix.

  An instance of \c MatrixEntry represents an entry of a constraint
  matrix. It contains the index of the variable to which the entry
  corresponds, the matrix position <tt>(i, j)</tt>, and the value of
  the entry.
*/
struct MatrixEntry {
  int var_idx;  //!< Index of the corresponding variable
  int i;  //!< Matrix row
  int j;  //!< Matrix column
  Number value;  //!< Entry value

  //! Creates a matrix entry with given data.
  MatrixEntry(int _var_idx, int _i, int _j, const Number& _value)
    : var_idx(_var_idx), i(_i), j(_j), value(_value) { }
};

/*!
  \brief A list of matrix entries.

  Implements a list of matrix entries. Iteration can be done using
  \cpp11 notation. Provides functions to add a single entry to the
  list or also the entries of a whole matrix.
*/
class MatrixEntryList {
  std::list<MatrixEntry> entries;

public:
  //! Constant iterator.
  typedef std::list<MatrixEntry>::const_iterator iterator;

  //! Clears the list.
  void clear() { entries.clear(); }

  //! Returns true if the list is empty.
  bool empty() const { return entries.empty(); }

  //! Returns iterator to the beginning.
  iterator begin() const { return entries.begin(); }

  //! Returns past-the-end iterator.
  iterator end() const { return entries.end(); }

  //! Swaps content with the other list.
  void swap(MatrixEntryList& other) { entries.swap(other.entries); }

  /* Utility functions */

  /*!
    \brief Adds a new entry to a list of matrix entries.
    
    Creates a new \c MatrixEntry instance given the variable index,
    position and value, and adds it to the list of entries. If the
    absolute value of the entry is less than or equal than \a
    output_eps, then the entry is not added.
  */
  void add_entry(double output_eps, int var_idx, int i, int j, 
		 const Number& value)
  {
    assert(i <= j);
    
    if (NumberOperations::abs(value) > output_eps)
      entries.emplace_back(var_idx, i, j, value);    
  }
  
  /*!
    \brief Adds all the entries of a given matrix to a list of entries.

    This function adds all entries of a given matrix, associated with a
    certain variable index of a problem, to the list of entries. It uses
    \c add_entry, so only entries whose absolute value is greater
    than \a output_eps are actually added. The matrix can have a mask,
    and it is used to speed-up processing.
  */
  void add_matrix_entries(double output_eps, int var_idx, 
			  const SDPMaskedMatrix& mm);
};

/*!
  \brief Internal, raw format for a constraint of an SDP.

  Contains all the information regarding a constraint of the
  semidefinite programming problem. This means that we have a list of
  instances of \c MatrixEntry containing the entries of the constraint
  matrices, the relation of the constraint (whether it is equality or
  inequality), and the right-hand side.
*/
struct RawConstraint {
  MatrixEntryList entries;  //!< Entries of the constraint matrices
  SDPRelation rel;  //!< Relation
  Number rhs;  //!< The righ-hand side

  //! Creates a new raw constraint given its relation and right-hand side.
  RawConstraint(SDPRelation _rel, const Number& _rhs)
    : rel(_rel), rhs(_rhs) { }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
