/*!
  \file sdpawriter.h
  
  \brief Class for outputting a problem in sparse SDPA format.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPAWRITER_H__
#define __SDPAWRITER_H__

#include <list>
#include <iostream>
#include <sdpsl/sdpproblem.h>
#include <sdpsl/sdpprogress.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Allows a problem to be output in sparse SDPA format.

  An instance of \c SDPAWriter can be used to output a problem in
  sparse SDPA format. Each variable occurring in the problem will
  correspond to a different block in the file. To assign block numbers
  to variables, variables are first ordered according to their
  indices, and then block numbers are assigned according to this
  order.

  \todo At the moment, if a problem contains many 1-by-1 variables,
  then many 1-by-1 blocks will be created. It would be better to group
  these variables in a diagonal block.
*/
class SDPAWriter {
  const SDPProblem& problem;
  mutable SDPProgressIndicator *progress;
  std::list<SDPVariable> ignore_list;

  void write(std::ostream& out) const;

public:
  /*!
    \brief Creates a writer given a problem.

    Only a reference to the problem is kept, so it must be guaranteed
    that the problem lives through the lifetime of the writer.
  */
  explicit SDPAWriter(const SDPProblem& _problem)
    : problem(_problem), progress(nullptr) { }

  /*!
    \brief Creates a writer given a problem and a progress indicator.

    Writing out a problem can take time, so a progress indicator may
    be provided to keep track of progress.

    \sa \c SDPProgressIndicator, \c ProgressBar
  */
  SDPAWriter(const SDPProblem& _problem, SDPProgressIndicator& _progress)
    : problem(_problem), progress(&_progress) { }

  /*! 
    \brief Writes the problem to the output stream provided.

    \throws ValueError if the problem is empty.
  */    
  friend std::ostream& operator<<(std::ostream& out, const SDPAWriter& writer)
  {
    writer.write(out);
    return out;
  }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
