/*!
  \file sdperror.h

  \brief Exceptions for \sdpsl, except for the polynomial classes.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPERROR_H__
#define __SDPERROR_H__

#include <string>
#include <sstream>
#include <exception>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

#ifdef __GNUC__
/* When using the GNU compiler, we use __FILE__ and __LINE__ to print nicer messages */

inline std::string make_error_msg(const std::string& msg, const std::string& filename, int lineno)
{
  std::stringstream ss;

  ss << filename << ", line " << lineno << ": " << msg;
  
  return ss.str();
}

#define ERROR_MSG(msg) make_error_msg(msg, __FILE__, __LINE__)
#define ERROR_INFO make_error_msg(std::string(), __FILE__, __LINE__)

#else

#define ERROR_MSG(msg) msg
#define ERROR_INFO std::string()

#endif  // #ifdef __GNUC__

//! Base class for exceptions of SDLSL
class SDPError : public std::exception {
  std::string what_text;

public:
  explicit SDPError(const std::string& txt = std::string()) : what_text(txt) { }

  virtual const char *what() const throw() { return what_text.c_str(); }
};

//! Invalid parameter passed to function.
class ValueError : public SDPError {
  
public:
  explicit ValueError(const std::string& txt = std::string()) : SDPError(txt) { }
};

//! Constraint is detected to be infeasible.
class InfeasibleConstraint : public SDPError {

public:
  explicit InfeasibleConstraint(const std::string& txt = std::string()) : SDPError(txt) { }
};

//! Basis of polynomials specified is not a basis or is insufficient.
class InvalidBasis : public SDPError { 
public:
  enum error_code {
    representation_error,
    repeated_monomial,
    incompatible_bases,
    singular_basis_change_matrix
  };

private:
  error_code _code;

public:
  explicit InvalidBasis(error_code code, const std::string& txt = std::string())
    : SDPError(txt), _code(code) { }

  error_code code() const { return _code; }
};

//! Feature not implemented.
class NotImplemented : public SDPError { 

public:
  explicit NotImplemented(const std::string& txt = std::string()) : SDPError(txt) { }
};

/*! 
  \brief Exception thrown when there is a read error.

  \bug This class has a deliberate memory leak, because I cannot
  figure out a super easy way of avoiding it, and because I honestly
  think it doesn't matter. The leak is in function \c what.
*/
class ReadError : public SDPError {
  int _lineno;
  
public:
  //! Constructs a \c ReadError given a message and the line number of the error.
  ReadError(const std::string& msg, int lineno = 0) : SDPError(msg), _lineno(lineno) { }

  /*!
    \brief Shows what message.

    \bug The returned string is allocated but never freed. This is a
    memory leak, but it should be ok because method \c what is not
    going to be called repeatedly, is it? Moreover, the program
    usually dies right afterwards!
  */
  virtual const char *what() const throw() 
  { 
    std::stringstream ss;

    ss << SDPError::what();

    if (_lineno)
      ss << ", at input line number " << _lineno;

    // Here we are deliberately causing a memory leak, oh my!
    std::string *foo = new std::string(ss.str());

    return foo->c_str();
  }  
  
  //! Returns the line number of the error.
  int lineno() const { return _lineno; }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
