/*!
  \file sdplincons.h

  \brief Classes for implementing linear constraints.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPLINCONS_H__
#define __SDPLINCONS_H__

#include <sdpsl/rawconstraint.h>
#include <sdpsl/sdplin.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Implements a linear constraint in an SDP.

  An instance of this class represents a linear constraint in a
  semidefinite programming problem (see \ref tut_sdp for a definition
  of the problems we consider); it contains the following:

  - A <b>left-hand side</b>, which is an instance of \c
    SDPLinearCombination;

  - A \b relation, which is either identity, less-than-or-equal, or
    greather-than-or-equal;

  - A <b>right-hand side</b> which is just a number.

  An instance of \c SDPLinearConstraint keeps the above data.

  \section sdplincons_con Constructing linear constraints

  Operators \c ==, \c <=, and \c >= are overloaded for instances of \c
  SDPLinearCombination and \c Number. This means that constraints can
  be constructed using a simple notation, as the following example
  illustrates:

  \include conex.cpp
*/
class SDPLinearConstraint {
  SDPLinearCombination _lhs;
  SDPRelation rel;
  Number _rhs;

public:
  /*!
    \brief  Constructs an instance of \c SDPLinearConstraint given a left-hand
    side, a relation, and a right-hand side.
  */
  SDPLinearConstraint(const SDPLinearCombination& C, SDPRelation R, const Number& num);
  
  /* Get and set */
  const SDPLinearCombination& lhs() const { return _lhs; }  //!< Returns the left-hand side.
  SDPRelation relation() const { return rel; }  //!< Returns the relation.
  const Number& rhs() const { return _rhs; }    //!< Returns the right-hand side.
  void rhs(const Number& num) { _rhs = num; }   //!< Sets the right-hand side.
};

//
// Prototypes
//

/*! \related SDPLinearConstraint
  \brief Creates a linear constraint with relation \c <=.
*/
extern SDPLinearConstraint operator<=(const SDPLinearCombination& lhs, const Number& rhs);

/*! \related SDPLinearConstraint
  \brief Creates a linear constraint with relation \c ==.
*/
extern SDPLinearConstraint operator==(const SDPLinearCombination& lhs, const Number& rhs);

/*! \related SDPLinearConstraint
  \brief Creates a linear constraint with relation >=.
*/
extern SDPLinearConstraint operator>=(const SDPLinearCombination& lhs, const Number& rhs);

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
