/*!
  \file sdppolcons.h

  \brief Class implementing polynomial constraints.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPPOLCONS_H__
#define __SDPPOLCONS_H__

#include <sdpsl/sdppol.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Implements a polynomial constraint in an SDP.

  An instance of this class represents a polynomial constraint in a
  semidefinite programming problem (see \ref tut_sdp
  for a definition of the problems we consider); it contains the following:

  - A <b>left-hand side</b> which is an instance of \c
    SDPPolynomialCombination;

  - A <b>right-hand side</b> which is a polynomial, an instance of \c
    Polynomial.
*/
class SDPPolynomialConstraint {
  SDPPolynomialCombination _lhs;
  Polynomial _rhs;

public:
  //! Constructs a polynomial constraint with given left-hand and right-hand sides.
  SDPPolynomialConstraint(const SDPPolynomialCombination& C, const Polynomial& p);
  
  /* Get and set */
  const SDPPolynomialCombination& lhs() const { return _lhs; }  //!< Returns left-hand side.
  const Polynomial& rhs() const { return _rhs; }  //!< Returns right-hand side.
  void rhs(const Polynomial& p) { _rhs = p; }     //!< Sets the right-hand side.
};

//
// Prototypes
//

/*!
  \related SDPPolynomialConstraint

  \brief Returns a polynomial constraint with given left-hand and right-hand sides.
*/
extern SDPPolynomialConstraint operator==(const SDPPolynomialCombination&, const Polynomial&);

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
