/*!
  \file sdpsolution.h

  \brief A map associating instances of \c SDPOutputMatrix to
  instances of \c SDPVariable.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <initializer_list>
#include <sdpsl/outmatrix.h>
#include <sdpsl/sdpvarmap.h>
#include <sdpsl/conv.h>

#ifndef __SDPSOLUTION_H__
#define __SDPSOLUTION_H__

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A map associating instances of \c SDPOutputMatrix to
  instances of \c SDPVariable.

  This is used to store solutions of SDPs. We use a template, instead
  of using the type \c Number, because of \c SDPSolutionAnalyzer,
  which can transform double solutions to GMP solutions and vice
  versa.
*/
template <class T>
class SDPSolution : public SDPVariableMap<SDPOutputMatrix<T> > {
  
public:
  //! Constructs an empty solution.
  SDPSolution() { }

  //! Copy constructor: copies all matrices and masks.
  SDPSolution(const SDPSolution& other)
  {
    for (auto& c : other) {
      Matrix<T> A = c.second.matrix();

      if (c.second.mask()) {
	MatrixMask mask = *c.second.mask();

	set_value(c.first, SDPOutputMatrix<T>(A, mask));
      }
      else
	set_value(c.first, SDPOutputMatrix<T>(A));
    }
  }
  
  /*!
    \brief Constructs a solution by converting another solution.

    This will convert each matrix of a solution of type \c U to a
    matrix of type \c T. A <em>converter class</em> like the one used
    for converting matrices (see \c Matrix) is needed.

    Each mask present is also copied, as is.
  */
  template <class U, class Converter>
  explicit SDPSolution(const SDPSolution<U>& other, const Converter& converter)
  {
    for (auto& c : other) {
      Matrix<T> A(c.second.matrix(), converter);

      if (c.second.mask()) {
	MatrixMask mask = *c.second.mask();

	set_value(c.first, SDPOutputMatrix<T>(A, mask));
      }
      else
	set_value(c.first, SDPOutputMatrix<T>(A));
    }
  }

  /*!
    \brief Associates a given instance of \c SDPOutputMatrix with 
    a variable.

    \throws ValueError if the matrix and the variable have different
    sizes.
  */
  virtual void set_value(const SDPVariable& X, const SDPOutputMatrix<T>& val)
  {
    if (val.matrix().num_rows() != X.size())
      throw ValueError(ERROR_MSG("Variable and matrix have different sizes"));
    
    SDPVariableMap<SDPOutputMatrix<T> >::set_value(X, val);
  }

  /*! 
    \brief Creates an entry with a new variable, without a mask.

    If the variable is already in the map, then its value is
    overwritten.
  */
  void create(const SDPVariable& X) { set_value(X, SDPOutputMatrix<T>(X)); }

  /*!
    \brief Creates entries with new variables, without a mask.
  */
  void create(const std::initializer_list<SDPVariable>& vars)
  {
    for (auto& X : vars)
      create(X);
  }

  /*!
    \brief Creates an entry with a new variable and a mask.

    If the variable is already in the map, then its value is
    overwritten.
  */
  void create_with_mask(const SDPVariable& X) { set_value(X, SDPOutputMatrix<T>(X, true)); }

  /*!
    \brief Creates entries with new variables and masks.
  */
  void create_with_mask(const std::initializer_list<SDPVariable>& vars)
  {
    for (auto& X : vars)
      create_with_mask(X);
  }

  /*!
    \brief Returns the value of the linear combination where the
    variables are substituted by the matrices in the solution.
  */
  T evaluate_at(const SDPLinearCombination& C) const
  {
    T val = 0;

    for (auto& c : C) {
      const SDPOutputMatrix<T> *ret = (*this)[c.first];

      if (!ret) continue;

      const Matrix<Number>& A = c.second.matrix();
      const MatrixMask *mask_A = c.second.mask();

      const Matrix<T>& X = ret->matrix();
      const MatrixMask *mask_X = ret->mask();

      if (mask_A)
	for (auto& p : *mask_A) {
	  if (!mask_X || mask_X->is_on(p.i, p.j))
	    val += gmp_conv<T>::convert(A[p.i][p.j] * X[p.i][p.j]);
	}
      else {
	for (int i = 0; i < A.num_rows(); i++)
	  for (int j = 0; j < A.num_cols(); j++)
	    if (!mask_X || mask_X->is_on(i, j))
	      val += gmp_conv<T>::convert(A[i][j] * X[i][j]);
      }
    }

    return val;
  }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
