/*!
  \file sdpvarmap.h

  \brief A map associating instances of a given type to instances of
  \c SDPVariable.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SDPVARMAP_H__
#define __SDPVARMAP_H__

#include <map>
#include <sdpsl/outmatrix.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!  
  \brief A map associating instances of type \c T to instances of
  \c SDPVariable.

  The map contains instances of <tt>std::pair<SDPVariable, T></tt>.
  Iteration is over such pairs.
*/
template <class T>
class SDPVariableMap {
  typedef std::map<SDPVariable, T> Map;

  Map map;

public:
  //! Creates an empty map.
  SDPVariableMap() { }

  //! Virtual destructor.
  virtual ~SDPVariableMap() { }

  //! Iterator
  typedef class Map::iterator iterator;

  //! Constant iterator
  typedef class Map::const_iterator const_iterator;

  //! Returns \c true if the variable is contained; \c false otherwise.
  bool contains(const SDPVariable& X) const { return map.find(X) != map.end(); }

  //! Returns iterator to the beginning.
  iterator begin() { return map.begin(); }
  
  //! Returns past-the-end iterator.
  iterator end() { return map.end(); }

  //! Returns const-iterator to the beginning.
  const const_iterator begin() const { return map.begin(); }

  //! Returns const past-the-end iterator.
  const const_iterator end() const { return map.end(); }

  /*!
    \brief Returns a pointer to the value associated with a given variable.

    If there is no value, a \c nullptr is returned.
  */
  virtual T *operator[](const SDPVariable& X)
  {
    auto it = map.find(X);

    if (it == map.end())
      return nullptr;

    return &(it->second);
  }

  /*!
    \brief Returns a const pointer to the value associated with a given variable.

    If there is no value, a \c nullptr is returned.
  */
  virtual const T *operator[](const SDPVariable& X) const
  {
    auto it = map.find(X);

    if (it == map.end())
      return nullptr;

    return &(it->second);
  }
  
  //! Associates a given instance of \c T to a variable.
  virtual void set_value(const SDPVariable& X, const T& val)
  {
    auto ret = map.insert(typename Map::value_type(X, val));

    if (!ret.second)
      ret.first->second = val;
  }
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
