/*!
  \file sparsechanger.h

  \brief Given a basis of polynomials, implements the interface of \c
  SparseBasisCoefficients using the given basis.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SPARSECHANGER_H__
#define __SPARSECHANGER_H__

#include <map>
#include <sdpsl/sdpenv.h>
#include <sdpsl/sdppolcons.h>
#include <sdpsl/sparsecoeff.h>
#include <sdpsl/sparsemon.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!  
  \brief Implements the interface of \c SparseBasisCoefficients
  using a given basis of polynomials.

  To add a polynomial constraint using a basis other than the standard
  monomial basis, one needs to implement the interface of \c
  SparseBasisCoefficients. \c SparseBasisChanger does that
  automatically given a basis of polynomials in which the constraint
  should be represented (cf. \ref polycon).

  \c SparseBasisChanger solves a linear system to invert the basis
  change matrix for the basis provided.  Environment variable \c
  tolerance_eps (cf. \c SDPEnvironment) is used to determine whether
  the basis change matrix is singular or not, that is, whether the
  basis provided is indeed a basis.

  \warning \c output_eps is being used in the constructor during
  matrix inversion, in order to decide which positions of the mask
  should be on, etc.
*/
class SparseBasisChanger : public SparseBasisCoefficients {
  typedef std::map<Monomial, int> BasisPosMap;

  SDPEnvironment env;
  int size;  // Size of the basis
  SparseMonomialBasis _mon_basis;
  Matrix<Number> Ainv;  // Inverse of basis change matrix
  std::vector<VectorMask> row_masks;  // Mask for each row of Ainv
  
public:
  /*!
    \brief Initializes the basis changer given a monomial basis and a
    basis of polynomials spanning the same subspace.

    \throws InvalidBasis if \a mon_basis and \a poly_basis do not span
    the same subspace, or if \a poly_basis is not a basis (i.e., if it
    is linearly dependent).
  */
  SparseBasisChanger(const SparseMonomialBasis& mon_basis,
		     const std::vector<Polynomial>& poly_basis,
		     const SDPEnvironment& e = SDPEnvironment());

  //! Returns the size of the basis; implements method of \c SparseBasisCoefficients.
  int basis_size() const { return size; }
  
  /*!
    \brief Implements method of \c SparseBasisCoefficients.

    Returns \c false if the polynomial \a p cannot be expanded on
    basis \a poly_basis given to the constructor.
  */
  bool basis_coefficients(const Polynomial& p, std::vector<Number>& coeffs,
			  VectorMask& mask) const;
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
