/*!
  \file sparsemon.h

  \brief Implements a basis of monomials.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SPARSEMON_H__
#define __SPARSEMON_H__

#include <map>
#include <vector>
#include <sdpsl/sparsecoeff.h>
#include <sdpsl/sdppolcons.h>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief This class implements a basis of monomials.

  This class is derived from \c SparseBasisCoefficients and can be
  used to implement a basis of monomials, that is, a basis of
  some subspace of the space of polynomials containing only monomials.  
*/
class SparseMonomialBasis : public SparseBasisCoefficients {
  typedef std::map<Monomial, int> BasisPosMap;
  
  BasisPosMap basis_pos;  // The position of each monomial in the basis
  std::vector<Monomial> _monomials;  // Monomials contained in the basis
  
public:
  //! Creates an empty basis.
  SparseMonomialBasis() { }

  /*!
    \brief Constructs a basis of monomials given a vector containing the monomials.
    
    \throws InvalidBasis if there are repeated monomials in \a basis.
  */
  explicit SparseMonomialBasis(const std::vector<Monomial>& basis);

  /*!
    \brief Constructs a basis of monomials given a polynomial constraint.
    
    The basis consists of all monomials occurring in any polynomial that
    appears in the constraint (both on the left and right-hand sides).
    
    \note
    To make the basis, the function has to run through all
    polynomials used in the constraint. In many cases, it might be more
    efficient to construct a monomial basis by giving the vector
    containing all monomials in the basis.
  */
  explicit SparseMonomialBasis(const SDPPolynomialConstraint& C);

  //! Implements method of \c SparseBasisCoefficients.
  int basis_size() const { return basis_pos.size(); }

  /*!
    \brief Implements the method of \c SparseBasisCoefficients.
    
    Given a polynomial \a p, this function puts on the vector \a coeffs
    the nonzero coefficients of the monomials occurring in \a p. The
    mask is turned on on the positions of the basis corresponding to the
    nonzero coefficients.
    
    \returns true if all monomials of \a p occur in the basis; false
    otherwise.
  */
  bool basis_coefficients(const Polynomial& p, std::vector<Number>& coeffs, 
			  VectorMask& mask) const;
  
  /*!
    \brief Adds a monomial to the basis.
    
    \throws InvalidBasis if the monomial is already in the basis.
  */
  void add_monomial(const Monomial& mon);

  //! Returns the vector of monomials of the basis.
  const std::vector<Monomial>& monomials() const { return _monomials; }

  /*!
    \brief Returns the index associated with a given monomial of the
    basis.

    To each monomial in the basis is associated an index starting at
    zero. Given a monomial, this function returns its index, if it is
    in the basis. If it is not in the basis, the function returns -1.
  */
  int mon_idx(const Monomial& mon) const;  
};
  
}  // namespace


#endif

// Local variables:
// mode: c++
// End:
