/*!
  \file sref.h

  \brief Mechanism for shared references.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SREF_H__
#define __SREF_H__

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief A shared reference to an object of type \c T.

  An instance of \c shared_ref<T> contains a shared reference to an
  instance of class \c T. Class \c T must implement the shared
  reference mechanism in order for this to work.

  \sa Class \c Matrix for an example of how shared references work.
  \sa Class \c const_shared_ref for shared references from \c const objects.
*/
template <class T>
class shared_ref {
  T obj;

public:
  //! Copy constructor: makes another reference.
  shared_ref(const shared_ref& other) : obj(other.obj, other.obj.take_data()) { }

  //! Creates a shared reference to instance \a ins.
  explicit shared_ref(T& ins) : obj(ins, ins.take_data()) { }

  //! Returns a pointer to the object referenced.
  T *ptr() { return &obj; }

  //! Structure dereference operator.
  T *operator->() { return &obj; }

  //! Returns a \cpp reference to the object referenced.
  T& operator*() { return obj; }

  //! Assignment of shared references is not implemented.
  shared_ref& operator=(const shared_ref&) = delete;
};

/*!
  \brief \c const version of \c shared_ref.

  This can be used to obtain shared references of an object to which
  only a constant reference is available.

  \sa \c shared_ref.
*/
template <class T>
class const_shared_ref {
  T obj;

public:
  //! Copy constructor: makes another reference.
  const_shared_ref(const const_shared_ref& other) : obj(other.obj, other.obj.take_data()) { }

  //! Creates a shared reference to instance \a ins.
  explicit const_shared_ref(const T& ins) : obj(ins, ins.take_data()) { }

  //! Returns a pointer to the object referenced.
  const T *ptr() const { return &obj; }

  //! Structure dereference.
  const T *operator->() const { return &obj; }

  //! Returns a \c C++ reference to the object referenced.
  const T& operator*() const { return obj; }

  //! Assignment of shared references is not implemented.
  const_shared_ref& operator=(const const_shared_ref&) = delete;
};

}  // namespace

#endif

// Local variables:
// mode: c++
// End:
