/*
  tensortemp.h

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __TENSORTEMP_H__
#define __TENSORTEMP_H__

#include <vector>
#include <initializer_list>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//
// DimBounds holds the lower and upper bound for a certain dimension
// of a Tensor
//
struct DimBounds {
  int lower, upper;
  
  DimBounds(int size) : lower(0), upper(size - 1) { }
  DimBounds(int l, int u) : lower(l), upper(u) { }
};

//
// Class TensorTemplate
//
class TensorTemplate {
  int mat_size;                   // Size of corresponding matrix
  int num_dimensions;             // Number of dimensions in the tensor
  std::vector<int> lower, upper;  // Lower and upper bounds of the dimensions
  std::vector<int> size;          // The size of each dimension
  std::vector<int> displace;      // Displacement for a given position

public:
  /* Constructor and destructor */
  explicit TensorTemplate(const std::initializer_list<DimBounds>&);

  /* Get and set */
  int matrix_size() const { return mat_size; }
  //  int get_lower(int k) const { return lower[k]; }
  //int get_displacement(int k) const { return displace[k]; }
  int idx(const int *) const;

  /* Iterators */
  class iterator {
    int mat_size;                   // Matrix size
    int num_dimensions;             // Number of dimensions
    const std::vector<int>& lower;  // Lower bounds for dimensions
		const std::vector<int>& upper;  // Upper bounds for dimensions
    std::vector<int> cur;           // Current position
    int cur_idx;                    // Current index

    explicit iterator(const TensorTemplate&, int = 0);

  public:
    /* Get and set */
    int idx() const { return cur_idx; }
    bool finished() const { return cur_idx == mat_size; }

    /* Operators */
    iterator& operator++() 
    {
      if (cur_idx == mat_size) return *this;

      int i;
      for (i = num_dimensions - 1; i >= 0; i--)
        if (cur[i] == upper[i]) cur[i] = lower[i];
        else break;
      
      if (i >= 0) cur[i]++;
      
      cur_idx++;
      
      return *this;
    }
    
    const std::vector<int>& operator*() const { return cur; }
    const std::vector<int> *operator->() const { return &cur; }
    int operator[](int i) const { return cur[i]; }
    
    bool operator==(const iterator& other) const { return cur_idx == other.cur_idx; }
    bool operator!=(const iterator& other) const { return cur_idx != other.cur_idx; }

    /* Friends */
    friend class TensorTemplate;
  };

  iterator begin() const { return iterator(*this); }
  iterator end() const { return iterator(*this, mat_size); }

  /* Friends */
  friend class TensorTemplate::iterator;
};

} // namespace

#endif

// Local variables:
// mode: c++
// End:
