/*!
  \file vectormask.h

  \brief A vector bit-mask.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __VECTORMASK_H__
#define __VECTORMASK_H__

#include <vector>

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief Implements a vector bit-mask.

  A \c VectorMask contains a vector of \c n bits and a list with all
  positions that are \e on.

  This class supports iteration over the list of positions that are
  turned on. Iteration takes constant time per position turned on.
  
  \par Complexity Checking whether a bit is on or off takes constant
  time. Turning a bit on takes constant time.

  \todo At the time, it is not possible to turn a position off. This
  has to be implemented in the future (though it is not used in the
  library).

  \sa \c MatrixMask
*/
class VectorMask {
  std::vector<bool> mask;    // The mask
  std::vector<int> on_pos;   // Positions where the mask is on

public:
  //! Constructs a mask of length \a n with all positions off.
  explicit VectorMask(int n) : mask(n, false) { }

  //! Returns the size of the mask.
  int size() const { return mask.size(); }  

  //! Returns true if position \a i is on.
  bool is_on(int i) const { return mask[i]; }  

  //! Turns on the bit at position \a i.
  void turn_on(int i) { if (!mask[i]) { mask[i] = true; on_pos.push_back(i); } }

  // //! Turns off the bit at position \a i.
  // void turn_off(int i) { if (mask[i]) { mask[i] = false; on_pos.remove(i); } }

  //! Clears the mask, turning all bits off.
  void clear();

  //! Constant iterator
  typedef std::vector<int>::const_iterator iterator;
  
  //! Returns iterator to the beginning.
  iterator begin() const { return on_pos.begin(); }

  //! Returns past-the-end iterator.
  iterator end() const { return on_pos.end(); }
};

} // namespace

#endif

// Local variables:
// mode: c++
// End:
