/*
  sagesdpsl.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <new>
#include <sstream>
#include <sagesdpsl.h>
#include <mpfsdpsl.h>
#include <mpfr.h>
#include <cstdlib>
#include <fstream>
#include <cstdio>

using namespace std;
using namespace mpf_sdp;

/*
  Sets the default GMP precision to prec.
*/
void set_gmp_prec(int prec)
{
  mpf_set_default_prec(prec);
}

/*
  Sets output_eps to eps.
*/
void set_output_eps(double eps)
{
  SDPEnvironment::set_output_eps(eps);
}

/*
  Sets tolerance_eps to eps.
*/
void set_tolerance_eps(double eps)
{
  SDPEnvironment::set_tolerance_eps(eps);
}

/*
  Returns the total number of polynomial variables.
*/
int get_num_variables()
{
  return PolynomialVariable::num_variables();
}

/*
  Returns number of monomials in the polynomial.
*/
int get_poly_size(SDPSLPoly p)
{
  return ((Polynomial *) p)->size();
}

/*
  Adds a new polynomial variable.
*/
void add_poly_variable(const char *name)
{
  PolynomialVariable x(name);
}

/*
  Creates a zero polynomial.
*/
SDPSLPoly create_polynomial()
{
  return new Polynomial;
}

/*
  Destroys a polynomial.
*/
void destroy_polynomial(SDPSLPoly p)
{
  delete (Polynomial *) p;
}

/*
  Adds a given monomial to the polynomial with a given coefficient.
*/
void add_monomial(SDPSLPoly p, int n, int *exponents, MPFRNumber coeff, int prec)
{
  Monomial mon;
  for (int i = 0; i < n; i++) {
    PolynomialVariable x(i);

    mon *= Monomial(x(exponents[i]));
  }

  mpf_class tmp(0, prec);
  mpfr_get_f(tmp.get_mpf_t(), (mpfr_srcptr) coeff, MPFR_RNDN);
  *((Polynomial *) p) += tmp * mon;
}

/*
  Returns arrays that represent a polynomial.
*/
void polynomial_to_array(SDPSLPoly p, int *n, int *ncoeffs, int **exponents, 
			 MPFRNumber **coeffs, int prec)
{
  Polynomial *p_ptr = (Polynomial *) p;

  int num_vars = PolynomialVariable::num_variables();
  int num_coeffs = p_ptr->size();

  int *ret_exps = (int *) malloc(num_vars * num_coeffs * sizeof(int));

  if (!ret_exps) throw bad_alloc();

  mpfr_t *ret_coeffs = (mpfr_t *) malloc(num_coeffs * sizeof(mpfr_t));

  if (!ret_coeffs) throw bad_alloc();

  int k = 0;
  for (auto& c : *p_ptr) {
    for (int i = 0; i < num_vars; i++)
      ret_exps[k * num_vars + i] = c.first[i];

    mpfr_init2(ret_coeffs[k], prec);
    mpfr_set_f(ret_coeffs[k], c.second.get_mpf_t(), MPFR_RNDN);

    k++;
  }

  *n = num_vars;
  *ncoeffs = num_coeffs;
  *exponents = ret_exps;
  *coeffs = (MPFRNumber *) ret_coeffs;
}

/*
  Creates an m x n matrix of numbers given an initial value.
*/
NumberMatrix create_number_matrix(int m, int n, int prec, MPFRNumber init)
{
  int old_prec = mpf_get_default_prec();
  mpf_set_default_prec(prec);

  mpf_class tmp;
  mpfr_get_f(tmp.get_mpf_t(), (mpfr_srcptr) init, MPFR_RNDN);

  Matrix<mpf_class> *ret = new Matrix<mpf_class>(m, n, tmp);

  mpf_set_default_prec(old_prec);

  return ret;
}

/*
  Destroys a matrix of numbers.
*/
void destroy_number_matrix(NumberMatrix A)
{
  delete (Matrix<mpf_class> *) A;
}

/*
  Puts in num the entry (i, j) of matrix A.
*/
void get_number_entry(NumberMatrix A, int i, int j, MPFRNumber num)
{
  Matrix<mpf_class> *mat = (Matrix<mpf_class> *) A;

  mpfr_set_f((mpfr_ptr) num, (*mat)[i][j].get_mpf_t(), MPFR_RNDN);
}

/*
  Sets the entry (i, j) of matrix A to num.
*/
void set_number_entry(NumberMatrix A, int i, int j, MPFRNumber num)
{
  Matrix<mpf_class> *mat = (Matrix<mpf_class> *) A;

  mpfr_get_f((*mat)[i][j].get_mpf_t(), (mpfr_srcptr) num, MPFR_RNDN);
}

/*
  Takes the entries of a numerical matrix from a vector in row-major
  order.  
*/
void set_number_entries(NumberMatrix A, MPFRNumber *entries)
{
  Matrix<mpf_class> *mat = (Matrix<mpf_class> *) A;
  int m = mat->num_rows(), n = mat->num_cols();

  for (int i = 0; i < m; i++)
    for (int j = 0; j < n; j++)
      mpfr_get_f((*mat)[i][j].get_mpf_t(), (mpfr_srcptr) entries[i * n + j],
		 MPFR_RNDN);
}

/*
  Returns a pointer to a newly-created matrix of polynomials of the
  given size.
*/
PolyMatrix create_poly_matrix(int m, int n)
{
  return new Matrix<Polynomial>(m, n);
}

/*
  Destroys a matrix of polynomials.
*/
void destroy_poly_matrix(PolyMatrix A)
{
  delete (Matrix<Polynomial> *) A;
}

/*
  Gets entry (i, j) of a matrix of polynomials.
*/
void get_poly_entry(PolyMatrix A, int i, int j, SDPSLPoly p)
{
  Matrix<Polynomial> *mat = (Matrix<Polynomial> *) A;
  Polynomial *p_ptr = (Polynomial *) p;

  *p_ptr = (*mat)[i][j];
}

/*
  Sets entry (i, j) of a matrix of polynomials.
*/
void set_poly_entry(PolyMatrix A, int i, int j, SDPSLPoly p)
{
  Matrix<Polynomial> *mat = (Matrix<Polynomial> *) A;
  Polynomial *p_ptr = (Polynomial *) p;

  (*mat)[i][j] = *p_ptr;
}

/*
  Returns a pointer to a new matrix mask of the given size.
*/
SDPSLMatrixMask create_matrix_mask(int m, int n)
{
  return new MatrixMask(m, n);
}

/*
  Destroys a given matrix mask.
*/
void destroy_matrix_mask(SDPSLMatrixMask mask)
{
  delete (MatrixMask *) mask;
}

/*
  Returns true iff entry (i, j) is turned on.
*/
int matrix_mask_is_on(SDPSLMatrixMask mask, int i, int j)
{
  return ((MatrixMask *) mask)->is_on(i, j);
}

/*
  Turns on entry (i, j).
*/
void matrix_mask_turn_on(SDPSLMatrixMask mask, int i, int j)
{
  ((MatrixMask *) mask)->turn_on(i, j);
}

/*
  Turns off entry (i, j).
*/
void matrix_mask_turn_off(SDPSLMatrixMask mask, int i, int j)
{
  ((MatrixMask *) mask)->turn_off(i, j);
}

/*
  Clears a given mask.
*/
void clear_matrix_mask(SDPSLMatrixMask mask)
{
  ((MatrixMask *) mask)->clear();
}

/*
  Returns a pointer to a new variable given its size.
*/
SDPSLVariable create_variable(int size)
{
  return new SDPVariable(size);
}

/*
  Destroys a variable.
*/
void destroy_variable(SDPSLVariable var)
{
  delete (SDPVariable *) var;
}

/*
  Returns the index of a variable.
*/
int get_variable_idx(SDPSLVariable var)
{
  return ((SDPVariable *) var)->idx();
}

/*
  Returns a pointer to a new and empty linear combination.
*/
SDPSLLinearCombination create_linear_combination()
{
  return new SDPLinearCombination;
}

/*
  Destroys a given linear combination.
*/
void destroy_linear_combination(SDPSLLinearCombination C)
{
  delete (SDPLinearCombination *) C;
}

/*
  Adds a matrix associated with a variable to the linear combination.
*/
void linear_combination_add_matrix(SDPSLLinearCombination C, SDPSLVariable X,
				   NumberMatrix A)
{
  SDPLinearCombination *C_ptr = (SDPLinearCombination *) C;
  SDPVariable *X_ptr = (SDPVariable *) X;
  Matrix<mpf_class> *A_ptr = (Matrix<mpf_class> *) A;

  *C_ptr += (*A_ptr) * (*X_ptr);
}

/*
  Adds a masked matrix associated with a variable to the linear
  combination.
*/
void linear_combination_add_matrix_mask(SDPSLLinearCombination C, SDPSLVariable X,
					NumberMatrix A, SDPSLMatrixMask mask)
{
  SDPLinearCombination *C_ptr = (SDPLinearCombination *) C;
  SDPVariable *X_ptr = (SDPVariable *) X;
  Matrix<mpf_class> *A_ptr = (Matrix<mpf_class> *) A;
  MatrixMask *mask_ptr = (MatrixMask *) mask;

  *C_ptr += SDPMaskedMatrix((*A_ptr), *mask_ptr) * (*X_ptr);
}

/*
  Returns a pointer to a new linear constraint given the left-hand
  side, the relation as a character, and the right-hand side.
*/
SDPSLLinearConstraint create_linear_constraint(SDPSLLinearCombination lhs, char rel,
					       MPFRNumber rhs)
{
  SDPLinearCombination *C = (SDPLinearCombination *) lhs;

  int prec = mpfr_get_prec((mpfr_srcptr) rhs);
  mpf_class tmp(0, prec);
  mpfr_get_f(tmp.get_mpf_t(), (mpfr_srcptr) rhs, MPFR_RNDN);

  switch (rel) {
  case '=':
    return new SDPLinearConstraint(*C, sdp_eq, tmp);
  case '<':
    return new SDPLinearConstraint(*C, sdp_leq, tmp);
  case '>':
    return new SDPLinearConstraint(*C, sdp_geq, tmp);
  }

  throw 1;  // We should never be here!

  return 0;
}

/*
  Destroys a linear constraint.
*/
void destroy_linear_constraint(SDPSLLinearConstraint C)
{
  delete (SDPLinearConstraint *) C;
}

/*
  Sets the right-hand side of a linear constraint.
*/
void linear_constraint_set_rhs(SDPSLLinearConstraint C, MPFRNumber rhs)
{
  int prec = mpfr_get_prec((mpfr_srcptr) rhs);
  mpf_class tmp(0, prec);
  mpfr_get_f(tmp.get_mpf_t(), (mpfr_srcptr) rhs, MPFR_RNDN);
  

  ((SDPLinearConstraint *) C)->rhs(tmp);
}

/*
  Returns a pointer to a new polynomial combination.
*/
SDPSLPolynomialCombination create_poly_combination()
{
  return new SDPPolynomialCombination;
}

/*
  Destroys a polynomial combination.
*/
void destroy_poly_combination(SDPSLPolynomialCombination C)
{
  delete (SDPPolynomialCombination *) C;
}

/*
  Adds a matrix associated with a variable.
*/
void poly_combination_add_matrix(SDPSLPolynomialCombination C, SDPSLVariable X,
				 PolyMatrix A)
{
  SDPPolynomialCombination *C_ptr = (SDPPolynomialCombination *) C;
  SDPVariable *X_ptr = (SDPVariable *) X;
  Matrix<Polynomial> *A_ptr = (Matrix<Polynomial> *) A;

  *C_ptr += (*A_ptr) * (*X_ptr);
}

/*
  Adds a matrix and a mask associated with a variable.
*/
void poly_combination_add_matrix_mask(SDPSLPolynomialCombination C, SDPSLVariable X,
				      PolyMatrix A, SDPSLMatrixMask mask)
{
  SDPPolynomialCombination *C_ptr = (SDPPolynomialCombination *) C;
  SDPVariable *X_ptr = (SDPVariable *) X;
  Matrix<Polynomial> *A_ptr = (Matrix<Polynomial> *) A;
  MatrixMask *mask_ptr = (MatrixMask *) mask;

  *C_ptr += SDPMaskedPolyMatrix(*A_ptr, *mask_ptr) * (*X_ptr);
}

/*
  Returns a pointer to a new polynomial constraint.
*/
SDPSLPolynomialConstraint create_poly_constraint(SDPSLPolynomialCombination lhs, 
						 SDPSLPoly p)
{
  return new SDPPolynomialConstraint(*((SDPPolynomialCombination *) lhs), 
				     *((Polynomial *) p));
}

/*
  Destroys a polynomial constraint.
*/
void destroy_poly_constraint(SDPSLPolynomialConstraint C)
{
  delete (SDPPolynomialConstraint *) C;
}

/*
  Returns a pointer to a new sparse monomial basis.
*/
SDPSLSparseMonomialBasis create_s_monomial_basis()
{
  return new SparseMonomialBasis;
}

/*
  Creates a monomial basis given a constraint.
*/
SDPSLSparseMonomialBasis create_s_monomial_basis_cons(SDPSLPolynomialConstraint C)
{
  return new SparseMonomialBasis(*((SDPPolynomialConstraint *) C));
}

/*
  Destroys a given sparse monomial basis.
*/
void destroy_s_monomial_basis(SDPSLSparseMonomialBasis basis)
{
  delete (SparseMonomialBasis *) basis;
}

/*
  Adds all monomials of a polynomial to the basis.
*/
void s_monomial_basis_add(SDPSLSparseMonomialBasis basis, SDPSLPoly p)
{
  SparseMonomialBasis *basis_ptr = (SparseMonomialBasis *) basis;
  Polynomial *p_ptr = (Polynomial *) p;

  for (auto& c : *p_ptr)
    basis_ptr->add_monomial(c.first);
}

/*
  Returns a pointer to a sparse basis changer.


*/
SDPSLSparseBasisChanger create_s_basis_changer(SDPSLSparseMonomialBasis mon_basis,
					       int n, SDPSLPoly *poly_basis, int *error)
{
  SparseMonomialBasis *mon_ptr = (SparseMonomialBasis *) mon_basis;
  vector<Polynomial> tmp(n);
  
  for (int i = 0; i < n; i++)
    tmp[i] = *((Polynomial *) poly_basis[i]);

  try {
    *error = 0;

    return new SparseBasisChanger(*mon_ptr, tmp);
  }
  catch(InvalidBasis e) {
    switch (e.code()) {
    case InvalidBasis::incompatible_bases:
      *error = INCOMPATIBLE_BASES_ERROR;
      break;

    case InvalidBasis::singular_basis_change_matrix:
      *error = SINGULAR_MATRIX_ERROR;
      break;

    default:
      *error = 0;
    }

    return 0;
  }
}

/*
  Destroys a basis changer.
*/
void destroy_s_basis_changer(SDPSLSparseBasisChanger changer)
{
  delete (SparseBasisChanger *) changer;
}

/*
  Returns a pointer to a new constraint container.
*/
SDPSLConstraintContainer create_constraint_container()
{
  return new SDPConstraintContainer;
}

/*
  Destroys a given constraint container.
*/
void destroy_constraint_container(SDPSLConstraintContainer cc)
{
  delete (SDPConstraintContainer *) cc;
}

/*
  Adds a linear constraint to a constraint container.

  Returns zero if successful, nonzero if the constraint is infeasible.
*/
int cc_add_linear(SDPSLConstraintContainer cc, SDPSLLinearConstraint constraint)
{
  SDPConstraintContainer *cc_ptr = (SDPConstraintContainer *) cc;
  SDPLinearConstraint *cons_ptr = (SDPLinearConstraint *) constraint;

  try {
    cc_ptr->add_constraint(*cons_ptr);
  }
  catch (InfeasibleConstraint) {
    return 1;
  }

  return 0;
}

/*
  Adds a polynomial constraint with the standard monomial basis. 

  Returns zero if successful, INFEASIBLE_CONSTRAINT if the constraint
  is infeasible, and REPRESENTATION_ERROR if some polynomial could not
  be expanded.
*/
int cc_add_poly(SDPSLConstraintContainer cc, SDPSLPolynomialConstraint constraint,
                int progress, const char *msg)
{
  SDPConstraintContainer *cc_ptr = (SDPConstraintContainer *) cc;
  SDPPolynomialConstraint *cons_ptr = (SDPPolynomialConstraint *) constraint;

  try {
    if (progress) {
      ProgressBar progress(msg, 40);
      cc_ptr->add_constraint(*cons_ptr, progress);
    }
    else
      cc_ptr->add_constraint(*cons_ptr);
  }
  catch (InfeasibleConstraint) {
    return INFEASIBLE_CONSTRAINT_ERROR;
  }
  catch (InvalidBasis) {
    return REPRESENTATION_ERROR;
  }

  return 0;
}

/*
  Adds a polynomial constraint with a basis changer.

  Returns zero if successful, INFEASIBLE_CONSTRAINT if the constraint
  is infeasible, and REPRESENTATION_ERROR if some polynomial could not
  be expanded.
*/
int cc_add_poly_with_basis(SDPSLConstraintContainer cc,
			   SDPSLPolynomialConstraint constraint,
			   SDPSLSparseBasisChanger changer,
                           int progress, const char *msg) 
{
  SDPConstraintContainer *cc_ptr = (SDPConstraintContainer *) cc;
  SDPPolynomialConstraint *cons_ptr = (SDPPolynomialConstraint *) constraint;
  SparseBasisChanger *changer_ptr = (SparseBasisChanger *) changer;

  try {
    if (progress) {
      ProgressBar progress(msg, 40);
      cc_ptr->add_constraint(*cons_ptr, *changer_ptr, progress);
    }
    else
      cc_ptr->add_constraint(*cons_ptr, *changer_ptr);
  }
  catch (InfeasibleConstraint) {
    return INFEASIBLE_CONSTRAINT_ERROR;
  }
  catch (InvalidBasis) {
    return REPRESENTATION_ERROR;
  }

  return 0;
}

/*
  Returns number of raw constraints in the container.
*/
int cc_get_num_constraints(SDPSLConstraintContainer cc)
{
  return ((SDPConstraintContainer *) cc)->constraints().size();
}

/*
  Clears the container.
*/
void cc_clear(SDPSLConstraintContainer cc)
{
  ((SDPConstraintContainer *) cc)->clear();
}

/*
  Returns a pointer to a new problem.
*/
SDPSLProblem create_problem()
{
  return new SDPProblem;
}

/*
  Destroys a problem.
*/
void destroy_problem(SDPSLProblem problem)
{
  delete (SDPProblem *) problem;
}

/*
  Sets the objective function of a problem.
*/
void problem_set_objective(SDPSLProblem problem, SDPSLLinearCombination C)
{
  ((SDPProblem *) problem)->set_objective(*((SDPLinearCombination *) C));
}

/*
  Adds a constraint container to a problem.
*/
void problem_add_container(SDPSLProblem problem, SDPSLConstraintContainer cc)
{
  *((SDPProblem *) problem) << *((SDPConstraintContainer *) cc);
}

/*
  Writes a problem in SDPA format to a file with a given precision.
  
  Returns:

  - ZERO -- if successful

  - FILE_ERROR -- if fails to open file

  - VALUE_ERROR -- if problem is empty
*/
int write_sdpa(SDPSLProblem problem, const char *filename, int prec)
{
  ofstream outfile(filename);

  if (!outfile) return FILE_ERROR;
  
  outfile.precision(prec);

  try {
    outfile << SDPAWriter(*((SDPProblem *) problem));
  }
  catch (ValueError) {
    return VALUE_ERROR;
  }

  return 0;
}

/*
  Returns a pointer to a new solution map.
*/
SDPSLSolution create_solution()
{
  return new SDPSolution<mpf_class>;
}

/*
  Destroys a solution map.
*/
void destroy_solution(SDPSLSolution sol)
{
  delete (SDPSolution<mpf_class> *) sol;
}

/*
  Adds a matrix to a solution.
*/
void solution_add_matrix(SDPSLSolution sol, SDPSLVariable X, NumberMatrix A)
{
  ((SDPSolution<mpf_class> *) sol)->set_value(*((SDPVariable *) X), SDPOutputMatrix<mpf_class>(*((Matrix<mpf_class> *) A)));
}

/*
  Reads a CSDP solution to a given problem.

  Returns FILE_ERROR if fails to open the file, READ_ERROR if there is
  a problem with the file, in which case error_line is set to the line
  where the error happened. If successful, returns 0.
*/
int csdp_problem_read_solution(SDPSLProblem problem, const char *filename,
			       SDPSLSolution sol, int *error_line)
{
  ifstream infile(filename);

  if (!infile)
    return FILE_ERROR;

  try {
    infile >> CSDPReader<mpf_class>(*((SDPProblem *) problem), *((SDPSolution<mpf_class> *) sol));
  }
  catch (ReadError e) {
    *error_line = e.lineno();

    return READ_ERROR;
  }

  return 0;
}

/*
  Reads a CSDP solution given a NULL-terminated list of variables.

  Returns FILE_ERROR if fails to open the file, READ_ERROR if there is
  a problem with the file, in which case error_line is set to the line
  where the error happened. If successful, returns 0.
*/
int csdp_list_read_solution(SDPSLVariable *vars, const char *filename,
			    SDPSLSolution sol, int *error_line)
{
  list<SDPVariable> var_list;

  for (SDPSLVariable *v = vars; *v; v++)
    var_list.push_back(*((SDPVariable *) *v));

  ifstream infile(filename);

  if (!infile)
    return FILE_ERROR;

  try {
    infile >> CSDPReader<mpf_class>(var_list, *((SDPSolution<mpf_class> *) sol));
  }
  catch (ReadError e) {
    *error_line = e.lineno();

    return READ_ERROR;
  }

  return 0;  
}

/*
  Computes the feasibility measures.

  - ccs -- NULL terminated array of constraint containers

  - sol -- solution

  - N -- contains number of measures

  - measures -- the measures
*/
void sdpsl_feasibility_measures(SDPSLConstraintContainer *ccs,
				SDPSLSolution sol, int *N,
				MPFRNumber **measures)
{
  list<const SDPConstraintContainer *> cc_list;

  for (SDPSLConstraintContainer *cc = ccs; *cc; cc++)
    cc_list.push_back((SDPConstraintContainer *) *cc);

  SDPSolution<mpf_class> *sol_ptr = (SDPSolution<mpf_class> *) sol;
  vector<mpf_class> mpf_measures;

  SDPSolutionAnalyzer analyzer;

  analyzer.feasibility_measures(cc_list, *sol_ptr, mpf_measures);

  /* Assembles the measures */
  mpfr_t *ret_measures = (mpfr_t *) malloc(mpf_measures.size() * sizeof(mpfr_t));

  if (!ret_measures) throw bad_alloc();

  int prec = mpf_get_default_prec();

  for (int i = 0; i < mpf_measures.size(); i++) {
    mpfr_init2(ret_measures[i], prec);
    mpfr_set_f(ret_measures[i], mpf_measures[i].get_mpf_t(), MPFR_RNDN);
  }

  *measures = (MPFRNumber *) ret_measures;
  *N = mpf_measures.size();    
}

/*
  Computes the eigenvalues of a matrix using LAPACK.

  The eigenvalues are placed in eigs, which is created for this
  purpose. The size of the vector is the size of the matrix.

  The function returns 0 if successful, nonzero if LAPACK fails.
*/
int sdpsl_double_eigenvalues(NumberMatrix A, SDPSLMatrixMask mask, 
			     double **eigs)
{
  Matrix<mpf_class> *A_ptr = (Matrix<mpf_class> *) A;
  MatrixMask *mask_ptr = (MatrixMask *) mask;
  vector<double> eigenvalues;

  try {
    SDPSolutionAnalyzer analyzer;

    if (mask_ptr)
      analyzer.double_eigenvalues(*A_ptr, *mask_ptr, eigenvalues);
    else
      analyzer.double_eigenvalues(*A_ptr, eigenvalues);
  }
  catch (...) {
    return 1;
  }

  /* We return the result */
  double *ret_eigs = (double *) malloc(eigenvalues.size() * sizeof(double));

  if (!ret_eigs) throw bad_alloc();

  for (int i = 0; i < eigenvalues.size(); i++)
    ret_eigs[i] = eigenvalues[i];

  *eigs = ret_eigs;

  return 0;
}

/*
  Projects a solution using arbitrary-precision arithmetic.

  Returns an error code depending on the exception thrown by the
  analyzer, or zero if successful.
*/
int sdpsl_gmp_project(SDPSLConstraintContainer *ccs,
		      SDPSLSolution sol, SDPSLSolution res)
{ 
  list<const SDPConstraintContainer *> cc_list;

  for (SDPSLConstraintContainer *cc = ccs; *cc; cc++)
    cc_list.push_back((SDPConstraintContainer *) *cc);

  SDPSolution<mpf_class> *sol_ptr = (SDPSolution<mpf_class> *) sol;
  SDPSolution<mpf_class> *res_ptr = (SDPSolution<mpf_class> *) res;

  SDPSolutionAnalyzer analyzer;

  try {
    analyzer.project_solution(cc_list, *sol_ptr, *res_ptr);
  }
  catch (ValueError) {
    return VALUE_ERROR;
  }
  catch (AnalyzerError) {
    return INFEASIBLE_SYSTEM;
  }

  return 0;
}
