/*
  sagesdpsl.h

  A C wrapper for SDPSL.

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SAGESDPSL_H__
#define __SAGESDPSL_H__

//
// Errors
//
#define FILE_ERROR                  1
#define VALUE_ERROR                 2
#define INCOMPATIBLE_BASES_ERROR    3
#define SINGULAR_MATRIX_ERROR       4
#define INFEASIBLE_CONSTRAINT_ERROR 5
#define REPRESENTATION_ERROR        6
#define READ_ERROR                  7
#define INFEASIBLE_SYSTEM           8

//
// Environment variables
//
extern "C" void set_gmp_prec(int prec);
extern "C" void set_output_eps(double eps);
extern "C" void set_tolerance_eps(double eps);

//
// Numbers
//
typedef void * MPFRNumber;

//
// Polynomials
//
typedef void * SDPSLPoly;

extern "C" int get_num_variables();
extern "C" int get_poly_size(SDPSLPoly p);
extern "C" void add_poly_variable(const char *name);
extern "C" SDPSLPoly create_polynomial();
extern "C" void destroy_polynomial(SDPSLPoly p);
extern "C" void add_monomial(SDPSLPoly p, int n, int *exponents, MPFRNumber coeff, int prec);
extern "C" void polynomial_to_array(SDPSLPoly p, int *n, int *ncoeffs, int **exponents, 
				    MPFRNumber **coeffs, int prec);

//
// Numerical matrices
//
typedef void * NumberMatrix;

extern "C" NumberMatrix create_number_matrix(int m, int n, int prec, MPFRNumber init);
extern "C" void destroy_number_matrix(NumberMatrix A);
extern "C" void get_number_entry(NumberMatrix A, int i, int j, MPFRNumber num);
extern "C" void set_number_entry(NumberMatrix A, int i, int j, MPFRNumber num);
extern "C" void set_number_entries(NumberMatrix A, MPFRNumber *entries);

//
// Matrices of polynomials
//
typedef void * PolyMatrix;

extern "C" PolyMatrix create_poly_matrix(int m, int n);
extern "C" void destroy_poly_matrix(PolyMatrix A);
extern "C" void get_poly_entry(PolyMatrix A, int i, int j, SDPSLPoly p);
extern "C" void set_poly_entry(PolyMatrix A, int i, int j, SDPSLPoly p);

//
// MatrixMask
//
typedef void * SDPSLMatrixMask;

extern "C" SDPSLMatrixMask create_matrix_mask(int m, int n);
extern "C" void destroy_matrix_mask(SDPSLMatrixMask mask);
extern "C" int matrix_mask_is_on(SDPSLMatrixMask mask, int i, int j);
extern "C" void matrix_mask_turn_on(SDPSLMatrixMask mask, int i, int j);
extern "C" void matrix_mask_turn_off(SDPSLMatrixMask mask, int i, int j);
extern "C" void clear_matrix_mask(SDPSLMatrixMask mask);

//
// SDPVariable
//
typedef void * SDPSLVariable;

extern "C" SDPSLVariable create_variable(int size);
extern "C" void destroy_variable(SDPSLVariable var);
extern "C" int get_variable_idx(SDPSLVariable var);

//
// SDPLinearCombination
//
typedef void * SDPSLLinearCombination;

extern "C" SDPSLLinearCombination create_linear_combination();
extern "C" void destroy_linear_combination(SDPSLLinearCombination C);
extern "C" void linear_combination_add_matrix(SDPSLLinearCombination C, SDPSLVariable X,
					      NumberMatrix A);
extern "C" void linear_combination_add_matrix_mask(SDPSLLinearCombination C, SDPSLVariable X,
						   NumberMatrix A, SDPSLMatrixMask mask);

//
// SDPLinearConstraint
//
typedef void * SDPSLLinearConstraint;

extern "C" SDPSLLinearConstraint create_linear_constraint(SDPSLLinearCombination lhs, char rel,
							  MPFRNumber rhs);
extern "C" void destroy_linear_constraint(SDPSLLinearConstraint C);
extern "C" void linear_constraint_set_rhs(SDPSLLinearConstraint C, MPFRNumber rhs);

//
// SDPPolynomialCombination
//
typedef void * SDPSLPolynomialCombination;

extern "C" SDPSLPolynomialCombination create_poly_combination();
extern "C" void destroy_poly_combination(SDPSLPolynomialCombination C);
extern "C" void poly_combination_add_matrix(SDPSLPolynomialCombination C, SDPSLVariable X,
					    PolyMatrix A);
extern "C" void poly_combination_add_matrix_mask(SDPSLPolynomialCombination C, SDPSLVariable X,
						 PolyMatrix A, SDPSLMatrixMask mask);

//
// SDPPolynomialConstraint
//
typedef void * SDPSLPolynomialConstraint;

extern "C" SDPSLPolynomialConstraint create_poly_constraint(SDPSLPolynomialCombination lhs, 
							    SDPSLPoly p);
extern "C" void destroy_poly_constraint(SDPSLPolynomialConstraint C);

//
// SparseMonomialBasis
//
typedef void * SDPSLSparseMonomialBasis;

extern "C" SDPSLSparseMonomialBasis create_s_monomial_basis();
extern "C" SDPSLSparseMonomialBasis create_s_monomial_basis_cons(SDPSLPolynomialConstraint C);
extern "C" void destroy_s_monomial_basis(SDPSLSparseMonomialBasis basis);
extern "C" void s_monomial_basis_add(SDPSLSparseMonomialBasis basis, SDPSLPoly p);

//
// SparseBasisChanger
//
typedef void * SDPSLSparseBasisChanger;

extern "C" SDPSLSparseBasisChanger create_s_basis_changer(SDPSLSparseMonomialBasis mon_basis,
							  int n, SDPSLPoly *poly_basis, int *error);
extern "C" void destroy_s_basis_changer(SDPSLSparseBasisChanger changer);

//
// SDPConstraintContainer
//
typedef void * SDPSLConstraintContainer;

extern "C" SDPSLConstraintContainer create_constraint_container();
extern "C" void destroy_constraint_container(SDPSLConstraintContainer cc);
extern "C" int cc_add_linear(SDPSLConstraintContainer cc, SDPSLLinearConstraint constraint);
extern "C" int cc_add_poly(SDPSLConstraintContainer cc, SDPSLPolynomialConstraint constraint,
                           int progress, const char *msg);
extern "C" int cc_add_poly_with_basis(SDPSLConstraintContainer cc,
				      SDPSLPolynomialConstraint constraint,
				      SDPSLSparseBasisChanger changer,
                                      int progress, const char *msg);
extern "C" int cc_get_num_constraints(SDPSLConstraintContainer cc);
extern "C" void cc_clear(SDPSLConstraintContainer cc);

//
// SDPProblem
//
typedef void * SDPSLProblem;

extern "C" SDPSLProblem create_problem();
extern "C" void destroy_problem(SDPSLProblem problem);
extern "C" void problem_set_objective(SDPSLProblem problem, SDPSLLinearCombination C);
extern "C" void problem_add_container(SDPSLProblem problem, SDPSLConstraintContainer cc);

//
// SDPAWriter
//
extern "C" int write_sdpa(SDPSLProblem problem, const char *filename, int prec);

//
// SDPSolution
//
typedef void * SDPSLSolution;

extern "C" SDPSLSolution create_solution();
extern "C" void destroy_solution(SDPSLSolution sol);
extern "C" void solution_add_matrix(SDPSLSolution sol, SDPSLVariable X, NumberMatrix A);

//
// CSDPReader
//
extern "C" int csdp_problem_read_solution(SDPSLProblem problem, const char *filename,
					  SDPSLSolution sol, int *error_line);
extern "C" int csdp_list_read_solution(SDPSLVariable *vars, const char *filename,
				       SDPSLSolution sol, int *error_line);

//
// SDPSolutionAnalyzer
//
extern "C" void sdpsl_feasibility_measures(SDPSLConstraintContainer *ccs,
					   SDPSLSolution sol, int *N,
					   MPFRNumber **measures);
extern "C" int sdpsl_double_eigenvalues(NumberMatrix A, SDPSLMatrixMask mask, 
					double **eigs);
extern "C" int sdpsl_gmp_project(SDPSLConstraintContainer *ccs,
				 SDPSLSolution sol, SDPSLSolution res);


#endif

// Local variables:
// mode: c++
// End:
