"""
A program to compute the semidefinite programming bound of D. de Laat,
F.M. de Oliveira Filho, and F. Vallentin for the densities of binary
sphere packings.

REFERENCES:

- D. de Laat, F.M. de Oliveira Filho, and F. Vallentin, Upper bounds
  for packings of spheres of several radii, arXiv:1206.2608, 2012,
  31pp.

AUTHOR:

- Fernando M. de Oliveira Filho <fmario AT gmail.com>
"""

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

from sage.sdpsl import *


def laguerre_polys(a, d, PR, x):
    """
    Return a list of the first few Laguerre polynomials.

    This function returns a list with the polynomials L_k^a(x) for k =
    0, ..., d. The polynomials are elements of the polynomial ring PR.
    """

    a = PR.base_ring()(a)
    
    ret = [ PR(1), -x + a + 1 ]
    for k in xrange(2, d + 1):
        ret.append((2 + (a - 1 - x) / k) * ret[-1]
                   - (1 + (a - 1) / k) * ret[-2])

    return ret


class BinaryPacking:
    """
    A class that generates binary packing SDP problems.

    EXAMPLES:

    One can get the SDPSL problem and then manipulate it as needed::

        sage: foo = BinaryPacking(2, 15, 0.7, 1.0)
        sage: problem = foo.make_problem()
        sage: solver = CSDPSolver()
        sage: solver.solve(problem, verbose = True)

    Or one can save the problem to an SDPA file directly::

        sage: foo.write_sdpa('problem.sdpa')
    """
    
    def __init__(self, n, d, r1 = 1/2, r2 = 1/2, prec = 256):
        """
        Create a binary packing problem generator.

        INPUT:

        - ``n`` - the dimension of the space, an integer >= 2.

        - ``d`` - degree for SOS computations, an odd integer >= 3.

        - ``r1`` - radius of first sphere.

        - ``r2`` - radius of the second sphere.

        - ``prec`` - precision for GMP.
        """
        
        n = int(n)
        d = int(d)

        if n < 2:
            raise ValueError('Dimension must be >= 2')

        if d % 2 == 0:
            raise ValueError('Max. degree must be odd')
        
        self.n = n
        self.d = d
        
        self.RH = RealField(prec)
        self.pi = self.RH.pi()
        
        self.r1 = self.RH(r1)
        self.r2 = self.RH(r2)

        self.vol1 = self.RH(2 * self.pi^(n / 2) * self.r1^n / (gamma(n / 2) * n))
        self.vol2 = self.RH(2 * self.pi^(n / 2) * self.r2^n / (gamma(n / 2) * n))        

        # Initializes SDPSL
        sdpsl_set_gmp_prec(prec)
        sdpsl_set_tolerance_eps(1e-40)
        sdpsl_set_output_eps(1e-30)

        # The polynomial ring
        self.PR = PolynomialRing(self.RH, 'x')
        x = self.PR.gen()

        # Bases of polynomials we work with
        self.mon_basis = [ x^(2 * k) for k in xrange(d + 1) ]
        
        tmp = laguerre_polys(n / 2.0 - 1, d, self.PR, 2 * self.pi * x^2)
        self.lag_basis = [ p / max(abs(c) for c in p.coefficients()) for p in tmp ]

        # Computes polynomials for Fourier inversion
        tmp = laguerre_polys(n / 2.0 - 1, d, self.PR, self.pi * x^2)
        self.fourier_polys = [ factorial(k) * self.pi^(-k) * tmp[k] for k in xrange(len(tmp)) ]

        # The moment matrix
        v = vector(self.PR, self.lag_basis[:floor(d / 2) + 1])            
        V = self.V = v.column() * v.row()

        # The matrices of Fourier inverses
        def inverse(p):
            ret = 0
            for c, k in zip(p.coefficients(), p.exponents()):
                ret += c * self.fourier_polys[k / 2]

            return ret
        
        F = self.F = matrix(self.PR, self.V.nrows(), self.V.ncols(), lambda i, j: inverse(self.V[i, j]))
        x2_F = self.x2_F = matrix(self.PR, self.V.nrows(), self.V.ncols(), lambda i, j: inverse(x^2 * self.V[i, j]))
        
        # Variables of the problem
        S0 = SDPVariable(d + 1)
        S1 = SDPVariable(d + 1)
        Q11 = SDPVariable(floor(d / 2) + 1)
        Q12 = SDPVariable(floor(d / 2) + 1)
        Q22 = SDPVariable(floor(d / 2) + 1)
        N = SDPVariable(2)

        # Y matrices
        Y11 = matrix(self.RH, [ [ 1, 0 ], [ 0, 0 ] ])
        Y22 = matrix(self.RH, [ [ 0, 0 ], [ 0, 1 ] ])
        Y12 = matrix(self.RH, [ [ 0, 0.5 ], [ 0.5, 0 ] ])

        #
        # We create the constraints
        #
        
        # Container for the constraints of the problem
        self.cc = SDPConstraintContainer()

        # The normalization constraint
        self.cc.add_constraint(Y11 * N + (-Y11.tensor_product(V(0))) * S0 == -self.vol1)
        self.cc.add_constraint(Y22 * N + (-Y22.tensor_product(V(0))) * S0 == -self.vol2)
        self.cc.add_constraint(Y12 * N + (-Y12.tensor_product(V(0))) * S0 == -sqrt(self.vol1 * self.vol2))
        
        # The SOS constraints
        try:
            changer = SDPSparseBasisChanger(self.mon_basis, self.lag_basis)
        except ValueError as ve:
            raise ValueError(ve.message + '\nTry decreasing tolerance_eps with the command: sdpsl_set_tolerance_eps')

        self.cc.add_constraint(Y11.tensor_product(F) * S0 + Y11.tensor_product(x2_F) * S1
                               + (x^2 - (self.r1 + self.r1)^2) * V * Q11 == 0, changer, progress = True)

        self.cc.add_constraint(Y22.tensor_product(F) * S0 + Y22.tensor_product(x2_F) * S1
                               + (x^2 - (self.r2 + self.r2)^2) * V * Q22 == 0, changer, progress = True)
        
        self.cc.add_constraint(Y12.tensor_product(F) * S0 + Y12.tensor_product(x2_F) * S1
                               + (x^2 - (self.r1 + self.r2)^2) * V * Q12 == 0, changer, progress = True)
        
        # The objective function
        s = SDPVariable(1) # For the maximum...
        
        self.cc.add_constraint(1 * s + (-Y11.tensor_product(F(0))) * S0
                               + (-Y11.tensor_product(x2_F(0))) * S1 >= 0)
        
        self.cc.add_constraint(1 * s + (-Y22.tensor_product(F(0))) * S0
                               + (-Y22.tensor_product(x2_F(0))) * S1 >= 0)

        self.obj = (-1) * s

    def make_problem(self):
        """
        Return the SDPSL problem that computes the bound.
        """
        
        problem = SDPProblem()

        problem.set_objective(self.obj)
        problem.add_container(self.cc)

        return problem

    def write_sdpa(self, filename):
        """
        Write the SDP problem computing the bound to a file.
        """
        
        problem = self.make_problem()
        problem.write_sdpa(filename, prec = 30)

        
# Local variables:
# mode: python
# End:
