"""
A function that returns an SDP whose optimal value is the global
minimum of a univariate real polynomial.

AUTHOR:

- Fernando M. de Oliveira Filho <fmario AT gmail.com>
"""

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

from sage.sdpsl import *

def poly_min_sos(p):
    """
    Return an SDPSL problem whose optimal value is the global minimum
    of the univariate polynomial ``p``, using the standard monomial
    basis.
    """
    
    PR = PolynomialRing(RR, 'x')
    x = PR.gen()

    try:
        q = PR(p)
    except:
        raise ValueError('p must be a univariate polynomial')

    d = p.degree()

    if d % 2 != 0:
        raise ValueError('p must have even degree')

    # Variables for the SDP
    X = SDPVariable(1 + d / 2)
    s, t = SDPVariable(1), SDPVariable(1)
    cc = SDPConstraintContainer()
    
    # The constraint
    v = vector(PR, [ x^k for k in xrange(1 + d / 2) ])
    R = v.column() * v.row()  # This is a moment matrix

    cc.add_constraint(R * X + 1 * s + (-1) * t == q)

    # We create the problem and return it
    problem = SDPProblem()

    problem.set_objective(1 * s + (-1) * t)
    problem.add_container(cc)

    return problem

def poly_min_laguerre(p):
    """
    Return an SDPSL problem whose optimal value is the global minimum
    of the univariate polynomial ``p``, using a basis of Laguerre
    polynomials.
    """
    
    PR = PolynomialRing(RealField(256), 'x')
    x = PR.gen()

    try:
        q = PR(p)
    except:
        raise ValueError('p must be a univariate polynomial')

    d = p.degree()

    if d % 2 != 0:
        raise ValueError('p must have even degree')

    # Variables for the SDP
    X = SDPVariable(1 + d / 2)
    s, t = SDPVariable(1), SDPVariable(1)
    cc = SDPConstraintContainer()

    # A basis of Laguerre polynomials
    changer = SDPSparseBasisChanger([ x^k for k in xrange(d + 1) ], [ PR(laguerre(k, x)) for k in xrange(d + 1) ])
    
    # The constraint
    v = vector(PR, [ x^k for k in xrange(1 + d / 2) ])
    R = v.column() * v.row()  # This is a moment matrix

    cc.add_constraint(R * X + 1 * s + (-1) * t == q, changer)

    # We create the problem and return it
    problem = SDPProblem()

    problem.set_objective(1 * s + (-1) * t)
    problem.add_container(cc)

    return problem


# Local variables:
# mode: python
# End:
