"""
A function to create an SDP that computes the Lovász theta number of a
SAGE graph.

AUTHOR:

- Fernando M. de Oliveira Filho <fmario AT gmail.com>
"""

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

from sage.sdpsl import *

def make_theta_sdp(G):
    """
    Create an SDPSL SDP problem that computes the theta number of a SAGE graph G.

    INPUT:

    - ``G`` -- the SAGE graph.

    Returns the problem created.

    EXAMPLE:

    Lets compute the theta number of the pentagon::

        sage: G = graphs.CycleGraph(5)
        sage: problem = make_theta_sdp(G)
        sage: solver = CSDPSolver()
        sage: ret = solver.solve(problem, verbose = True)
        sage: ret.primal_value
        2.23606800000000
    """
    
    # Vertices are not necessarily integers; we associate integers with the vertices.
    vidx = dict(zip(G.vertices(), range(G.num_verts())))

    # Constraint generation
    cc = SDPConstraintContainer()
    
    X = SDPVariable(G.num_verts())
    J = SDPMatrix(G.num_verts(), initial_value = 1)

    # Creates the constraints for the edges
    mask = SDPMatrixMask(G.num_verts())
    edge_constraint = (J, mask) * X == 0

    for u, v in G.edges(labels = False):
        mask[vidx[u], vidx[v]] = True
        cc.add_constraint(edge_constraint)
        mask[vidx[u], vidx[v]] = False

    # Then the trace constraint
    cc.add_constraint(identity_matrix(RR, G.num_verts()) * X == 1)

    # And the problem
    problem = SDPProblem()

    problem.set_objective(J * X)
    problem.add_container(cc)

    return problem
        

# Local variables:
# mode: python
# End:
