#
# A wrapper for CSDP with SDPSL
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

import re, os, tempfile, subprocess

def read_csdp_solution(filename, variable_src, solution):
    """
    read_csdp_solution(filename, variable_src, solution)
    
    Read a solution to an SDP in CSDP format.

    INPUT:

    - ``filename`` -- the file containing the solution.

    - ``variable_src`` -- to read a solution we need to know the
      variables used to generate the problem. This argument is either
      an instance of :class:`SDPProblem`, from which one can obtain a
      list of variables, or else a list of variables (instances of
      :class:`SDPVariable`).

    - ``solution`` -- is an instance of :class:`SDPSolution` that will
      contain the solution. Notice that only variables that occur in
      ``solution`` will be set; all others will be ignored. This does
      not mean, however, that only the variables in ``solution`` need
      to be given in ``variable_src``: all variables of the problem
      have to be specified there.
    """
    cdef int lineno
    cdef SDPVariable_wrapper var_wrapper
    cdef SDPProblem_wrapper prob_wrapper 
    cdef SDPMatrixSolution_wrapper sol_wrapper
    cdef SDPSLVariable *var_vector

    tmpsol = solution._get_sdp_matrix_map()
    sol_wrapper = tmpsol._solmap

    if isinstance(variable_src, SDPProblem):
        prob_wrapper = variable_src._problem

        try:
            sig_on()
            ret = csdp_problem_read_solution(prob_wrapper._problem, filename, 
                                             sol_wrapper._solmap, cython.address(lineno))
        finally:
            sig_off()
    else:
        n = len(variable_src)
        var_vector = <SDPSLVariable *> malloc((n + 1) * cython.sizeof(SDPSLVariable))

        if var_vector is NULL:
            raise MemoryError('Failed to allocate memory for var_vector')

        for i from 0 <= i < n:
            X = variable_src[i]

            if not isinstance(X, SDPVariable):
                free(var_vector)
                raise ValueError('Variables must be instances of SDPVariable')

            var_wrapper = X._var
            var_vector[i] = var_wrapper._var

        var_vector[n] = NULL

        try:
            sig_on()
            ret = csdp_list_read_solution(var_vector, filename,
                                          sol_wrapper._solmap, cython.address(lineno))
        finally:
            free(var_vector)
            sig_off()

    if ret == FILE_ERROR:
        raise IOError('Cannot open file "%s"' % filename)
    elif ret == READ_ERROR:
        raise IOError('"%s", line %d: error reading from file' % (filename, lineno))

    solution._set_from_matrix_map(tmpsol)

    
def run_cmd(cmd, verbose = False):
    p = subprocess.Popen(cmd, bufsize = -1, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    ret_lines = [ ]
    
    while True:
        p.poll()

        line = p.stdout.readline()
        line_err = p.stderr.readline()
        
        if not line and not line_err and not p.returncode is None:
            break
        elif line:
            if verbose: print line,
                
            ret_lines.append(line)

    return (p.returncode, ret_lines)


class CSDPResult:
    """
    Contains information about a CSDP run.

    The following members can be accessed:

    - ``ret_code`` -- the CSDP return code.

    - ``msg`` -- a message corresponding to the return code.

    - ``primal_value`` -- the primal value of the problem extracted
      from the CSDP output; ``None`` if the extraction fails.

    - ``dual_value`` -- the dual value of the problem extracted from
      the CSDP output; ``None`` if the extraction fails.

    - ``csdp_output`` -- the output from CSDP.
    """
    def __init__(self, ret_code, msg, primal_value, dual_value, csdp_output):
        self.ret_code = ret_code
        self.msg = msg
        self.primal_value = primal_value
        self.dual_value = dual_value
        self.csdp_output = csdp_output
        

class CSDPSolver:
    """
    A wrapper for CSDP.

    This wrapper uses files to communicate with CSDP; a future
    implementation could use the CSDP library instead.
    """
    def __init__(self, csdp = 'csdp'):
        """
        __init__(self, csdp = 'csdp')

        Create a new CSDP solver.

        INPUT:

        - ``csdp`` -- shell command to execute CSDP. The default value
          is ``csdp``, which works if CSDP's directory is in the PATH.

        EXCEPTIONS:

        - ``ValueError`` -- if the given command for CSDP cannot be called.
        """
        try:
            subprocess.check_output(csdp, stderr = subprocess.STDOUT)
        except subprocess.CalledProcessError:
            pass
        except OSError:
            raise ValueError('Cannot execute CSDP')

        self._csdp = csdp

    def solve(self, problem, solution = None, prob_file = None, sol_file = None, verbose = False):
        """
        solve(self, problem, solution = None, prob_file = None, sol_file = None, verbose = False)

        Run CSDP to solve a problem and (optionally) read the solution.

        INPUT:

        - ``problem`` -- an instance of :class:`SDPProblem` containing
          the problem to be solved.

        - ``prob_file`` -- name of file to which the problem will be
          saved. If ``None``, then a temporary file is created (and
          then deleted).

        - ``sol_file`` -- name of file to which the solution will be
          written. If ``None`` then a temporary file is created (and
          then deleted).

        - ``verbose`` -- if ``True``, then the output of CSDP is shown
          as it is generated.

        OUTPUT:

        - ``solution`` -- an instance of :class:`SDPSolution` that
          will contain the solution of the problem. Only the matrices
          corresponding to variables contained in the solution are
          read. If this argument is ``None``, then no solution is
          read.

        The function returns an instance of :class:`CSDPResult`
        containing information about the execution of CSDP.
        """
        if not isinstance(problem, SDPProblem):
            raise ValueError('Problem must be instance of SDPProblem')

        if not solution is None and not isinstance(solution, SDPSolution):
            raise ValueError('Solution must be an instance of SDPSolution')

        # Creates temporary files as needed
        del_prob = False
        del_sol = False
        
        if prob_file is None:
            del_prob = True
            handle, prob_file = tempfile.mkstemp()
            os.close(handle)

        if not solution is None and sol_file is None:
            del_sol = True
            handle, sol_file = tempfile.mkstemp()
            os.close(handle)

        try:
            # Writes problem to file
            problem.write_sdpa(prob_file)
            
            # Runs CSDP
            if solution is None:
                ret_code, output = run_cmd([ self._csdp, prob_file ], verbose)
            else:
                ret_code, output = run_cmd([ self._csdp, prob_file, sol_file ], verbose)

            # Fills the solution
            if not solution is None:
                read_csdp_solution(sol_file, problem, solution)
        finally:
            if del_prob: os.remove(prob_file)
            if del_sol: os.remove(sol_file)

        # Reads CSDP output to get primal and dual values
        primal_value = None
        dual_value = None
        
        for line in output:
            if re.match('Primal objective value:', line):
                fields = re.split(':', line)

                try:
                    primal_value = RealField(SDPEnvironment.gmp_prec)(fields[1].strip())
                except:
                    pass
            elif re.match('Dual objective value:', line):
                fields = re.split(':', line)

                try:
                    dual_value = RealField(SDPEnvironment.gmp_prec)(fields[1].strip())
                except:
                    pass

        # We decide for a return message
        if ret_code == 0:    msg = 'SDP solved'
        elif ret_code == 1:  msg = 'Problem is primal infeasible'
        elif ret_code == 2:  msg = 'Problem is dual infeasible'
        elif ret_code == 3:  msg = 'Solved with reduced accuracy'
        elif ret_code == 4:  msg = 'Maximum iterations reached'
        elif ret_code == 5:  msg = 'Stuck at edge of primal feasibility'
        elif ret_code == 6:  msg = 'Stuck at edge of dual infeasibility'
        elif ret_code == 7:  msg = 'Lack of progress'
        elif ret_code == 8:  msg = 'X, Z, or O was singular'
        elif ret_code == 9:  msg = 'Detected Nan or Inf values'
        elif ret_code == 11: msg = 'File error'
        else:                msg = 'Unkown error'

        return CSDPResult(ret_code, msg, primal_value, dual_value, ''.join(output))

# Local variables:
# mode: python
# End:

