#
# SparseMonomialBasis and SparseBasisChanger
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

cdef class SDPSparseMonomialBasis_wrapper:
    """
    Wrapper for SparseMonomialBasis from SDPSL.
    """
    cdef SDPSLSparseMonomialBasis _basis

    def __cinit__(self, C = None):
        cdef SDPPolynomialConstraint_wrapper C_wrapper
        
        if C is None:
            self._basis = create_s_monomial_basis()
        else:
            C_wrapper = C._constraint

            try:
                sig_on()
                self._basis = create_s_monomial_basis_cons(C_wrapper._constraint)
            finally:
                sig_off()

    def __dealloc__(self):
        destroy_s_monomial_basis(self._basis)

    def add_monomials(self, p):
        cdef SDPPolynomial_wrapper p_wrapper = p._p
        s_monomial_basis_add(self._basis, p_wrapper._p)


class SDPSparseMonomialBasis:
    """
    An SDPSL SparseMonomialBasis.

    At the moment this class is only used internally.
    """   
    def __init__(self, initial_monomials):
        """
        Initializes a sparse monomial basis.

        INPUT:

        - ``initial_monomials`` -- a list of monomials, which for us
          are polynomials having exactly one monomial. The
          coefficients are of course ignored. Alternatively, an
          instance of SDPPolynomialConstraint can be given, and then
          all unique monomials in the left-hand side and right-hand
          side of the constraint are added to the basis.
        """
        if isinstance(initial_monomials, SDPPolynomialConstraint):
            self._basis = SDPSparseMonomialBasis_wrapper(initial_monomials)
        else:
            exponents = set()
            PR = PolynomialRing(RealField(53), 'x')

            new_monomials = [ ]
            for mon in initial_monomials:
                try:
                    mon = make_real_poly(mon)
                except:
                    raise ValueError('Element in list cannot be converted to a polynomial')

                if len(mon.exponents()) != 1:
                    raise ValueError('Element in list not a monomial')

                monomial = mon.exponents()[0]
                
                try:
                    foo = iter(monomial)
                except:
                    monomial = (monomial,)
                
                if monomial in exponents:
                    raise ValueError('Duplicate monomial in list')

                exponents.add(monomial)
                new_monomials.append(mon)

            p = sum(new_monomials)
            self._basis = SDPSparseMonomialBasis_wrapper()
            self._basis.add_monomials(SDPPolynomial(p))


cdef class SDPSparseBasisChanger_wrapper:
    """
    Wrapper for SparseBasisChanger from SDPSL.
    """   
    cdef SDPSLSparseBasisChanger _changer

    def __cinit__(self, monomial_basis, poly_basis):
        cdef int n, error 
        cdef SDPSLPoly *tmp_poly
        cdef SDPPolynomial_wrapper p_wrapper
        cdef SDPSparseMonomialBasis_wrapper mon_wrapper

        n = len(poly_basis)
        
        try:
            converted_polys = [ SDPPolynomial(p) for p in poly_basis ]
        except:
            raise ValueError('Basis element is not a real polynomial')

        sdpsl_mon_basis = SDPSparseMonomialBasis(monomial_basis)

        try:
            sig_on()
        
            tmp_poly = <SDPSLPoly *> malloc(n * cython.sizeof(SDPSLPoly))

            if tmp_poly is NULL:
                raise MemoryError('Failed to allocate memory for tmp_poly')
            
            for k from 0 <= k < n:
                p_wrapper = converted_polys[k]._p
                tmp_poly[k] = p_wrapper._p

            mon_wrapper = sdpsl_mon_basis._basis
            self._changer = create_s_basis_changer(mon_wrapper._basis, n, tmp_poly, 
                                                   cython.address(error))
        finally:
            free(tmp_poly)
            sig_off()

        if self._changer is NULL:
            if error == INCOMPATIBLE_BASES_ERROR:
                raise ValueError('Bases given do not span same subspace')
            elif error == SINGULAR_MATRIX_ERROR:
                raise ValueError('Basis change matrix is singular with tolerance_eps = %s' 
                                 % str(SDPEnvironment.tolerance_eps))
            else:
                raise ValueError('Unknown error')

    def __dealloc__(self):
        destroy_s_basis_changer(self._changer)

        
class SDPSparseBasisChanger:
    """
    An SDPSL sparse basis changer.

    Polynomial constraints can be represented in different bases
    (cf. :ref:`otherbases`). This class implements a basis changer
    between a basis of monomials and a basis of polynomials spanning
    the same subspace. It can later be used to add polynomial
    constraints using the polynomial basis specified (see
    :class:`SDPConstraintContainer`).

    EXAMPLES::

        sage: PR = PolynomialRing(RR, 'x')
        sage: x = PR.gen()
        sage: SDPSparseBasisChanger([ x, x^2, x^3 ], [ x + x^2, 2 * x + x^3, x ])
        SDPSL sparse basis changer between:

        Monomial basis   = [x, x^2, x^3]

        Polynomial basis = [x^2 + x, x^3 + 2.00000000000000*x, x]
        
    An exception is raised if the polynomials are linearly dependent::

        sage: SDPSparseBasisChanger([ x, x^2, x^3 ], [ x + x^2, 2 * x + x^3, x - x^2 + x^3 ])
        ValueError: Basis change matrix is singular with tolerance_eps = 1e-15
    """
    def __init__(self, monomial_basis, poly_basis):
        """
        __init__(self, monomial_basis, poly_basis)

        Create a sparse basis changer. 

        INPUT:

        - ``monomial_basis`` -- either a list of distinct monomials,
          which are polynomials or real Laurent polynomials having
          exactly one nonzero coefficient, or a polynomial constraint
          (an instance of :class:`SDPPolynomialConstraint`), in which
          case the list consists of all monomials occurring in
          polynomials in the constraint.

        - ``poly_basis`` -- a list of polynomials spanning the same
          subspace as the monomials in ``monomial_basis``.

        EXCEPTIONS: 

        - ``ValueError`` -- if ``monomial_basis`` and ``poly_basis``
          do not span the same subspace, or if ``poly_basis`` is not a
          basis.

        .. note::

           Environment parameter ``tolerance_eps`` (see
           :ref:`sdpenvironment`) is used to determine whether the
           basis change matrix is singular, that is, whether
           ``poly_basis`` is really a basis or is linearly
           dependent. 

           This means that numerical issues may cause SDPSL to decide
           that a basis is not a basis. In these cases, the value of
           ``tolerance_eps`` should be made smaller to fix the
           problem, but in any case this should be a warning that
           numerical issues are happening, and that maybe the
           precision of the computations must be increased by raising
           the GMP precision (see :func:`sdpsl_set_gmp_prec`).
        """
        self._monomial_basis = copy(monomial_basis)
        self._poly_basis = [ make_real_poly(p) for p in poly_basis ]
        self._changer = SDPSparseBasisChanger_wrapper(monomial_basis, self._poly_basis)

    def monomial_basis(self):
        """
        monomial_basis(self)
        
        Return the monomial basis.
        """
        return list(self._monomial_basis)

    def polynomial_basis(self):
        """
        polynomial_basis(self)

        Return polynomial basis.
        """
        return list(self._poly_basis)
        
    def __repr__(self):
        if len(self._monomial_basis) < 10:
            return str(self)

        return 'SDPSL sparse basis changer for basis of size %d' % len(self._monomial_basis)

    def __str__(self):
        return ('SDPSL sparse basis changer between:\n\n'
                'Monomial basis   = %s\n\n'
                'Polynomial basis = %s') % (str(self._monomial_basis), str(self._poly_basis))


# Local variables:
# mode: python
# End:
