#
# SDPSL SDPLinearCombination and SDPLinearConstraint
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

#from sage.matrix.matrix import is_Matrix

#from sage.rings.real_mpfr import is_RealField, is_RealNumber
#from sage.rings.real_mpfr cimport RealNumber, RealField, RealField_class

class SDPLinearCombination:
    """
    An SDPSL linear combination.

    An SDPSL linear combination represents the left-hand side of a
    linear constraint or the objective function of a problem; see
    :ref:`sdpslsdps`.

    A linear combination is actually a map assigning numerical
    matrices, or pairs of matrices and masks, to variables. Such a map
    may be built using an addition-multiplication notation::
    
        sage: A = random_matrix(RR, 10)
        sage: B = SDPMatrix(random_matrix(RR, 15))
        sage: X = SDPVariable(10)
        sage: Y = SDPVariable(15)
        sage: A * X + B * Y
        SDPSL linear combination with 2 term(s)

    Here, ``A * X`` gives an instance of
    :class:`SDPLinearCombination`. Addition of linear combinations
    means that the internal dictionaries are merged; matrices are NOT
    added. If two linear combinations being added share a variable,
    then an exception is raised. Note from the example above that both
    SAGE matrices over ``RealField`` and SDPSL's :class:`SDPMatrix`
    can be used. 

    Masks can also be specified::

        sage: A = random_matrix(RR, 10)
        sage: mask = SDPMatrixMask(10)
        sage: X = SDPVariable(10)
        sage: C = (A, mask) * X
        SDPSL linear combination with 1 term(s)
    
    Linear combinations can also be manipulated using the ``[]``
    operator. Above, instead of ``(A, mask) * X`` we could have
    written::

        sage: C = SDPLinearCombination()
        sage: C[X] = (A, mask)

    Actually, everything that is not a SAGE matrix or an SDPSL matrix
    is given to the SAGE matrix constructor in an attempt to transform
    it to a matrix. So you can also write things like::

        sage: X = SDPVariable(2)
        sage: [ [ 1, 2 ], [ 3, 4 ] ] * X
        SDPSL linear combination with 1 term(s)
        sage: _[X]
        [1.00000000000000 2.00000000000000]
        [3.00000000000000 4.00000000000000]

    In general, writing::

        sage: X = SDPVariable(2)
        sage: A * X

    where ``A`` is not a matrix, is the same as writing::

        sage: matrix(RF, X.size(), X.size(), A) * X

    where ``RF`` is a ``RealField`` with the default GMP precision
    (see :func:`sdpsl_set_gmp_prec`).

    This behavior is very useful when 1-by-1 variables are concerned::

        sage: X = SDPVariable(1)
        sage: C = 1 * X
        sage: C[X]
        [1.00000000000000]

    Finally, linear combinations can also be added to polynomial
    combinations, resulting in polynomial combinations; see
    :class:`SDPPolynomialCombination` for more information.
    
    .. note:: 

       SAGE matrices are transformed to instances of
       :class:`SDPMatrix` when a linear combination is used (for
       instance by :class:`SDPLinearConstraint`). If you use SDPSL
       matrices from the start, then this conversion is not
       needed. See documentation of :class:`SDPLinearConstraint` for
       more information.

    Linear combinations can be used to create linear constraints if
    one uses the comparison operators ==, >=, and <=. See the
    documentation of :class:`SDPLinearConstraint` for more information.
    """
    def _add_pair(self, p, unique = True):
        try:
            X, foo = p
        except:
            raise ValueError('Invalid pair: %s' % repr(p))

        if not isinstance(X, SDPVariable):
            raise ValueError('Invalid pair: %s' % repr(p))

        if unique and X in self._data:
            raise NotImplementedError('Duplicate variable given')

        if type(foo) is tuple:
            try:
                A, mask = foo
            except:
                A = foo
                mask = None
        else:
            A = foo
            mask = None

        if not mask is None and not isinstance(mask, SDPMatrixMask):
            raise ValueError('Mask must be instance of SDPMatrixMask')

        if not isinstance(A, SDPMatrix):
            if not is_Matrix(A) or not is_RealField(A.base_ring()):
                # Try a conversion
                try:
                    A = sage_matrix(RealField(SDPEnvironment.gmp_prec), X.size(), X.size(), A)
                except:
                    raise ValueError('Could not convert the given data to a real matrix')

                A = SDPMatrix(A)

        if not A.is_square():
            raise ValueError('Matrices must be square')

        if A.nrows() != X.size():
            raise ValueError('Size of matrix differs from size of variable')

        if not mask is None:
            if A.nrows() != mask.nrows() or A.ncols() != mask.ncols():
                raise ValueError('Sizes of matrix and mask must agree')

        self._data[X] = (A, mask)
    
    def __init__(self, initial_list = [ ]):
        """
        __init__(self, initial_list = [ ])
        
        Create a linear combination.

        INPUT:

        - ``initial_list`` -- a list of pairs ``(X, val)``, where
          ``X`` is an instance of :class:`SDPVariable` and ``val`` is
          a matrix or a pair ``(A, mask)`` where ``A`` is a matrix and
          ``mask`` is an instance of :class:`SDPMatrixMask`. If ``A``
          is not a matrix, then it is passed on to the SAGE matrix
          constructor (see examples above).
        """
        self._data = { }

        for p in initial_list:
            self._add_pair(p)

    def evaluate_at_solution(self, solution):
        """
        evaluate_at_solution(self, solution)

        Return the value of the linear combination at a given solution.

        INPUT:

        - ``solution`` -- the solution at which the combination will
          be evaluated. This is an instance of :class:`SDPSolution`.
        """
        if not isinstance(solution, SDPSolution):
            raise ValueError('Solution must be instance of SDPSolution')

        ret = RealNumber(RealField(SDPEnvironment.gmp_prec), 0)
        for X, foo in self._data.items():
            A, mask = foo

            try:
                B = solution[X]
            except KeyError:
                continue

            n = A.nrows()
            
            if mask is None:
                for i from 0 <= i < n:
                    for j from 0 <= j < n:
                        ret += A[i, j] * B[i, j]
            else:
                for i, j in mask.on_list():
                    ret += A[i, j] * B[i, j]

        return ret            
            
    def __iadd__(self, other):
        """
        __iadd__(self, other)

        Join ``other`` to self, where other is another linear combination.
        """
        if isinstance(other, SDPLinearCombination):
            for key in other._data.keys():
                if key in self._data:
                    raise NotImplementedError('Linear combinations share a variable')

            self._data = dict(self._data.items() + other._data.items())

            return self
        else:
            return NotImplemented

    def __add__(self, other):
        if not isinstance(other, SDPLinearCombination):
            return NotImplemented
        
        C = SDPLinearCombination(self._data.items())
        C += other

        return C

    def __radd__(self, other):
        if not isinstance(other, SDPLinearCombination):
            return NotImplemented

        C = SDPLinearCombination(self._data.items())
        C += other

        return C

    def __repr__(self):
        return 'SDPSL linear combination with %d term(s)' % len(self._data)

    def __len__(self):
        """
        __len__(self)
        
        Return number of terms in the linear combination.
        """
        return len(self._data)

    def __contains__(self, var):
        """
        __contains__(self, var)
        
        Return True iff a variable occurs in the combination.
        """
        if not isinstance(var, SDPVariable):
            raise ValueError('Instance of SDPVariable expected')

        return var in self._data
    
    def variables(self):
        """
        variables(self)
        
        Return a list with the variables that occur in the linear
        combination.
        """
        return self._data.keys()

    def __getitem__(self, var):
        """
        __getitem__(self, var)
        
        Return the matrix associated with a variable. If there is
        also a mask associated, then a pair is returned; otherwise
        just the matrix.
        """
        if not isinstance(var, SDPVariable):
            raise ValueError('Key must be an instance of SDPVariable')

        ret = self._data.get(var, None)

        if ret is None:
            raise KeyError('Variable not found')

        if ret[1] is None:
            return ret[0]
        else:
            return ret

    def __setitem__(self, var, val):
        """
        __setitem__(self, var, val)
        
        Associate a matrix to a given variable. A mask can be also
        specified by giving a pair ``(matrix, mask)`` instead of just
        the matrix. If ``matrix`` is not a SAGE matrix over
        ``RealField`` or an :class:`SDPMatrix`, then it is given to
        the SAGE matrix constructor; see the class description for
        examples.
        """
        if not isinstance(var, SDPVariable):
            raise ValueError('Key must be an instance of SDPVariable')

        self._add_pair((var, val), unique = False)

    def __eq__(self, other):
        """
        __eq__(self, other)
        
        Return a linear equality constraint with self as left-hand
        side and ``other`` as the right-hand side.
        """
        if is_RealNumber(other):
            rhs = other
        else:
            try:
                rhs = RealNumber(RealField(SDPEnvironment.gmp_prec), other)
            except:
                raise ValueError('Cannot convert right-hand side')

        return SDPLinearConstraint(self, '==', rhs)

    def __le__(self, other):
        """
        __le__(self, other)
        
        Return a linear <= constraint with self as left-hand side and
        ``other`` as the right-hand side.
        """
        if is_RealNumber(other):
            rhs = other
        else:
            try:
                rhs = RealNumber(RealField(SDPEnvironment.gmp_prec), other)
            except:
                raise ValueError('Cannot convert right-hand side')

        return SDPLinearConstraint(self, '<=', rhs)

    def __ge__(self, other):
        """
        __ge__(self, other)
        
        Return a linear >= constraint with self as left-hand side and
        ``other`` as the right-hand side.
        """
        if is_RealNumber(other):
            rhs = other
        else:
            try:
                rhs = RealNumber(RealField(SDPEnvironment.gmp_prec), other)
            except:
                raise ValueError('Cannot convert right-hand side')

        return SDPLinearConstraint(self, '>=', rhs)        


cdef SDPSLLinearCombination make_linear_combination(C):
    """
    Creates an SDPSL linear combination from an instance of
    SDPLinearCombination.

    SAGE matrices are converted; SDPSL matrices are used as they are.
    """
    cdef SDPSLLinearCombination ret
    cdef SDPVariable_wrapper var_wrapper
    cdef SDPMatrix_wrapper mat_wrapper
    cdef SDPMatrixMask_wrapper mask_wrapper

    ret = create_linear_combination()

    try:
        for var, mat in C._data.items():
            A, mask = mat

            var_wrapper = var._var

            # The matrix might have to be converted...
            if isinstance(A, SDPMatrix):
                mat_wrapper = A._matrix
            else:
                foo = SDPMatrix(A)
                mat_wrapper = foo._matrix

            if mask is None:
                linear_combination_add_matrix(ret, var_wrapper._var, mat_wrapper._matrix)
            else:
                mask_wrapper = mask._mask
                linear_combination_add_matrix_mask(ret, var_wrapper._var, mat_wrapper._matrix,
                                                   mask_wrapper._mask)
    except:
        destroy_linear_combination(ret)
        raise
                
    return ret


cdef class SDPLinearConstraint_wrapper:
    """
    Wrapper for SDPLinearConstraint from SDPSL.
    """
    cdef SDPSLLinearConstraint _constraint
    
    def __cinit__(self, lhs, rel, rhs):
        cdef SDPSLLinearCombination C
        cdef char rel_char
        cdef RealNumber tmp

        C = make_linear_combination(lhs)

        if rel == '==': rel_char = '='
        elif rel == '>=': rel_char = '>'
        else: rel_char = '<'

        tmp = rhs

        self._constraint = create_linear_constraint(C, rel_char, tmp.value)

        destroy_linear_combination(C)

    def __dealloc__(self):
        destroy_linear_constraint(self._constraint)

    def set_rhs(self, rhs):
        cdef RealNumber tmp = rhs
        linear_constraint_set_rhs(self._constraint, tmp.value)


class SDPLinearConstraint:
    """
    A linear constraint in an SDPSL problem.

    Linear constraints contain a linear combination as left-hand side,
    a relation (one of ==, <=, or >=), and a real number as right-hand
    side.

    Constraints can be created using comparison operators with linear
    combinations::

        sage: X = SDPVariable(5)
        sage: A = random_matrix(RR, 5)
        sage: C1 = A * X == 1
        sage: C2 = A * X <= 1
        sage: C3 = A * X >= 1

    The data of the linear combination is copied; changes to the
    original will not affect the constraint. SAGE matrices in the
    linear combination are transformed into SDPSL matrices, and
    changes to the SAGE matrices have no effect on the
    constraint. SDPSL matrices, however, are not converted, and if
    they are changed then the content of the constraint is also
    changed; the same happens with masks.

    Consider for instance the following setup::

        sage: X, Y = SDPVariable(5), SDPVariable(6)
        sage: A = random_matrix(RR, 5)
        sage: mask = SDPMatrixMask(5)
        sage: B = SDPMatrix(random_matrix(RR, 6))
        sage: C = (A, mask) * X + B * Y == 0

    Changes to ``A`` will not reflect on the constraint ``C``, but
    changes to ``mask`` and ``B`` will. For instance::

        sage: mask[0, 1] = True
        sage: B[0,0] = 3

    These effectively change constraint ``C``. This bahavior can be
    used to efficiently generate many constraints that share a same
    structure; one example is given in :ref:`lovasztheta`.
    """
    def __init__(self, lhs, rel, rhs):
        """
        __init__(self, lhs, rel, rhs)
        
        Create a linear constraint.
        
        INPUT:

        - ``lhs`` -- this is the left-hand side, an instance of :class:`SDPLinearCombination`.

        - ``rel`` -- this is the relation, a string that is either '==', '<=' or '>='.

        - ``rhs`` -- this is the right-hand side, a number.
        """
        if not isinstance(lhs, SDPLinearCombination):
            raise ValueError('Left-hand side must be instance of SDPLinearCombination')

        if rel != '==' and rel != '<=' and rel != '>=':
            raise ValueError('Unknown relation')

        new_rhs = None
        if not is_RealNumber(rhs):
            try:
                new_rhs = RealNumber(RealField(SDPEnvironment.gmp_prec), rhs)
            except:
                raise TypeError('Cannot convert right-hand side to a RealNumber')
        elif rhs.base_ring().prec() < SDPEnvironment.gmp_prec:
            new_rhs = RealNumber(RealField(SDPEnvironment.gmp_prec), rhs)

        if new_rhs is None:
            self._constraint = SDPLinearConstraint_wrapper(lhs, rel, rhs)
        else:
            self._constraint = SDPLinearConstraint_wrapper(lhs, rel, new_rhs)

        self._variables = lhs.variables()

    def variables(self):
        """
        variables(self)
        
        Return a list of variables used in the constraint.
        """
        return list(self._variables)

    def set_rhs(self, rhs):
        """
        Set the right-hand side of the constraint to ``rhs``.
        """
        new_rhs = None
        if not is_RealNumber(rhs):
            try:
                new_rhs = RealNumber(RealField(SDPEnvironment.gmp_prec), rhs)
            except:
                raise TypeError('Cannot convert right-hand side to a RealNumber')
        elif rhs.base_ring().prec() < SDPEnvironment.gmp_prec:
            new_rhs = RealNumber(RealField(SDPEnvironment.gmp_prec), rhs)

        if new_rhs is None:
            self._constraint.set_rhs(rhs)
        else:
            self._constraint.set_rhs(new_rhs)
    
    def __repr__(self):
        return 'SDPSL linear constraint with %d term(s)' % len(self._variables)
    

# Local variables:
# mode: python
# End:
