#
# SDPPolynomialCombination and SDPPolynomialConstraint
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

class SDPPolynomialCombination:
    """
    An SDPSL polynomial combination.

    This is a map associating matrices of polynomials, or pairs of
    matrices and masks, to variables. They appear as the left-hand
    side of polynomial constraints; see :ref:`sdpslsdps`.

    Polynomial combinations can be created using an
    addition-multiplication notation::
    
        sage: PR = PolynomialRing(RR, 'x')
        sage: A = random_matrix(PR, 10)
        sage: B = SDPPolyMatrix(random_matrix(PR, 15))
        sage: X = SDPVariable(10)
        sage: Y = SDPVariable(15)
        sage: A * X + B * Y
        SDPSL polynomial combination with 2 term(s)

    Here, ``A * X`` gives an instance of
    :class:`SDPPolynomialCombination`. Addition of polynomial
    combinations means that the internal dictionaries are merged;
    matrices are NOT added. If two polynomial combinations being added
    share a variable, then an exception is raised. Note from the
    example above that both SAGE matrices over rings of polynomials
    (over the ``RealField``) and SDPSL's :class:`SDPPolyMatrix` can be
    used.

    Masks can also be specified::

        sage: PR = PolynomialRing(RR, 'x')
        sage: A = random_matrix(PR, 10)
        sage: mask = SDPMatrixMask(10)
        sage: X = SDPVariable(10)
        sage: C = (A, mask) * X
        SDPSL polynomial combination with 1 term(s)
    
    Polynomial combinations can also be manipulated using the []
    operator. Above, instead of ``(A, mask) * X`` we could have
    written::

        sage: C = SDPPolynomialCombination()
        sage: C[X] = (A, mask)

    Unlike :class:`SDPLinearCombination`, data given as the matrix is
    never passed to SAGE's matrix constructor. In any case, numerical
    matrices, that is, SAGE matrices over ``RealField`` and instances
    of :class:`SDPMatrix` can also be given. These matrices are
    transformed into matrices of constant polynomials. This means in
    particular that polynomial combinations can be added (joined) to
    linear combinations, giving as a result a polynomial combination::

        sage: X = SDPVariable(1), Y = SDPVariable(2)
        sage: PR = PolynomialRing(RR, 'x')
        sage: A = random_matrix(PR, 2)
        sage: 1 * X + A * Y
        sage: _[X]
        [1.00000000000000]

    Above, ``1 * X`` is actually a linear combination before the
    addition.
        
    .. note::

        SAGE matrices are transformed to instances of
        :class:`SDPPolyMatrix` when a polynomial combination is used
        (for instance by :class:`SDPPolynomialConstraint`). If you use
        SDPSL matrices from the start, then this conversion is not
        needed. See the documentation of
        :class:`SDPPolynomialConstraint` for more information.

    Polynomial combinations can be used to create linear constraints
    if one uses the comparison operator ==. See the documentation of
    :class:`SDPPolynomialConstraint` for more information.
    """
    def _add_pair(self, p, unique = True):
        try:
            X, foo = p
        except:
            raise TypeError('Cannot add "%s" to the polynomial combination' % repr(p))

        if not isinstance(X, SDPVariable):
            raise ValueError('Variable must be an instance of SDPVariable')

        if unique and X in self._data:
            raise NotImplementedError('Duplicate variable in initial_list')

        if type(foo) is tuple:
            try:
                A, mask = foo
            except:
                A = foo
                mask = None
        else:
            A = foo
            mask = None

        if not mask is None and not isinstance(mask, SDPMatrixMask):
            raise ValueError('Mask must be instance of SDPMatrixMask')

        if not is_poly_matrix(A):
            if isinstance(A, SDPMatrix) or is_Matrix(A):
                # We try a conversion; FIXME: this is inneficient code!
                if isinstance(A, SDPMatrix):
                    A = A.sage_matrix()

                # FIXME: is this the desired behavior?
                ring = A.base_ring()
                if not is_RealField(ring):
                    ring = RealField(SDPEnvironment.gmp_prec)

                try:
                    PR = PolynomialRing(ring, 'x')
                    A = sage_matrix(PR, A)
                except:                
                    raise ValueError('Could not convert matrix to a matrix of real polynomials')

                A = SDPPolyMatrix(A)
            else:
                raise ValueError('Could not convert matrix to a matrix of polynomials')

        if not A.is_square():
            raise ValueError('Matrix must be square')

        if A.nrows() != X.size():
            raise ValueError('Size of matrix differs from size of variable')

        if not mask is None:
            if A.nrows() != mask.nrows() or A.ncols() != mask.ncols():
                raise ValueError('Sizes of matrix and mask must agree')

        self._data[X] = (A, mask)
    
    def __init__(self, initial_list = [ ]):
        """
        __init__(self, initial_list = [ ])

        Create a polynomial combination.

        INPUT:

        - ``initial_list`` -- a list of pairs ``(X, val)``, where
          ``X`` is an instance of :class:`SDPVariable` and val is a
          matrix or a pair ``(A, mask)`` where ``A`` is a matrix and
          ``mask`` is a mask of the same size. Matrices can be either
          instances of :class:`SDPPolyMatrix` or SAGE matrices of
          polynomials with entries in rings of polynomials or Laurent
          polynomials over ``RealField``.
        """
        self._data = { }

        for p in initial_list:
            self._add_pair(p)
            
    def __iadd__(self, other):
        """
        __iadd__(self, other)

        Join ``other`` to self, where ``other`` is another polynomial combination.
        """
        if isinstance(other, SDPPolynomialCombination):
            for key in other._data.keys():
                if key in self._data:
                    raise NotImplementedError('Polynomial combinations share a variable')

            self._data = dict(self._data.items() + other._data.items())

            return self
        elif isinstance(other, SDPLinearCombination):
            for key in other._data.keys():
                if key in self._data:
                    raise NotImplementedError('Polynomial combinations share a variable')

            for p in other._data.items():
                self._add_pair(p)

            return self
        else:
            return NotImplemented

    def __add__(self, other):
        if not isinstance(other, SDPLinearCombination) \
            and not isinstance(other, SDPPolynomialCombination):
            return NotImplemented
        
        C = SDPPolynomialCombination(self._data.items())
        C += other

        return C

    def __radd__(self, other):
        if not isinstance(other, SDPLinearCombination) \
            and not isinstance(other, SDPPolynomialCombination):
            return NotImplemented

        C = SDPPolynomialCombination(self._data.items())
        C += other

        return C

    def __repr__(self):
        return 'SDPSL polynomial combination with %d term(s)' % len(self._data)

    def __len__(self):
        """
        __len__(self)

        Return number of terms in the polynomial combination.
        """
        return len(self._data)

    def __contains__(self, var):
        """
        __contains__(self, var)

        Return True iff a ``var`` occurs in the combination.
        """
        if not isinstance(var, SDPVariable):
            raise ValueError('Instance of SDPVariable expected')

        return var in self._data
    
    def variables(self):
        """
        variables(self)

        Return a list with the variables that occur in the linear
        combination.
        """
        return self._data.keys()

    def __getitem__(self, var):
        """
        __getitem__(self, var)

        Return the matrix associated with a variable ``var``. If
        there is also a mask associated, then a pair is returned;
        otherwise just the matrix.
        """
        if not isinstance(var, SDPVariable):
            raise ValueError('Key must be an instance of SDPVariable')

        ret = self._data.get(var, None)

        if ret is None:
            raise KeyError('Variable not found')

        if ret[1] is None:
            return ret[0]
        else:
            return ret

    def __setitem__(self, var, val):
        """
        __setitem__(self, var, val)

        Associate a matrix to a given variable ``var``. A mask can be
        also specified by giving a pair ``(matrix, mask)`` instead of
        just the matrix.
        """
        if not isinstance(var, SDPVariable):
            raise ValueError('Key must be an instance of SDPVariable')

        self._add_pair((var, val), unique = False)

    def __eq__(self, other):
        """
        __eq__(self, other)

        Return a polynomial constraint having self as the left-hand
        side and ``other`` as the right-hand side.
        """
        return SDPPolynomialConstraint(self, other)


cdef SDPSLPolynomialCombination make_polynomial_combination(C):
    """
    Creates an SDPSLPolynomialCombination from an instance of
    SDPPolynomialCombination.

    SAGE matrices are converted to SDPSL matrices in the process.
    """
    cdef SDPSLPolynomialCombination ret
    cdef SDPVariable_wrapper var_wrapper
    cdef SDPPolyMatrix_wrapper mat_wrapper
    cdef SDPMatrixMask_wrapper mask_wrapper

    ret = create_poly_combination()

    try:
        for var, mat in C._data.items():
            A, mask = mat

            var_wrapper = var._var

            # The matrix might have to be converted...
            if isinstance(A, SDPPolyMatrix):
                mat_wrapper = A._matrix
            else:
                foo = SDPPolyMatrix(A)
                mat_wrapper = foo._matrix

            if mask is None:
                poly_combination_add_matrix(ret, var_wrapper._var, mat_wrapper._matrix)
            else:
                mask_wrapper = mask._mask
                poly_combination_add_matrix_mask(ret, var_wrapper._var, mat_wrapper._matrix,
                                                 mask_wrapper._mask)
    except:
        destroy_poly_combination(ret)
        raise

    return ret

    
cdef class SDPPolynomialConstraint_wrapper:
    """
    Wrapper for SDPPolynomialConstraint from SDPSL.
    """
    cdef SDPSLPolynomialConstraint _constraint
    
    def __cinit__(self, lhs, rhs):
        cdef SDPSLPolynomialCombination C
        cdef SDPPolynomial_wrapper rhs_wrapper = rhs._p
        
        C = make_polynomial_combination(lhs)

        self._constraint = create_poly_constraint(C, rhs_wrapper._p)

        destroy_poly_combination(C)

    def __dealloc__(self):
        destroy_poly_constraint(self._constraint)


class SDPPolynomialConstraint:
    """
    A polynomial constraint in an SDPSL problem.

    Polynomial constraints contain a polynomial combination as
    left-hand side and a real polynomial or Laurent polynomial as the
    right-hand side.

    Constraints can be created using comparison operator ``==`` with
    polynomial combinations::

        sage: PR = PolynomialRing(RR, 'x')
        sage: X = SDPVariable(5)
        sage: A = random_matrix(PR, 5)
        sage: A * X == 1
        SDPSL polynomial constraint with 1 term(s)

    If the right-hand side is not a real polynomial or Laurent
    polynomial, like above, then it is attempted to convert it to a
    real polynomial.
        
    The data of the polynomial combination is copied; changes to the
    original will not affect the constraint. SAGE matrices in the
    polynomial combination are transformed into SDPSL matrices, and
    changes to the SAGE matrices have no effect on the
    constraint. SDPSL matrices, however, are not converted, and if
    they are changed then the content of the constraint is also
    changed; the same happens with masks. This behavior is analogous
    to the behavior of :class:`SDPLinearConstraint`.
    """

    def __init__(self, lhs, rhs):
        """
        __init__(self, lhs, rhs)

        Create a linear constraint.
        
        INPUT:

        - ``lhs`` -- this is the left-hand side, an instance of
          SDPPolynomialCombination.

        - ``rhs`` -- this is the right-hand side, a real polynomial or
          a Laurent polynomial. If this is not the case, then a
          conversion is attempted.
        """
        if not isinstance(lhs, SDPPolynomialCombination):
            raise ValueError('Left-hand side must be instance of SDPPolynomialCombination')

        new_rhs = make_real_poly(rhs)

        self._constraint = SDPPolynomialConstraint_wrapper(lhs, SDPPolynomial(new_rhs))
        self._variables = lhs.variables()

    def variables(self):
        """
        variables(self)

        Return a list of variables used in the constraint.
        """
        return list(self._variables)
        
    def __repr__(self):
        return 'SDPSL polynomial constraint with %d term(s)' % len(self._variables)
    

# Local variables:
# mode: python
# End:
