#
# SDPSolution
#

# Copyright (c) 2014 Fernando M. de Oliveira Filho
# <http://www.ime.usp.br/~fmario>
# 
# This file is part of SDPSL.
# 
# SDPSL is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
# 
# SDPSL is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# You should have received a copy of the GNU General Public License
# along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.

cdef class SDPMatrixSolution_wrapper:
    """
    Wrapper for SDPSolution from SDPSL.
    """
    cdef SDPSLSolution _solmap

    def __cinit__(self, sol):
        """
        Makes an SDPSL solution map from a dictionary associating
        instances of SDPMatrix to instances of SDPVariable.
        """
        cdef SDPVariable_wrapper var_wrapper
        cdef SDPMatrix_wrapper mat_wrapper
        
        self._solmap = create_solution()
    
        for X, A in sol.items():
            var_wrapper = X._var
            mat_wrapper = A._matrix
            
            solution_add_matrix(self._solmap, var_wrapper._var,
                                mat_wrapper._matrix)

    def __dealloc__(self):
        destroy_solution(self._solmap)

        
class SDPMatrixSolution:
    """
    A solution containing only SDPMatrix instances.
    """
    def __init__(self, sol):
        """
        Creates a new solution.

        INPUT:

        - ``sol`` -- is a dictionary associating matrices to
        variables. The instances of SDPMatrix are kept; SAGE matrices
        are transformed into SDPMatrix.  
        """
        self._data = { }
        
        for X, A in sol.items():
            if isinstance(A, SDPMatrix):
                self._data[X] = A
            else:
                self._data[X] = SDPMatrix(A)

        self._solmap = SDPMatrixSolution_wrapper(self._data)        


class SDPSolution:
    """
    A solution of a semidefinite programming problem.

    This class basically keeps a map that associates numerical
    matrices to variables. Both instances of SDPMatrix and SAGE real
    number matrices can be used. If SAGE matrices are used, then
    functions that process solutions have to transform them to
    instances of SDPMatrix.

    EXAMPLES:

    Matrices can be associated to variables with the ``[]`` operator::

        sage: X, Y = SDPVariable(2), SDPVariable(3)
        sage: S = SDPSolution()
        sage: S[X] = random_matrix(RR, 2)
        sage: S[Y] = SDPMatrix(3)
        sage: S
        SDPSL solution map with 2 variable(s)

    Matrices can be created in the solution given a list of variables,
    either through the constructor or through the method
    ``create_matrix``::

        sage: S = SDPSolution([ X, Y ])

    This creates matrices of the sizes of ``X`` and ``Y`` and
    associates the matrices to these variables. It is equivalent to::

        sage: S = SDPSolution()
        sage: S.create_matrix([ X, Y ])
    """    
    def __init__(self, initial_vars = [ ]):
        """
        __init__(self, initial_vars = [ ])
        
        Create a solution map.

        INPUT:

        - ``initial_vars`` -- a list of variables or pairs ``(X, A)``
          where ``X`` is a variable and ``A`` is a numerical matrix
          (either :class:`SDPMatrix` or a SAGE matrix of real
          numbers). When a pair ``(X, A)`` is given, matrix ``A`` is
          associated with ``X``; when just a variable is given, an
          instance of :class:`SDPMatrix` is created and associated
          with it.
        """
        self._data = { }
        
        for foo in initial_vars:
            try:
                X, A = foo
            except:
                X = foo
                A = None

            if not isinstance(X, SDPVariable):
                raise ValueError('List must contain variables or pairs (variable, matrix)')

            if A is None:
                self._data[X] = SDPMatrix(X.size())
            else:
                if not isinstance(A, SDPMatrix):
                    if not is_Matrix(A) or not is_RealField(A.base_ring()):
                        raise ValueError('SAGE matrix over RealField or SDPMatrix expected')

                if not A.is_square():
                    raise ValueError('Matrix must be square')

                if A.nrows() != X.size():
                    raise ValueError('Size of matrix has to equal size of variable')
                    
                self._data[X] = A

    def create_matrix(self, var_list):
        """
        create_matrix(self, var_list)
        
        Create matrices and associates them with the given variables.

        For each variable in ``var_list``, this function creates an
        instance of :class:`SDPMatrix` of the appropriate size and
        associates it with the variable.
        """
        for X in var_list:
            if not isinstance(X, SDPVariable):
                raise ValueError('List must contain only variables')

            self._data[X] = SDPMatrix(X.size())

    def variables(self):
        """
        variables(self)
        
        Return a list of variables occurring in the solution.
        """
        return self._data.keys()
        
    def _get_sdp_matrix_map(self):
        return SDPMatrixSolution(self._data)

    def _set_from_matrix_map(self, sol):
        for X, A in sol._data.items():
            B = self._data[X]
            
            if not isinstance(B, SDPMatrix):
                A.fill_sage_matrix(B)
        
    def __getitem__(self, X):
        """
        __getitem__(self, X)
        
        Return the matrix associated with a given variable.
        """
        if not isinstance(X, SDPVariable):
            raise ValueError('Key must be a variable')

        ret = self._data.get(X, None)

        if ret is None:
            raise KeyError('Variable not in solution')
        else:
            return ret

    def __setitem__(self, X, A):
        """
        __setitem__(self, X, A)
        
        Set the matrix associated with a given variable.
        """
        if not isinstance(X, SDPVariable):
            raise ValueError('Key must be a variable')

        if not isinstance(A, SDPMatrix):
            if not is_Matrix(A) or not is_RealField(A.base_ring()):
                raise ValueError('SAGE matrix over RealField or SDPMatrix expected')

        if not A.is_square():
            raise ValueError('Matrix must be square')

        if A.nrows() != X.size():
            raise ValueError('Size of matrix has to equal size of variable')

        self._data[X] = A

    def __contains__(self, X):
        """
        __contains__(self, X)
        
        Return True iff variable ``X`` is in the solution.
        """
        if not isinstance(X, SDPVariable):
            raise ValueError('Key must be variable')

        return X in self._data

    def __len__(self):
        """
        __len__(self)

        Return the number of variables in the solution.
        """
        return len(self._data)
    
    def __repr__(self):
        return 'SDPSL solution map with %d variable(s)' % len(self._data)

        
# Local variables:
# mode: python
# End:

