/*
  analyzer.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <set>
#include <sdpsl/analyzer.h>
#include <sdpsl/conv.h>
#include <sdpsl/qrdec.h>

#include <iostream>

using namespace std;

//
// LAPACK prototypes
//
extern "C" int dsyev_(char *jobz, char *uplo, int *N, double *A, int *lda, 
                      double *w, double *work, int *lwork, int *info);

extern "C" void dgelsy_(int *M, int *N, int *nrhs, double *A, int *lda, 
			double *b, int *ldb, int *jpvt, double *rcond, 
			int *rank, double *work, int *lwork, int *info);

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//
// Private functions
//
template <class T>
static double do_double_eigenvalues(const Matrix<T>& A, const MatrixMask *mask, 
				    std::vector<double>& eigenvalues);

template <class T>
static T do_feasibility_measures(const std::list<const SDPConstraintContainer *>& containers,
				 const SDPSolution<T>& solution,
				 std::vector<T>& measures);

template <class T>
static void do_project_solution(const std::list<const SDPConstraintContainer *>& containers,
				const SDPSolution<T>& solution, SDPSolution<double>& result,
				const SDPEnvironment& env);

template <class T>
static void do_project_solution(const std::list<const SDPConstraintContainer *>& containers,
				const SDPSolution<T>& solution, SDPSolution<mpf_class>& result,
				const SDPEnvironment& env);

//
// Static functions
//

/*
  do_double_eigenvalues

  This function uses LAPACK to compute the eigenvalues of a symmetric
  matrix.
*/
template <class T>
double do_double_eigenvalues(const Matrix<T>& A, const MatrixMask *mask, 
			     std::vector<double>& eigenvalues)
{
  if (!A.is_square())
    throw ValueError(ERROR_MSG("Matrix is not square"));

  if (mask)
    if (A.num_rows() != mask->num_rows() 
        || A.num_cols() != mask->num_cols())
      throw ValueError(ERROR_MSG("Matrix and mask differ in size"));

  /* We create a new matrix of doubles, converting the old one. */
  int n = A.num_rows();
  Matrix<double> A_tmp(n, n, 0.0);
  
  if (mask)
    for (auto& p : *mask)
      A_tmp[p.i][p.j] = convert_number<double>(A[p.i][p.j]);
  else
    for (int i = 0; i < n; i++)
      for (int j = 0; j < n; j++)
	A_tmp[i][j] = convert_number<double>(A[i][j]);
  
  /* Data for LAPACK call */
  char jobz = 'N';
  char uplo = 'U';
  int N = n;
  int lda = n;
  int lwork;
  int info;

  if (eigenvalues.size() < N)
    eigenvalues.resize(N);

  /* Query for LWORK */
  lwork = -1;
  double work_query;

  dsyev_(&jobz, &uplo, &N, A_tmp.vector_data(), &lda, eigenvalues.data(),
	 &work_query, &lwork, &info);

  if (info) throw AnalyzerError(ERROR_MSG("LAPACK: Failed to query for LWORK"));

  /* Finds eigenvalues */
  lwork = (int) work_query;
  vector<double> work(lwork);

  dsyev_(&jobz, &uplo, &N, A_tmp.vector_data(), &lda, eigenvalues.data(),
	 work.data(), &lwork, &info);

  if (info) throw AnalyzerError(ERROR_MSG("LAPACK: Call failed"));

  /* Finds minimum eigenvalue to return */
  double min = eigenvalues[0];
  
  for (int i = 1; i < n; i++)
    if (eigenvalues[i] < min) min = eigenvalues[i];

  return min;
}

/*
  do_feasibility_measures

  We use a template to implement the computation of feasibility
  measures, this way avoiding duplicating code. Class gmp_conv plays an
  important role in this.
*/
template <class T>
T do_feasibility_measures(const std::list<const SDPConstraintContainer *>& containers,
			  const SDPSolution<T>& solution,
			  std::vector<T>& measures)
{
  /* 
     We assemble a list with the matrices of the solution 

     WARNING: here we have a potential place where many variables
     would make the code inefficient.
  */
  vector<const Matrix<T> *> sol_matrices(SDPVariable::num_variables(), nullptr);
  vector<const MatrixMask *> sol_masks(SDPVariable::num_variables(), nullptr);

  for (auto& c : solution) {
    sol_matrices[c.first.idx()] = &c.second.matrix();
    sol_masks[c.first.idx()] = c.second.mask();
  }

  /* Then we compute how many constraints will be processed */
  int total_constraints = 0;
  for (auto cc : containers)
    total_constraints += cc->constraints().size();

  /* Then we compute the measures */
  measures.resize(total_constraints);

  int k = 0;
  T max = 0;  // The maximum violation
  for (auto cc : containers) {
    for (auto& c : cc->constraints()) {
      measures[k] = 0;

      /* Here, recall solutions are symmetric */
      for (auto& entry : c.entries)
	if (sol_matrices[entry.var_idx]) {
	  if (!sol_masks[entry.var_idx] || sol_masks[entry.var_idx]->is_on(entry.i, entry.j)) {
	    if (entry.i != entry.j)
	      measures[k] += gmp_conv<T>::convert(2 * (*sol_matrices[entry.var_idx])[entry.i][entry.j] * entry.value);
	    else
	      measures[k] += gmp_conv<T>::convert((*sol_matrices[entry.var_idx])[entry.i][entry.j] * entry.value);
	  }
	}

      if (c.rel == sdp_eq)
	measures[k] = NumberOperations::abs(gmp_conv<T>::convert(c.rhs - measures[k]));
      else if (c.rel == sdp_leq) {
	measures[k] -= convert_number<T>(c.rhs);

	if (measures[k] < 0) measures[k] = 0;
      }
      else if (c.rel == sdp_geq) {
	measures[k] = gmp_conv<T>::convert(c.rhs - measures[k]);

	if (measures[k] < 0) measures[k] = 0;
      }

      if (measures[k] > max) max = measures[k];

      k++;
    }
  }

  return max;
}

/*
  ProjectionSystem

  To project a solution onto a subspace given by some equality
  constraints, we have to build a matrix system related to the given
  constraints and the solution to be projected. This matrix can then
  be fed into LAPACK routines or our own routines to compute the
  projection.

  The key to understanding how the system is built is this: we
  identify a tuple (X_1, ..., X_r) of symmetric matrices, X_j being an
  n_j by n_j matrix, with a vector in R^N, where

  N = sum_{j = 1}^r (n_j * (n_j + 1)) / 2

  is the sum of the number of diagonal and upper-diagonal entries in
  all X_j. Then the system given by the constraints can be naturally
  interpreted as A x = b, with A an m by N matrix, and b in R^m.

  This class can be used to build such a system. A solution to the
  system is a vector x in R^N, and this class can also be used to
  convert such a vector into a tuple of symmetric matrices.

  Template parameters:

  T : type of input solution
  U : type of numbers used in the system
*/
template <class T, class U>
class ProjectionSystem {
  SDPEnvironment env;

  /* Set of variables and their displacements in the system */
  set<SDPVariable> var_set;
  vector<int> displace;  // For each variable, the total size of
			 // variables coming before it in the tuple

  /* The system */
  int m, n;     // Number of rows and columns
  vector<U> A;  // The system matrix
  vector<U> b;  // The right-hand side and output array for LAPACK
  vector<U> z;  // The solution to be projected

public:
  /* 
     Given a list of constraint containers and a solution to be
     projected, builds the system for the projection.

     If lapack is true, then we build a system to be used with an
     LAPACK projection routine. This means that the system matrix is
     given in column-major order, and that the right-hand side is
     different (see intro in SDPSolutionAnalyzer).

     If lapack is false, then the right-hand side is the right-hand
     side of the constraints, and the matrix is given in row-major
     order.

     Throws:

     - ValueError if the system is empty or if it is overdetermined.
  */
  ProjectionSystem(const std::list<const SDPConstraintContainer *>& containers,
		   const SDPSolution<T>& solution, const SDPEnvironment& e,
		   bool lapack)
    : env(e), displace(SDPVariable::num_variables(), -1)
  {
    /* First we determine the size of the system that has to be
       solved. We ignore inequality constraints.
    */

    /* Number of rows */
    m = 0;
    for (auto cc : containers)
      for (auto& c : cc->constraints())
	if (c.rel == sdp_eq) m++;
  
    if (m == 0)
      throw ValueError(ERROR_MSG("Empty system for projection"));

    /* To avoid making too big of a matrix, we will compute the set of
       variables actually used in the constraints and the
       solution. Only these variables will be included in the matrix.
    */
    for (auto cc : containers)
      for (auto& X : cc->variables())
	var_set.insert(X);

    for (auto& c : solution)
      var_set.insert(c.first);

    /* We compute the number of columns, and also take the chance to
       compute a displacement vector.
    */
    {
      int last_size = 0, last_displace = 0;

      for (auto& X : var_set) {
	last_displace = displace[X.idx()] = last_displace + (last_size + 1) * last_size / 2;
	last_size = X.size();
      }

      n = last_displace + (last_size + 1) * last_size / 2;
    }

    if (m > n)
      throw ValueError(ERROR_MSG("System for projection is overdetermined"));

    /* Then we create vectors for solution matrices and masks */
    vector<const Matrix<T> *> sol_matrices(SDPVariable::num_variables(), nullptr);
    vector<const MatrixMask *> sol_masks(SDPVariable::num_variables(), nullptr);

    for (auto& c : solution) {
      sol_matrices[c.first.idx()] = &c.second.matrix();
      sol_masks[c.first.idx()] = c.second.mask();
    }
    
    /* 
       We create and fill the vector that contains the matrix for the
       system, in column-major order if LAPACK is used, and in
       row-major order if our own algorithm is used.

       The right-hand side also differs according to whether LAPACK is
       used or not.
    */
    A.resize(m * n, 0);
    b.resize(n);

    int row = 0;
    for (auto cc : containers)
      for (auto& c : cc->constraints()) {
	if (c.rel != sdp_eq) continue;
      
	b[row] = convert_number<U>(c.rhs);
      
	for (auto& entry : c.entries) {
	  int size = SDPVariable::variable_size(entry.var_idx);
	  int col = displace[entry.var_idx] + (2 * size - entry.i + 1) * entry.i / 2 + entry.j - entry.i;
	  
	  assert(col >= 0 && col < n);

	  int fac = 1;
	  if (entry.i != entry.j) fac = 2;

	  if (lapack) {	    
	    if (sol_matrices[entry.var_idx])
	      if (!sol_masks[entry.var_idx] || sol_masks[entry.var_idx]->is_on(entry.i, entry.j))
		b[row] -= gmp_conv<U>::convert(fac * (*sol_matrices[entry.var_idx])[entry.i][entry.j] * entry.value);

	    A[row + col * m] = gmp_conv<U>::convert(fac * entry.value);
	  }
	  else
	    A[row * n + col] = gmp_conv<U>::convert(fac * entry.value);

	}
    
	row++;
      }

    /* We fill z with the solution given */
    z.resize(n, 0);

    for (auto& c : solution) {
      int size = c.first.size();
      const Matrix<T>& X = c.second.matrix();
      const MatrixMask *mask = c.second.mask();

      if (mask)
	for (auto& p : *mask) {
	  if (p.i > p.j) continue;  // Recall solutions are symmetric

	  int col = displace[c.first.idx()] + (2 * size - p.i + 1) * p.i / 2 + p.j - p.i;

	  assert(col >= 0 && col < n);

	  z[col] = convert_number<U>(X[p.i][p.j]);
	}
      else 
	for (int i = 0; i < size; i++)
	  for (int j = i; j < size; j++) {
	    int col = displace[c.first.idx()] + (2 * size - i + 1) * i / 2 + j - i;

	    assert(col >= 0 && col < n);

	    z[col] = convert_number<U>(X[i][j]);
	  }
    }
  }

  /* Access to the system */
  int num_rows() const { return m; }
  int num_cols() const { return n; }

  vector<U>& get_A() { return A; }
  vector<U>& get_b() { return b; }
  vector<U>& get_z() { return z; }

  /* 
     Assembles an instance of SDPSolution from the vector z. 

     Environment variable output_eps is used to set the mask of a
     solution matrix if a mask is provided.
  */
  template <class Out>
  void make_solution(SDPSolution<Out>& result) const
  {
    /* First we create the matrices that are needed and reset the
       others.
    */
    for (auto& c : result) {
      Matrix<Out>& A = c.second.matrix();
      
      if (var_set.find(c.first) == var_set.end()) {
	if (c.second.mask()) 
	  c.second.mask()->clear();
	else
	  for (int i = 0; i < A.num_rows(); i++)
	    for (int j = 0; j < A.num_cols(); j++)
	      A[i][j] = 0;
      }
    }
    
    for (auto& X : var_set)
      if (!result.contains(X))
        result.create(X);  // We don't create masks.
    
    /* Then we fill the solution matrices */
    double output_eps = env.output_eps();
    
    for (auto& X : var_set) {
      int k = X.idx();
      int size = X.size();
      
      assert(result[X]);
      
      Matrix<Out>& R = result[X]->matrix();
      MatrixMask *mask = result[X]->mask();
      
      for (int i = 0; i < size; i++)
        for (int j = i; j < size; j++) {
	  int col = displace[k] + (2 * size - i + 1) * i / 2 + j - i;
	  
	  assert(col >= 0 && col < n);
	  
	  R[i][j] = R[j][i] = convert_number<Out>(z[col]);
	  
	  if (mask && NumberOperations::abs(z[col]) > output_eps) {
	    mask->turn_on(i, j);
	    mask->turn_on(j, i);
	  }
        }
    }  
  }
};

/*
  do_project_solution - LAPACK version
*/

/*
  Since we use LAPACK to do the projection, we need vectors of
  doubles. But we assemble our vectors using mpf_class when T =
  mpf_class, and this gives us a complication. The following template
  trickery aims at solving this problem. Note we avoid copying the
  vectors if they are already vectors of doubles.
*/
template <class T>
class double_vector { };

template <>
class double_vector<double> {
  vector<double>& v;

public:
  double_vector(vector<double>& vec) : v(vec) { }

  vector<double>& get_vector() { return v; }
};

template <>
class double_vector<mpf_class> {
  vector<double> v;

public:
  double_vector(vector<mpf_class>& vec) 
    : v(vec.size())
  {
    for (int i = 0; i < vec.size(); i++)
      v[i] = vec[i].get_d();
  }

  vector<double>& get_vector() { return v; }
};

/* 
   The projection function uses environment parameter tolerance_eps as
   LAPACK call RCOND parameter, which is used to determine the rank of
   the matrix A.
*/
template <class T>
void do_project_solution(const std::list<const SDPConstraintContainer *>& containers,
			 const SDPSolution<T>& solution, SDPSolution<double>& result,
			 const SDPEnvironment& env)
{
  ProjectionSystem<T, T> S(containers, solution, env, true);

  int m = S.num_rows();
  int n = S.num_cols();

  /* We need vectors of doubles here; see explanation at the beginning of function. */
  double_vector<T> A_tmp(S.get_A());
  double_vector<T> b_tmp(S.get_b());

  /* We run LAPACK */
  int M = m;
  int N = n;
  int nrhs = 1;
  int lda = m;
  int ldb = n;
  vector<int> jpvt(n);
  double rcond = env.tolerance_eps(); 
  int rank;
  double work_query;
  int lwork;
  int info;

  /* Workspace query */
  lwork = -1;
  
  dgelsy_(&M, &N, &nrhs, A_tmp.get_vector().data(), &lda, b_tmp.get_vector().data(), &ldb, jpvt.data(),
  	  &rcond, &rank, &work_query, &lwork, &info);

  if (info || work_query < 1.0)
    throw AnalyzerError(ERROR_MSG("Failure querying LAPACK"));

  lwork = (int) work_query;
  vector<double> work(lwork);

  /* The call */
  dgelsy_(&M, &N, &nrhs, A_tmp.get_vector().data(), &lda, b_tmp.get_vector().data(), &ldb, jpvt.data(),
  	  &rcond, &rank, work.data(), &lwork, &info); 

  if (info) 
    throw AnalyzerError(ERROR_MSG("LAPACK call failed"));

  /* Then we assemble the projected solution */
  vector<T>& z = S.get_z();

  for (int i = 0; i < n; i++) 
    z[i] += b_tmp.get_vector()[i];

  S.make_solution(result);
}

/*
  Projects a solution using QR decomposition with GMP
  arbitrary-precision arithmetic.

  We apply the QR decomposition to the transpose of the system matrix
  A. The system matrix is given in row-major order, so the same vector
  gives the transpose in column-major order, what is expected by our
  column-oriented QR algorithm.

  Parameter tolerance_eps is used to find out the rank of the matrix
  in the QR decomposition, and also to determine whether the system is
  feasible, in case A is rank-defficient.
*/
template <class T>
void do_project_solution(const std::list<const SDPConstraintContainer *>& containers,
			 const SDPSolution<T>& solution, SDPSolution<mpf_class>& result,
			 const SDPEnvironment& env)
{
  /* Assembles the projection system */
  ProjectionSystem<T, mpf_class> S(containers, solution, env, false);

  int m = S.num_rows();
  int n = S.num_cols();
  vector<mpf_class>& b = S.get_b();
  vector<mpf_class>& z = S.get_z();

  /* Does the QR decomposition of A transpose */
  QRDecomposition<mpf_class> QR(S.get_A(), n, m, env.tolerance_eps());

  int r = QR.rank();  // Rank of the matrix A
  mpf_class sfactor = QR.scaling_factor();

  /* We find a solution of the system, solving R_11^T y = P^T b */
  vector<mpf_class> y(n, 0);
  for (int i = 0; i < r; i++) {
    y[i] = b[QR.perm(i)] / sfactor;

    for (int j = 0; j < i; j++)
      y[i] -= QR.R(j, i) * y[j];

    y[i] /= QR.R(i, i);
  }

  /* Checks if the system has indeed a solution */
  for (int i = r; i < m; i++) {
    mpf_class val = 0;

    for (int j = 0; j < r; j++)
      val += QR.R(j, i) * y[j];

    if (NumberOperations::abs(val - b[QR.perm(i)] / sfactor) > env.tolerance_eps())
      throw AnalyzerError(ERROR_MSG("Projection system is infeasible"));
  }

  /* Then we compute v = A^T y^* */
  vector<mpf_class> v = z;

  QR.Q_transpose_mul(v);
  for (int i = 0; i < r; i++) v[i] -= y[i];
  for (int i = r; i < n; i++) v[i] = 0;

  QR.Q_mul(v);

  /* And the projection is... */
  for (int i = 0; i < n; i++) z[i] -= v[i];  

  S.make_solution(result);
}


//
// Methods from SDPSolutionAnalyzer
//

/*
  feasibility_measures
*/
double SDPSolutionAnalyzer::feasibility_measures(const std::initializer_list<const SDPConstraintContainer *>& containers,
						 const SDPSolution<double>& solution,
						 std::vector<double>& measures) const
{
  return do_feasibility_measures(list<const SDPConstraintContainer *>(containers), solution, measures);
}

mpf_class SDPSolutionAnalyzer::feasibility_measures(const std::list<const SDPConstraintContainer *>& containers,
						    const SDPSolution<mpf_class>& solution,
						    std::vector<mpf_class>& measures) const
{
  return do_feasibility_measures(containers, solution, measures);
}

mpf_class SDPSolutionAnalyzer::feasibility_measures(const std::initializer_list<const SDPConstraintContainer *>& containers,
						    const SDPSolution<mpf_class>& solution,
						    std::vector<mpf_class>& measures) const
{
  return do_feasibility_measures(list<const SDPConstraintContainer *>(containers), solution, measures);
}

/*
  Versions of double_eigenvalues
*/
double SDPSolutionAnalyzer::double_eigenvalues(const Matrix<double>& A, std::vector<double>& eigenvalues) const
{
  return do_double_eigenvalues(A, nullptr, eigenvalues);
}

double SDPSolutionAnalyzer::double_eigenvalues(const Matrix<double>& A, const MatrixMask& mask, 
					       std::vector<double>& eigenvalues) const
{
  return do_double_eigenvalues(A, &mask, eigenvalues);
}

double SDPSolutionAnalyzer::double_eigenvalues(const Matrix<mpf_class>& A, std::vector<double>& eigenvalues) const
{
  return do_double_eigenvalues(A, nullptr, eigenvalues);
}

double SDPSolutionAnalyzer::double_eigenvalues(const Matrix<mpf_class>& A, const MatrixMask& mask, 
					       std::vector<double>& eigenvalues) const
{
  return do_double_eigenvalues(A, &mask, eigenvalues);
}

/*
  Versions of project_solution
*/
void SDPSolutionAnalyzer::project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
					   const SDPSolution<double>& solution, SDPSolution<double>& result) const
{
  do_project_solution(list<const SDPConstraintContainer *>(containers), solution, result, env);
}

void SDPSolutionAnalyzer::project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
					   const SDPSolution<mpf_class>& solution, SDPSolution<double>& result) const
{
  do_project_solution(list<const SDPConstraintContainer *>(containers), solution, result, env);
}

void SDPSolutionAnalyzer::project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
					   const SDPSolution<double>& solution, SDPSolution<mpf_class>& result) const
{
  do_project_solution(list<const SDPConstraintContainer *>(containers), solution, result, env);
}

void SDPSolutionAnalyzer::project_solution(const std::list<const SDPConstraintContainer *>& containers,
					   const SDPSolution<mpf_class>& solution, SDPSolution<mpf_class>& result) const
{
  do_project_solution(containers, solution, result, env);
}

void SDPSolutionAnalyzer::project_solution(const std::initializer_list<const SDPConstraintContainer *>& containers,
					   const SDPSolution<mpf_class>& solution, SDPSolution<mpf_class>& result) const
{
  do_project_solution(list<const SDPConstraintContainer *>(containers), solution, result, env);
}

}  // namespace
