/*
  polynomial.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sdpsl/polynomial.h>
#include <sdpsl/polyerror.h>

using namespace std;

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

//
// Static variables
//
const Number Polynomial::zero = 0;

//
// Method definition
//

/*
  Constructors
*/

Polynomial::Polynomial(const Number& val)
{
  if (val != 0)
    monomials[Monomial()] = val;
}

Polynomial::Polynomial(const PolynomialVariable& x)
{
  monomials[Monomial(x)] = 1;
}

Polynomial::Polynomial(const PolynomialVariable::Power& pw)
{
  monomials[Monomial(pw)] = 1;
}

#ifdef __MPF_BUILD__
Polynomial::Polynomial(double val)
{
  if (val != 0)
    monomials[Monomial()] = val;
}
#endif

Polynomial::Polynomial(const Monomial& mon)
{
  monomials[mon] = 1;
}

/*
  add_monomial

  Adds a monomial mon with the given coefficient coeff to the current
  polynomial. If the monomial in question is already contained in the
  polynomial, we verify whether after the addition the resulting
  coefficient becomes zero, in which case the monomial is removed.

  \remark Zero means zero! It is conceivable that numerical error
  would cause an operation that is supposed to give zero as result to
  give a nonzero number, but we do not check for that.
*/
void Polynomial::add_monomial(const Number& coeff, const Monomial& mon)
{
  if (coeff == 0) return;

  auto result = monomials.insert(value_type(mon, coeff));

  if (!result.second) {
    result.first->second += coeff;

    if (result.first->second == 0)
      monomials.erase(result.first);
  }
}

Polynomial& Polynomial::operator+=(const Polynomial& other)
{
  /* We never know if other == *this, so we must take care */
  for (auto it = other.monomials.begin(); it != other.monomials.end(); ) {
    auto result = monomials.insert(*it);

    if (!result.second)
      result.first->second += it->second;

    ++it;

    /* This will never happen if other == *this */
    if (result.first->second == 0)
      monomials.erase(result.first);
  }

  return *this;
}

Polynomial operator+(const Polynomial& p, const Polynomial& q)
{
  Polynomial ret = p;
  ret += q;
  return ret;
}

Polynomial& Polynomial::operator-=(const Polynomial& other)
{
  /* We never know if other == *this, so we must take care */
  for (auto it = other.monomials.begin(); it != other.monomials.end(); ) {
    auto result = monomials.insert(value_type(it->first, Number(-it->second)));

    if (!result.second)
      result.first->second -= it->second;

    ++it;

    if (result.first->second == 0)
      monomials.erase(result.first);
  }

  return *this;
}

Polynomial operator-(const Polynomial& p, const Polynomial& q)
{
  Polynomial ret = p;
  ret -= q;
  return ret;
}

Polynomial& Polynomial::operator*=(const Polynomial& other)
{
  *this = *this * other;

  return *this;
}

Polynomial operator*(const Polynomial& p, const Polynomial& q)
{
  Polynomial ret;

  for (auto& c1 : p.monomials)
    for (auto& c2 : q.monomials)
      ret.add_monomial(c1.second * c2.second, c1.first * c2.first);

  return ret;
}

Polynomial Polynomial::operator-() const
{
  Polynomial ret = *this;

  for (auto& c : ret.monomials)
    c.second *= -1;

  return ret;
}

Polynomial& Polynomial::operator/=(const Number& val)
{
  for (auto& c : monomials)
    c.second /= val;

  return *this;
}

Polynomial operator/(const Polynomial& p, const Number& val)
{
  Polynomial ret = p;
  ret /= val;
  return ret;
}

std::ostream& operator<<(std::ostream& out, const Polynomial& p)
{
  if (p.monomials.empty())
    out << "0";
  else {
    bool first = true;

    for (auto it = p.monomials.crbegin(); it != p.monomials.crend(); ++it) {
      if (!first)
	out << " " << InsideCoefficient(it->second);
      else {
	first = false;
	out << it->second;
      }

      if (!it->first.is_constant())
	out << " * " << it->first;
    }
  }

  return out;
}

Number Polynomial::univariate_eval(const PolynomialVariable& x, const Number& val) const
{
  Number ret = 0;
  Number xpower = 1;
  
  int last_exponent = 0;
  for (auto& c : monomials) {
    if (c.first[x] < last_exponent)
      throw PolyError::ValueError("Polynomial is multivariate or has negative exponents");

    for ( ; last_exponent < c.first[x]; last_exponent++)
      xpower *= val;

    ret += xpower * c.second;
  }
  
  return ret;
}

}  // namespace
