/*
  sdpawriter.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sdpsl/sdpawriter.h>
#include <set>

using namespace std;

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*!
  \brief This function does the job of outputting the problem to a stream.

  At the moment we are still allowing variables to be ignored via the
  ignore_list, but this feature will probably be removed in the
  future.

  \todo Decide whether to allow or remove the ignore_list feature.
*/
void SDPAWriter::write(std::ostream& out) const
{
  /* Computes total number of constraints and inequalities */
  int total_constraints = 0;
  int total_ineqs = 0;

  for (auto& cc : problem.containers()) {
    total_constraints += cc->constraints().size();
    total_ineqs += cc->num_ineqs();
  }

  if (progress)
    progress->set_max(total_constraints);

  /* We make a set with all the variables occurring in the SDP */
  set<SDPVariable> variables;

  for (auto& var : problem.objective_variables())
    variables.insert(var);

  for (auto& cc : problem.containers())
    for (auto& var : cc->variables())
      variables.insert(var);

  if (total_constraints == 0 || variables.size() == 0)
    throw ValueError(ERROR_MSG("Problem has no variables or constraints"));

  /* We compute the block numbers corresponding to variables, given
     the ignore_list. Block numbers are already +1.
  */
  vector<int> block_number(SDPVariable::num_variables(), -1);

  for (auto& var : variables)
    block_number[var.idx()] = 0;  // These are the variables actually used

  for (auto& var : ignore_list)
    block_number[var.idx()] = -1;  // Marks that this variable is ignored

  int num_blocks = 0;

  for (int i = 0; i < block_number.size(); i++)
    if (block_number[i] == 0)
      block_number[i] = ++num_blocks;
  
  int slackb = num_blocks + 1;  // The block for slack variables
  
  /* Writes the header */
  out << total_constraints << endl;
  
  if (total_ineqs)
    out << num_blocks + 1 << endl;
  else
    out << num_blocks << endl;

  /* TODO: This could be improved, using the set above. */
  for (int i = 0; i < block_number.size(); i++)
    if (block_number[i] != -1)
      out << SDPVariable::variable_size(i) << " ";

  if (total_ineqs)
    out << -total_ineqs;

  out << endl;

  /* The right-hand side. We take the opportunity to check that the
     number of inequalities is correct.
  */
  int check_ineqs = 0;
  for (auto& cc : problem.containers())
    for (auto& rc : cc->constraints()) {
      assert(!rc.entries.empty());
      
      if (rc.rel != sdp_eq)
	check_ineqs++;
      
      out << rc.rhs << " ";
    }

  out << endl;

  assert(total_ineqs == check_ineqs);

  /* Writes the constraint entries */
  if (progress) progress->start();

  int cur_constraint = 0;
  int cur_ineq = 0;

  for (auto& cc : problem.containers())
    for (auto& rc : cc->constraints()) {
      cur_constraint++;
      
      for (auto& entry : rc.entries)
	if (block_number[entry.var_idx] != -1)
	  out << cur_constraint << " "
	      << block_number[entry.var_idx] << " "
	      << entry.i + 1 << " " << entry.j + 1 << " "
	      << entry.value << endl;
      
      /* And we deal with a possible inequality */
      if (rc.rel == sdp_leq) {
	cur_ineq++;
	
	out << cur_constraint << " "
	    << slackb << " " << cur_ineq << " " << cur_ineq << " "
	    << 1 << endl;
      }
      else if (rc.rel == sdp_geq) {
	cur_ineq++;
	
	out << cur_constraint << " "
	    << slackb << " " << cur_ineq << " " << cur_ineq << " "
	    << -1 << endl;
      }      
      
      if (progress) ++(*progress);
    }

  assert(cur_constraint == total_constraints);
  assert(cur_ineq == total_ineqs);

  /* Writes the objective function */
  for (auto& entry : problem.objective_entries())
    if (block_number[entry.var_idx] != -1)
      out << 0 << " " << block_number[entry.var_idx] << " " << entry.i + 1 << " "
	  << entry.j + 1 << " " << entry.value << endl;

  if (progress) progress->finish();  
}

}  // namespace
