/*
  sdpcon.cpp

  Copyright (c) 2014 Fernando M. de Oliveira Filho
  <http://www.ime.usp.br/~fmario>
  
  This file is part of SDPSL.
  
  SDPSL is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your
  option) any later version.
  
  SDPSL is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with SDPSL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sdpsl/sdpcon.h>
#include <cassert>
#include <vector>
#include <sdpsl/sparsemon.h>

using namespace std;

#ifdef __MPF_BUILD__
namespace mpf_sdp {
#else
namespace sdp {
#endif

/*
  add_variable

  Checks whether a given variable has already been added to the
  container; if not then it is added.
*/
void SDPConstraintContainer::add_variable(const SDPVariable& X)
{
  if (X.idx() >= data->variables_contained.size())
    data->variables_contained.resize(X.idx() + 1, false);

  if (!data->variables_contained[X.idx()]) {
    data->variables_contained[X.idx()] = true;
    data->variables.push_back(X);
  }
}

/*
  do_add_constraint

  In this function we do all the work of adding polynomial
  constraints. Functions add_constraint then provide wrappers for
  this function, depending on whether a progress indicator is
  specified or not.
*/
void SDPConstraintContainer::do_add_constraint(const SDPPolynomialConstraint& C, 
                                               const SparseBasisCoefficients& sc,
                                               SDPProgressIndicator *progress)
{
  int d = sc.basis_size();
  vector<Number> coeffs_upper(d), coeffs_lower(d);
  VectorMask mask_upper(d), mask_lower(d);
  vector<MatrixEntryList> degree_entries(d);
  double output_eps = env.output_eps();

  /* First we determine the maximum value for the progress indicator */
  if (progress) {
    int max_value = 0;
    
    for (auto& s : C.lhs()) {
      const Matrix<Polynomial>& A = s.second.matrix();
      
      if (!s.second.mask())
        max_value += ((1 + A.num_rows()) * A.num_rows()) / 2;
      else
        max_value += s.second.mask()->on_count();
    }

    progress->set_max(max_value);
  }

  /* We add the variables that occur in the constraint to the container */
  for (auto& s : C.lhs())
    add_variable(s.first);
  
  /* We process the constraint */
  if (progress) progress->start();

  for (auto& s : C.lhs()) {
    const Matrix<Polynomial>& A = s.second.matrix();
    const MatrixMask *mask = s.second.mask();
    
    if (!mask) {
      for (int i = 0; i < A.num_rows(); i++)
        for (int j = i; j < A.num_cols(); j++) {
          if (!sc.basis_coefficients(A[i][j], coeffs_upper, mask_upper))
            throw InvalidBasis(InvalidBasis::representation_error, ERROR_INFO);

          if (i != j && !sc.basis_coefficients(A[j][i], coeffs_lower, mask_lower))
            throw InvalidBasis(InvalidBasis::representation_error, ERROR_INFO);

          /* Processes upper mask */
          for (int k : mask_upper)
            if (i == j)
              degree_entries[k].add_entry(output_eps, s.first.idx(),
					  i, j, coeffs_upper[k]);
            else if (mask_lower.is_on(k))
              degree_entries[k].add_entry(output_eps, s.first.idx(),
					  i, j, (coeffs_upper[k] + coeffs_lower[k]) / 2);
            else 
              degree_entries[k].add_entry(output_eps, s.first.idx(),
					  i, j, coeffs_upper[k] / 2);

          /* Processes lower mask */
          if (i != j)
            for (int k : mask_lower)
              if (!mask_upper.is_on(k))  // The case mask_upper.is_on(k) is already treated above!
                degree_entries[k].add_entry(output_eps, s.first.idx(),
					    i, j, coeffs_lower[k] / 2);
	
          if (progress) ++(*progress);
        }
    }
    else {
      for (auto& p : *mask) {
        int i, j;
        mask_upper.clear();
        mask_lower.clear();

        if (p.i <= p.j) {
          i = p.i; j = p.j;

          if (!sc.basis_coefficients(A[i][j], coeffs_upper, mask_upper))
            throw InvalidBasis(InvalidBasis::representation_error, ERROR_INFO);

          if (i != j && mask->is_on(j, i))
            if (!sc.basis_coefficients(A[j][i], coeffs_lower, mask_lower))
              throw InvalidBasis(InvalidBasis::representation_error, ERROR_INFO);
        }
        else {
          i = p.j; j = p.i;  // We want to always have i <= j

          if (!mask->is_on(i, j))
            if (!sc.basis_coefficients(A[j][i], coeffs_lower, mask_lower))
              throw InvalidBasis(InvalidBasis::representation_error, ERROR_INFO);

	  /* If mask->is_on(i, j), then we already took care of things
	     in the previous case. In this case, both mask_upper and
	     mask_lower will be empty and everything works fine.
	  */
        }

        assert(i <= j);

        /* Processes upper mask */
        for (int k : mask_upper)
          if (i == j)
            degree_entries[k].add_entry(output_eps, s.first.idx(),
					i, j, coeffs_upper[k]);
          else if (mask_lower.is_on(k))
            degree_entries[k].add_entry(output_eps, s.first.idx(),
					i, j, (coeffs_upper[k] + coeffs_lower[k]) / 2);
          else 
            degree_entries[k].add_entry(output_eps, s.first.idx(),
					i, j, coeffs_upper[k] / 2);
        
        /* Processes lower mask */
        for (int k : mask_lower)  // Lower mask is empty if i == j
          if (!mask_upper.is_on(k))  // The case mask_upper.is_on(k) is already treated above!
            degree_entries[k].add_entry(output_eps, s.first.idx(),
					i, j, coeffs_lower[k] / 2);
	
        if (progress) ++(*progress);
      }
    }
  }
  
  if (progress) progress->finish();
  
  if (!sc.basis_coefficients(C.rhs(), coeffs_upper, mask_upper))
    throw InvalidBasis(InvalidBasis::representation_error, ERROR_INFO);
  
  for (int k = 0; k < d; k++)
    if (degree_entries[k].empty()) {
      if (mask_upper.is_on(k) && NumberOperations::abs(coeffs_upper[k]) > output_eps)
	throw InfeasibleConstraint(ERROR_INFO);
    }
    else {
      if (mask_upper.is_on(k))
	data->constraints.push_back(RawConstraint(sdp_eq, coeffs_upper[k]));
      else
	data->constraints.push_back(RawConstraint(sdp_eq, 0));
      
      data->constraints.back().entries.swap(degree_entries[k]);
    }
}


/*
  Several versions of add_constraint.
*/

void SDPConstraintContainer::add_constraint(const SDPLinearConstraint& C)
{
  double output_eps = env.output_eps();

  data->constraints.push_back(RawConstraint(C.relation(), C.rhs()));
  RawConstraint& cur = data->constraints.back();

  for (auto& s : C.lhs()) {
    add_variable(s.first);
    cur.entries.add_matrix_entries(output_eps, s.first.idx(), s.second);
  }

  if (cur.entries.empty()) {
    data->constraints.pop_back();

    if (C.relation() == sdp_leq && C.rhs() < -output_eps)
      throw InfeasibleConstraint(ERROR_INFO);
    else if (C.relation() == sdp_geq && C.rhs() > output_eps)
      throw InfeasibleConstraint(ERROR_INFO);
    else if (C.relation() == sdp_eq && NumberOperations::abs(C.rhs()) > output_eps)
      throw InfeasibleConstraint(ERROR_INFO);
  }
  else if (C.relation() != sdp_eq)
    _num_ineqs++;
}

void SDPConstraintContainer::add_constraint(const SDPPolynomialConstraint& C)
{
  do_add_constraint(C, SparseMonomialBasis(C), nullptr);
}

void SDPConstraintContainer::add_constraint(const SDPPolynomialConstraint& C, 
					    SDPProgressIndicator& progress)
{
  do_add_constraint(C, SparseMonomialBasis(C), &progress);
}

void SDPConstraintContainer::add_constraint(const SDPPolynomialConstraint& C, const SparseBasisCoefficients& sc)
{
  do_add_constraint(C, sc, 0);
}

void SDPConstraintContainer::add_constraint(const SDPPolynomialConstraint& C, 
					    const SparseBasisCoefficients& sc, 
					    SDPProgressIndicator& progress)
{
  do_add_constraint(C, sc, &progress);
}

} // namespace
